@@ -1,6 +1,6 @@
 /*
  * ProFTPD - mod_sftp 'publickey' user authentication
- * Copyright (c) 2008-2010 TJ Saunders
+ * Copyright (c) 2008-2011 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,14 +14,14 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders and other respective copyright holders
  * give permission to link this program with OpenSSL, and distribute the
  * resulting executable, without including the source code for OpenSSL in the
  * source distribution.
  *
- * $Id: auth-publickey.c,v 1.4.2.1 2010/06/17 17:57:08 castaglia Exp $
+ * $Id: auth-publickey.c,v 1.9 2011/08/04 21:15:19 castaglia Exp $
  */
 
 #include "mod_sftp.h"
@@ -73,14 +73,27 @@ static int send_pubkey_ok(const char *algo, const char *pubkey_data,
   return 0;
 }
 
-int sftp_auth_publickey(struct ssh2_packet *pkt, const char *orig_user,
-    const char *user, const char *service, char **buf, uint32_t *buflen,
-    int *send_userauth_fail) {
+int sftp_auth_publickey(struct ssh2_packet *pkt, cmd_rec *pass_cmd,
+    const char *orig_user, const char *user, const char *service, char **buf,
+    uint32_t *buflen, int *send_userauth_fail) {
   int have_signature, pubkey_type;
   char *pubkey_algo = NULL, *pubkey_data;
   uint32_t pubkey_len;
   struct passwd *pw;
 
+  if (pr_cmd_dispatch_phase(pass_cmd, PRE_CMD, 0) < 0) {
+    (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+      "authentication request for user '%s' blocked by '%s' handler",
+      orig_user, pass_cmd->argv[0]);
+
+    pr_cmd_dispatch_phase(pass_cmd, POST_CMD_ERR, 0);
+    pr_cmd_dispatch_phase(pass_cmd, LOG_CMD_ERR, 0);
+
+    *send_userauth_fail = TRUE;
+    errno = EPERM;
+    return 0;
+  }
+
   have_signature = sftp_msg_read_bool(pkt->pool, buf, buflen);
 
   if (sftp_interop_supports_feature(SFTP_SSH2_FEAT_HAVE_PUBKEY_ALGO)) {
@@ -90,7 +103,7 @@ int sftp_auth_publickey(struct ssh2_packet *pkt, const char *orig_user,
   pubkey_data = sftp_msg_read_data(pkt->pool, buf, buflen, pubkey_len);
 
   if (pubkey_algo == NULL) {
-    /* The client did not send the string identify the public key algorithm.
+    /* The client did not send the string identifying the public key algorithm.
      * Thus we need to extract the algorithm string from the public key data.
      */
     pubkey_algo = sftp_msg_read_string(pkt->pool, &pubkey_data, &pubkey_len);
@@ -99,12 +112,20 @@ int sftp_auth_publickey(struct ssh2_packet *pkt, const char *orig_user,
   pr_trace_msg(trace_channel, 9, "client sent '%s' public key %s",
     pubkey_algo, have_signature ? "with signature" : "without signature");
 
-  if (strcmp(pubkey_algo, "ssh-rsa") == 0) {
+  if (strncmp(pubkey_algo, "ssh-rsa", 8) == 0) {
     pubkey_type = EVP_PKEY_RSA;
 
-  } else if (strcmp(pubkey_algo, "ssh-dss") == 0) {
+  } else if (strncmp(pubkey_algo, "ssh-dss", 8) == 0) {
     pubkey_type = EVP_PKEY_DSA;
 
+  /* XXX This is where we would add support for X509 public keys, e.g.:
+   *
+   *  x509v3-ssh-dss
+   *  x509v3-ssh-rsa
+   *  x509v3-sign (older)
+   *
+   */
+
   } else {
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
       "unsupported public key algorithm '%s' requested, rejecting request",
@@ -161,6 +182,12 @@ int sftp_auth_publickey(struct ssh2_packet *pkt, const char *orig_user,
     char *buf2, *ptr2, *signature_data;
     uint32_t buflen2, bufsz2, id_len, signature_len;
 
+    /* XXX This should become a more generic "is this key data
+     * usable/acceptable?" check (and take the pubkey_type parameter), so that
+     * that is where we would check the validity/usability of an X509v3 cert
+     * (if a cert), or if the key is on the blacklist (if a key).
+     */
+
     if (sftp_blacklist_reject_key(pkt->pool, pubkey_data, pubkey_len)) {
       *send_userauth_fail = TRUE;
       errno = EPERM;
@@ -176,6 +203,15 @@ int sftp_auth_publickey(struct ssh2_packet *pkt, const char *orig_user,
      * done authenticating.
      */
 
+    /* XXX Need to pass the pubkey_type here as well, so that the
+     * verification routines can handle different databases of keys/certs.
+     * 
+     * For X509v3 certs, we will want a way to enforce/restrict which
+     * user names can be used with the provided cert.  Perhaps a database
+     * mapping cert fingerprints to user names/UIDs?  Configurable callback
+     * check (HOOK?), for modules to enforce.
+     */
+
     if (sftp_keystore_verify_user_key(pkt->pool, user, pubkey_data,
         pubkey_len) < 0) {
       *send_userauth_fail = TRUE;
