@@ -2,7 +2,7 @@
  * ProFTPD: mod_sql_postgres -- Support for connecting to Postgres databases.
  * Time-stamp: <1999-10-04 03:21:21 root>
  * Copyright (c) 2001 Andrew Houghton
- * Copyright (c) 2004-2009 TJ Saunders
+ * Copyright (c) 2004-2011 TJ Saunders
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,22 +16,18 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Andrew Houghton and other respective copyright
  * holders give permission to link this program with OpenSSL, and distribute
  * the resulting executable, without including the source code for OpenSSL in
  * the source distribution.
  *
- * $Id: mod_sql_postgres.c,v 1.37 2009/02/05 18:53:54 castaglia Exp $
- */
-
-/*
+ * $Id: mod_sql_postgres.c,v 1.54 2011/10/06 15:56:46 castaglia Exp $
  * $Libraries: -lm -lpq $
  */
 
-/* 
- * Internal define used for debug and logging.  All backends are encouraged
+/* Internal define used for debug and logging.  All backends are encouraged
  * to use the same format.
  */
 #define MOD_SQL_POSTGRES_VERSION	"mod_sql_postgres/4.0.4"
@@ -56,6 +52,11 @@
  */
 #ifdef PR_USE_NLS
 extern const char *pg_encoding_to_char(int encoding);
+
+/* And this is the mod_sql_postgres-specific function mapping iconv
+ * locales to Postgres encodings.
+ */
+static const char *get_postgres_encoding(const char *encoding);
 #endif
 
 /* 
@@ -176,7 +177,7 @@ static void _sql_check_cmd(cmd_rec *cmd, char *msg) {
       ": '%s' was passed an invalid cmd_rec. Shutting down.", msg);
     sql_log(DEBUG_WARN, "'%s' was passed an invalid cmd_rec. Shutting down.",
       msg);
-    end_login(1);
+    pr_session_end(0);
   }    
 
   return;
@@ -257,6 +258,86 @@ static modret_t *_build_data(cmd_rec *cmd, db_conn_t *conn) {
   return mod_create_data( cmd, (void *) sd );
 }
 
+#ifdef PR_USE_NLS
+static const char *get_postgres_encoding(const char *encoding) {
+
+  /* XXX Hack to deal with Postgres' incredibly broken behavior when
+   * handling the 'ASCII' encoding.  Specifically, Postgres chokes on
+   * 'ASCII', and instead insists on calling it 'SQL_ASCII' (which, of
+   * course, is not even close to being a valid encoding name according
+   * to libiconv.)
+   *
+   * Treat 'ANSI_X3.4-1968' (another common name/alias for ASCII) the
+   * same; rename it to 'SQL_ASCII' for Postgres' benefit.  And same for
+   * 'US-ASCII'.
+   */
+  if (strcasecmp(encoding, "ANSI_X3.4-1968") == 0 ||
+      strcasecmp(encoding, "ASCII") == 0 ||
+      strcasecmp(encoding, "US-ASCII") == 0) {
+    return "SQL_ASCII";
+  }
+
+  /* And other commonly used charsets for which Postgres has their own names. */
+
+  if (strcasecmp(encoding, "CP1251") == 0 ||
+      strcasecmp(encoding, "WINDOWS-1251") == 0) {
+    return "WIN1251";
+  }
+
+  if (strcasecmp(encoding, "KOI-8") == 0 ||
+      strcasecmp(encoding, "KOI8-R") == 0 ||
+      strcasecmp(encoding, "KOI8") == 0 ||
+      strcasecmp(encoding, "KOI8R") == 0) {
+    return "KOI";
+  }
+
+  if (strcasecmp(encoding, "CP866") == 0) {
+    return "WIN866";
+  }
+
+  if (strcasecmp(encoding, "ISO-8859-1") == 0) {
+    return "LATIN1";
+  }
+
+  if (strcasecmp(encoding, "ISO-8859-15") == 0) {
+    return "LATIN9";
+  }
+
+  if (strcasecmp(encoding, "EUC-CN") == 0 ||
+      strcasecmp(encoding, "EUCCN") == 0) {
+    return "EUC_CN";
+  }
+
+  if (strcasecmp(encoding, "EUC-JP") == 0 ||
+      strcasecmp(encoding, "EUCJP") == 0) {
+    return "EUC_JP";
+  }
+
+  if (strcasecmp(encoding, "EUC-KR") == 0 ||
+      strcasecmp(encoding, "EUCKR") == 0) {
+    return "EUC_KR";
+  }
+
+  if (strcasecmp(encoding, "EUC-TW") == 0 ||
+      strcasecmp(encoding, "EUCTW") == 0) {
+    return "EUC_TW";
+  }
+
+  if (strcasecmp(encoding, "SHIFT-JIS") == 0 ||
+      strcasecmp(encoding, "SHIFT_JIS") == 0) {
+    return "SJIS";
+  }
+
+  if (strcasecmp(encoding, "UTF8") == 0 ||
+      strcasecmp(encoding, "UTF-8") == 0 ||
+      strcasecmp(encoding, "UTF8-MAC") == 0) {
+    return "UTF8";
+  }
+
+  return encoding;
+}
+#endif
+
 /*
  * cmd_open: attempts to open a named connection to the database.
  *
@@ -277,6 +358,7 @@ static modret_t *_build_data(cmd_rec *cmd, db_conn_t *conn) {
 MODRET cmd_open(cmd_rec *cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
+  const char *server_version = NULL;
 
   sql_log(DEBUG_FUNC, "%s", "entering \tpostgres cmd_open");
 
@@ -300,16 +382,77 @@ MODRET cmd_open(cmd_rec *cmd) {
   /* if we're already open (connections > 0) increment connections 
    * reset our timer if we have one, and return HANDLED 
    */
-  if ((entry->connections > 0) && 
-      (PQstatus(conn->postgres) == CONNECTION_OK)) {
-    entry->connections++;
-    if (entry->timer) 
-      pr_timer_reset(entry->timer, &sql_postgres_module);
+  if (entry->connections > 0) { 
+    if (PQstatus(conn->postgres) == CONNECTION_OK) {
+      entry->connections++;
 
-    sql_log(DEBUG_INFO, "connection '%s' count is now %d", entry->name,
-      entry->connections);
-    sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
-    return PR_HANDLED(cmd);
+      if (entry->timer) {
+        pr_timer_reset(entry->timer, &sql_postgres_module);
+      }
+
+      sql_log(DEBUG_INFO, "connection '%s' count is now %d", entry->name,
+        entry->connections);
+      sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
+      return PR_HANDLED(cmd);
+
+    } else {
+      char *reason;
+      size_t reason_len;
+
+      /* Unless we've been told not to reconnect, try to reconnect now.
+       * We only try once; if it fails, we return an error.
+       */
+      if (!(pr_sql_opts & SQL_OPT_NO_RECONNECT)) {
+        PQreset(conn->postgres);
+
+        if (PQstatus(conn->postgres) == CONNECTION_OK) {
+          entry->connections++;
+
+          if (entry->timer) {
+            pr_timer_reset(entry->timer, &sql_postgres_module);
+          }
+
+          sql_log(DEBUG_INFO, "connection '%s' count is now %d", entry->name,
+            entry->connections);
+          sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
+          return PR_HANDLED(cmd);
+        }
+      }
+
+      reason = PQerrorMessage(conn->postgres);
+      reason_len = strlen(reason);
+
+      /* Postgres might give us an empty string as the reason; not helpful. */
+      if (reason_len == 0) {
+        reason = "(unknown)";
+        reason_len = strlen(reason);
+      }
+
+      /* The error message returned by Postgres is usually appended with
+       * a newline.  Let's prettify it by removing the newline.  Note
+       * that yes, we are overwriting the pointer given to us by Postgres,
+       * but it's OK.  The Postgres docs say that we're not supposed to
+       * free the memory associated with the returned string anyway.
+       */
+      reason = pstrdup(session.pool, reason);
+
+      if (reason[reason_len-1] == '\n') {
+        reason[reason_len-1] = '\0';
+        reason_len--;
+      }
+
+      sql_log(DEBUG_INFO, "lost connection to database: %s", reason);
+
+      entry->connections = 0;
+      if (entry->timer) {
+        pr_timer_remove(entry->timer, &sql_postgres_module);
+        entry->timer = 0;
+      }
+
+      sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
+      return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+        "lost connection to database");
+    }
   }
 
   /* make sure we have a new conn struct */
@@ -321,25 +464,20 @@ MODRET cmd_open(cmd_rec *cmd) {
     return _build_error( cmd, conn );
   }
 
+#if defined(PG_VERSION_STR)
+  sql_log(DEBUG_FUNC, "Postgres client: %s", PG_VERSION_STR);
+#endif
+
+  server_version = PQparameterStatus(conn->postgres, "server_version");
+  if (server_version != NULL) {
+    sql_log(DEBUG_FUNC, "Postgres server version: %s", server_version);
+  }
+
 #ifdef PR_USE_NLS
   if (pr_encode_get_encoding() != NULL) {
-    const char *encoding = pr_encode_get_encoding();
+    const char *encoding;
 
-    /* XXX Hack to deal with Postgres' incredibly broken behavior when
-     * handling the 'ASCII' encoding.  Specifically, Postgres chokes on
-     * 'ASCII', and instead insists on calling it 'SQL_ASCII' (which, of
-     * course, is not even close to being a valid encoding name according
-     * to libiconv.)
-     *
-     * Treat 'ANSI_X3.4-1968' (another common name/alias for ASCII) the
-     * same; rename it to 'SQL_ASCII' for Postgres' benefit.  And same for
-     * 'US-ASCII'.
-     */
-    if (strcmp(encoding, "ANSI_X3.4-1968") == 0 ||
-        strcmp(encoding, "ASCII") == 0 ||
-        strcmp(encoding, "US-ASCII") == 0) {
-      encoding = "SQL_ASCII";
-    }
+    encoding = get_postgres_encoding(pr_encode_get_encoding());
 
     /* Configure the connection for the current local character set. */
     if (PQsetClientEncoding(conn->postgres, encoding) < 0) {
@@ -350,21 +488,32 @@ MODRET cmd_open(cmd_rec *cmd) {
 
     sql_log(DEBUG_FUNC, "Postgres connection character set now '%s' "
       "(from '%s')", pg_encoding_to_char(PQclientEncoding(conn->postgres)),
-      pr_encode_get_charset());
+      pr_encode_get_encoding());
   }
 #endif /* !PR_USE_NLS */
 
   /* bump connections */
   entry->connections++;
 
-  /* set up our timer if necessary */
-  if (entry->ttl > 0) {
+  if (pr_sql_conn_policy == SQL_CONN_POLICY_PERSESSION) {
+    /* If the connection policy is PERSESSION... */
+    if (entry->connections == 1) {
+      /* ...and we are actually opening the first connection to the database;
+       * we want to make sure this connection stays open, after this first use
+       * (as per Bug#3290).  To do this, we re-bump the connection count.
+       */
+      entry->connections++;
+    } 
+ 
+  } else if (entry->ttl > 0) { 
+    /* Set up our timer if necessary */
+
     entry->timer = pr_timer_add(entry->ttl, -1, &sql_postgres_module,
       sql_timer_cb, "postgres connection ttl");
     sql_log(DEBUG_INFO, "connection '%s' - %d second timer started",
       entry->name, entry->ttl);
 
-    /* timed connections get re-bumped so they don't go away when cmd_close
+    /* Timed connections get re-bumped so they don't go away when cmd_close
      * is called.
      */
     entry->connections++;
@@ -574,10 +723,15 @@ MODRET cmd_defineconnection(cmd_rec *cmd) {
       "named connection already exists");
   }
 
-  entry->ttl = (cmd->argc == 5) ? 
-    (int) strtol(cmd->argv[4], (char **)NULL, 10) : 0;
-  if (entry->ttl < 0) 
-    entry->ttl = 0;
+  if (cmd->argc == 5) { 
+    entry->ttl = (int) strtol(cmd->argv[4], (char **) NULL, 10);
+    if (entry->ttl >= 1) {
+      pr_sql_conn_policy = SQL_CONN_POLICY_TIMER;
+ 
+    } else {
+      entry->ttl = 0;
+    }
+  }
 
   entry->timer = 0;
   entry->connections = 0;
@@ -1152,6 +1306,11 @@ MODRET cmd_escapestring(cmd_rec * cmd) {
   modret_t *cmr = NULL;
   char *unescaped = NULL;
   char *escaped = NULL;
+  cmd_rec *close_cmd;
+  size_t unescaped_len = 0;
+#ifdef HAVE_POSTGRES_PQESCAPESTRINGCONN
+  int pgerr = 0;
+#endif
 
   sql_log(DEBUG_FUNC, "%s", "entering \tpostgres cmd_escapestring");
 
@@ -1179,14 +1338,24 @@ MODRET cmd_escapestring(cmd_rec * cmd) {
     return cmr;
   }
 
-  /* Note: the PQescapeString() function appeared in the C API as of
-   * Postgres-7.2.
-   */
   unescaped = cmd->argv[1];
+  unescaped_len = strlen(unescaped);
   escaped = (char *) pcalloc(cmd->tmp_pool, sizeof(char) *
-    (strlen(unescaped) * 2) + 1);
+    (unescaped_len * 2) + 1);
 
-  PQescapeString(escaped, unescaped, strlen(unescaped));
+#ifdef HAVE_POSTGRES_PQESCAPESTRINGCONN
+  PQescapeStringConn(conn->postgres, escaped, unescaped, unescaped_len, &pgerr);
+  if (pgerr != 0) {
+    sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
+    return _build_error(cmd, conn);
+  }
+#else
+  PQescapeString(escaped, unescaped, unescaped_len);
+#endif
+
+  close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
+  cmd_close(close_cmd);
+  SQL_FREE_CMD(close_cmd);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
   return mod_create_data(cmd, (void *) escaped);
@@ -1237,11 +1406,14 @@ MODRET cmd_checkauth(cmd_rec * cmd) {
 
   conn = (db_conn_t *) entry->data;
 
+  sql_log(DEBUG_WARN, MOD_SQL_POSTGRES_VERSION
+    ": Postgres does not support the 'Backend' SQLAuthType");
+
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_checkauth");
 
   /* PostgreSQL doesn't provide this functionality */
-
-  return PR_ERROR(cmd);
+  return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+    "Postgres does not support the 'Backend' SQLAuthType");
 }
 
 /*
@@ -1287,8 +1459,11 @@ MODRET cmd_prepare(cmd_rec *cmd) {
   }
 
   conn_pool = (pool *) cmd->argv[0];
-  conn_cache = make_array((pool *) cmd->argv[0], DEF_CONN_POOL_SIZE,
-    sizeof(conn_entry_t));
+
+  if (conn_cache == NULL) {
+    conn_cache = make_array((pool *) cmd->argv[0], DEF_CONN_POOL_SIZE,
+      sizeof(conn_entry_t *));
+  }
 
   return mod_create_data(cmd, NULL);
 }
@@ -1336,7 +1511,7 @@ static void sql_postgres_mod_load_ev(const void *event_data,
     if (sql_register_backend("postgres", sql_postgres_cmdtable) < 0) {
       pr_log_pri(PR_LOG_NOTICE, MOD_SQL_POSTGRES_VERSION
         ": notice: error registering backend: %s", strerror(errno));
-      end_login(1);
+      pr_session_end(0);
     }
   }
 }
@@ -1349,7 +1524,7 @@ static void sql_postgres_mod_unload_ev(const void *event_data,
     if (sql_unregister_backend("postgres") < 0) {
       pr_log_pri(PR_LOG_NOTICE, MOD_SQL_POSTGRES_VERSION
         ": notice: error unregistering backend: %s", strerror(errno));
-      end_login(1);
+      pr_session_end(0);
     }
 
     /* Unregister ourselves from all events. */
@@ -1372,9 +1547,15 @@ static int sql_postgres_init(void) {
 }
 
 static int sql_postgres_sess_init(void) {
-  conn_pool = make_sub_pool(session.pool);
-  conn_cache = make_array(make_sub_pool(session.pool), DEF_CONN_POOL_SIZE,
-    sizeof(conn_entry_t));
+  if (conn_pool == NULL) {
+    conn_pool = make_sub_pool(session.pool);
+    pr_pool_tag(conn_pool, "Postgres connection pool");
+  }
+
+  if (conn_cache == NULL) {
+    conn_cache = make_array(make_sub_pool(session.pool), DEF_CONN_POOL_SIZE,
+      sizeof(conn_entry_t *));
+  }
 
   return 0;
 }
