@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,14 +23,15 @@
  */
 
 /* Feature list management
- * $Id: feat.h,v 1.3 2003/03/09 02:07:00 castaglia Exp $
+ * $Id: feat.h,v 1.5 2011/05/23 20:35:35 castaglia Exp $
  */
 
 #ifndef PR_FEAT_H
 #define PR_FEAT_H
 
-void pr_feat_add(const char *);
+int pr_feat_add(const char *);
 const char *pr_feat_get(void);
 const char *pr_feat_get_next(void);
+int pr_feat_remove(const char *);
 
 #endif /* PR_FEAT_H */
