@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -25,8 +25,7 @@
  */
 
 /* ProFTPD internal implemenation of syslog(3) routines
- *
- * $Id: pr-syslog.h,v 1.4 2003/07/25 01:21:58 jwm Exp $
+ * $Id: pr-syslog.h,v 1.6 2011/05/23 20:35:35 castaglia Exp $
  */
 
 #include "conf.h"
@@ -83,6 +82,9 @@ int pr_openlog(const char *ident, int option, int facility);
 /* Set the log mask level.  */
 int pr_setlogmask(int mask);
 
+/* Set the facility of the system logger.  */
+int pr_setlogfacility(int facility);
+
 /* Generate a log message using the given format string and option arguments.
  */
 void pr_syslog(int sockfd, int pri, const char *fmt, ...);
