@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003, 2004 The ProFTPD Project team
+ * Copyright (c) 2001-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -25,7 +25,7 @@
  */
 
 /* General options
- * $Id: proftpd.h,v 1.53 2005/02/26 17:28:58 castaglia Exp $
+ * $Id: proftpd.h,v 1.71 2011/09/21 05:40:04 castaglia Exp $
  */
 
 #ifndef PR_PROFTPD_H
@@ -63,6 +63,7 @@ typedef int (*callback_t)(CALLBACK_FRAME);
 struct conn_struc;
 struct cmd_struc;
 struct config_struc;
+struct modret_struc;
 
 typedef struct {
   pool *pool;
@@ -73,9 +74,6 @@ typedef struct {
   pr_netaddr_t data_addr;		/* Remote data address */
   unsigned short data_port;		/* Remote data port */
 
-  unsigned char ident_lookups;		/* Is RFC931 (ident) protocol used? */
-  char *ident_user;			/* User identified by ident protocol */
-
   const char *auth_mech;		/* Name of the authentication
                                          * module/mechanism that successfully
                                          * authenticated the client
@@ -92,6 +90,9 @@ typedef struct {
                                          * for current operation
                                          */
 
+  /* Process ID of the session */
+  pid_t pid;
+
   /* The uid/gids are manipulated by the PRIVS macros in
    * privs.h
    */
@@ -123,6 +124,8 @@ typedef struct {
   char *proc_prefix;			/* The "prefix" of our process name */
 
   int wtmp_log;				/* Are we logging to wtmp? */
+  int multiline_rfc2228;		/* Are we using RFC2228-style multiline responses ? */
+
   struct conn_struc *c;			/* Control connection */
   struct conn_struc *d;			/* Data connection */
 
@@ -134,6 +137,9 @@ typedef struct {
   char *anon_user;			/* E-mail address sent to us */
 
   char *curr_cmd;                       /* Current FTP command */
+  int curr_cmd_id;			/* Current FTP command ID */
+  struct cmd_struc *curr_cmd_rec;       /* Current command */
+
   int curr_phase;                       /* Current handler phase */
 
   off_t restart_pos;			/* Restart marked position */
@@ -176,6 +182,24 @@ typedef struct {
    */
   unsigned int total_files_xfer;
 
+  /* Total number of "raw" protocol bytes read in from the network for
+   * the session.
+   */
+  off_t total_raw_in;
+
+  /* Total number of "raw" protocol bytes written out to the network for
+   * the session.
+   */
+  off_t total_raw_out;
+
+  /* Reason code for end of session/disconnection; in reality, the values
+   * come from the pr_disconnect_reason_e enum in session.h.
+   */
+  int disconnect_reason;
+
+  /* Module which disconnected/ended the session */
+  struct module_struc *disconnect_module;
+
 } session_t;
 
 /* Daemon identity values, defined in main.c */
@@ -192,8 +216,6 @@ extern array_header *daemon_gids;
 extern session_t	session;
 extern char AddressCollisionCheck;
 extern char ServerType;
-extern char MultilineRFC2228;
-extern const char	*pwdfname,*grpfname;
 
 /* Session/State flags */
 
@@ -226,7 +248,7 @@ extern const char	*pwdfname,*grpfname;
 #define SERVER_STANDALONE	1
 
 /* Signals */
-#define RECEIVED_SIG_REHASH	0x0001
+#define RECEIVED_SIG_RESTART	0x0001
 #define RECEIVED_SIG_EXIT	0x0002
 #define RECEIVED_SIG_SHUTDOWN	0x0004
 #define RECEIVED_SIG_SEGV	0x0008
@@ -239,19 +261,37 @@ extern const char	*pwdfname,*grpfname;
 #define RECEIVED_SIG_ALRM	0x0400
 
 /* Timers */
-#define TIMER_LOGIN		1
-#define TIMER_IDLE		2
-#define TIMER_NOXFER		3
-#define TIMER_STALLED		4
-#define TIMER_SESSION		5
+#define PR_TIMER_LOGIN		1
+#define PR_TIMER_IDLE		2
+#define PR_TIMER_NOXFER		3
+#define PR_TIMER_STALLED	4
+#define PR_TIMER_SESSION	5
+
+/* Developer code */
+
+#ifdef PR_DEVEL_TIMING
+# define PR_DEVEL_CLOCK(code) \
+  { \
+    int local_errno; \
+    struct timeval local_before, local_after, local_since; \
+    timerclear(&local_before); \
+    timerclear(&local_after); \
+    timerclear(&local_since); \
+    (void) gettimeofday(&local_before, NULL); \
+    (code); \
+    local_errno = errno; \
+    (void) gettimeofday(&local_after, NULL); \
+    timersub(&local_after, &local_before, &local_since); \
+    (void) pr_trace_msg("timing", 9, "code at %s:%d took %lu sec, %lu usec", __FILE__, __LINE__, (unsigned long) local_since.tv_sec, (unsigned long) local_since.tv_usec); \
+    errno = local_errno; \
+  }
+#else
+# define PR_DEVEL_CLOCK(code)   (code)
+#endif /* PR_DEVEL_TIMING */
 
 /* Misc Prototypes */
-
-void end_login(int);
 void pr_signals_handle(void);
 void session_exit(int, void *, int, void *);
-void session_set_idle(void);
-void pr_rehash_register_handler(void *, void(*)(void *));
 void set_daemon_rlimits(void);
 void set_session_rlimits(void);
 
