@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project and other respective copyright
  * holders give permission to link this program with OpenSSL, and distribute
@@ -22,9 +22,8 @@
  * the source distribution.
  */
 
-/* Scoreboard routines.
- *
- * $Id: scoreboard.h,v 1.13 2004/11/02 18:18:58 castaglia Exp $
+/* Scoreboard routines
+ * $Id: scoreboard.h,v 1.21 2011/05/23 20:35:35 castaglia Exp $
  */
 
 #ifndef PR_SCOREBOARD_H
@@ -32,7 +31,7 @@
 
 /* PR_SCOREBOARD_VERSION is used for checking for scoreboard compatibility
  */
-#define PR_SCOREBOARD_VERSION        		0x01040002
+#define PR_SCOREBOARD_VERSION        		0x01040003
 
 /* Structure used as a header for scoreboard files.
  */
@@ -74,14 +73,30 @@ typedef struct {
   char sce_client_name[PR_TUNABLE_SCOREBOARD_BUFFER_SIZE];
 
   char sce_class[32];
+  char sce_protocol[32];
   char sce_cwd[PR_TUNABLE_SCOREBOARD_BUFFER_SIZE];
 
-  char sce_cmd[5];
+  char sce_cmd[65];
   char sce_cmd_arg[PR_TUNABLE_SCOREBOARD_BUFFER_SIZE];
 
   time_t sce_begin_idle, sce_begin_session;
 
-  off_t sce_xfer_size, sce_xfer_done, sce_xfer_len;
+  /* Records the number of bytes to be transferred, and the number of bytes
+   * transferred so far.  These two numbers are used to calculate the
+   * percentage of data transferred in the ftptop/ftpwho utilities.
+   *
+   * Note that for uploads, we do not know the full size of the data being
+   * uploaded, hence we cannot show a percentage; we can only display/record
+   * how many bytes have been transferred so far.
+   */
+  off_t sce_xfer_size;
+  off_t sce_xfer_done;
+
+  /* Records the number of bytes transferred, and the elapsed time.  These
+   * two fields are used to calculate the transfer rate as displayed by
+   * the ftptop/ftpwho utilities.
+   */
+  off_t sce_xfer_len;
   unsigned long sce_xfer_elapsed;
 
 } pr_scoreboard_entry_t;
@@ -106,6 +121,7 @@ typedef struct {
 #define PR_SCORE_BEGIN_SESSION	14
 #define PR_SCORE_XFER_LEN	15
 #define PR_SCORE_XFER_ELAPSED	16
+#define PR_SCORE_PROTOCOL	17
 
 /* Scoreboard error values */
 #define PR_SCORE_ERR_BAD_MAGIC		-2
@@ -113,7 +129,9 @@ typedef struct {
 #define PR_SCORE_ERR_NEWER_VERSION	-4
 
 const char *pr_get_scoreboard(void);
+const char *pr_get_scoreboard_mutex(void);
 int pr_set_scoreboard(const char *);
+int pr_set_scoreboard_mutex(const char *);
 
 int pr_close_scoreboard(void);
 void pr_delete_scoreboard(void);
@@ -121,11 +139,15 @@ int pr_open_scoreboard(int);
 int pr_restore_scoreboard(void);
 int pr_rewind_scoreboard(void);
 
-int pr_scoreboard_add_entry(void);
-int pr_scoreboard_del_entry(unsigned char);
 pid_t pr_scoreboard_get_daemon_pid(void);
 time_t pr_scoreboard_get_daemon_uptime(void);
-pr_scoreboard_entry_t *pr_scoreboard_read_entry(void);
-int pr_scoreboard_update_entry(pid_t, ...);
+int pr_scoreboard_scrub(void);
+
+int pr_scoreboard_entry_add(void);
+int pr_scoreboard_entry_del(unsigned char);
+pr_scoreboard_entry_t *pr_scoreboard_entry_read(void);
+const char *pr_scoreboard_entry_get(int);
+int pr_scoreboard_entry_kill(pr_scoreboard_entry_t *, int);
+int pr_scoreboard_entry_update(pid_t, ...);
 
 #endif /* PR_SCOREBOARD_H */
