@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2004-2007 The ProFTPD Project team
+ * Copyright (c) 2004-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,7 +23,7 @@
  */
 
 /* Table management
- * $Id: table.h,v 1.6 2007/10/04 01:46:58 castaglia Exp $
+ * $Id: table.h,v 1.8 2011/05/23 20:35:35 castaglia Exp $
  */
 
 #ifndef PR_TABLE_H
@@ -271,4 +271,7 @@ pr_table_t *pr_table_nalloc(pool *p, int flags, unsigned int nchains);
  */
 void *pr_table_pcalloc(pr_table_t *tab, size_t sz);
 
+/* Internal use only. */
+int table_handling_signal(int bool);
+
 #endif /* PR_TABLE_H */
