@@ -27,7 +27,7 @@
  * SUCH DAMAGE.
  */
 
-/* $Id: pr-syslog.c,v 1.22 2008/08/25 22:16:00 castaglia Exp $
+/* $Id: pr-syslog.c,v 1.24 2011/05/01 04:32:27 castaglia Exp $
  */
 
 #include "conf.h"
@@ -50,6 +50,60 @@ static int log_mask = 0xff;
 extern char *__progname;
 #endif /* HAVE___PROGNAME */
 
+#if defined(SOLARIS2_9) || defined(SOLARIS2_10)
+/* These tables are used for populating the stupid Solaris 9/10 syslog
+ * "header".
+ */
+
+struct {
+  int facility;
+  const char *name;
+
+} syslog_facility_names[] = {
+  { LOG_AUTHPRIV,	"auth" },
+#ifdef HAVE_LOG_FTP
+  { LOG_FTP,		"ftp" },
+#endif
+#ifdef HAVE_LOG_CRON
+  { LOG_CRON,		"cron" },
+#endif
+  { LOG_DAEMON,		"daemon" },
+  { LOG_KERN,		"kern" },
+  { LOG_LOCAL0,		"local0" },
+  { LOG_LOCAL1,		"local1" },
+  { LOG_LOCAL2,		"local2" },
+  { LOG_LOCAL3,		"local3" },
+  { LOG_LOCAL4,		"local4" },
+  { LOG_LOCAL5,		"local5" },
+  { LOG_LOCAL6,		"local6" },
+  { LOG_LOCAL7,		"local7" },
+  { LOG_LPR,		"lpr" },
+  { LOG_MAIL,		"mail" },
+  { LOG_NEWS,		"news" },
+  { LOG_USER,		"user" },
+  { LOG_UUCP,		"uucp" },
+  { 0,			NULL }
+};
+
+struct {
+  int level;
+  const char *name;
+
+} syslog_level_names[] = {
+  { PR_LOG_EMERG,	"emerg" },
+  { PR_LOG_ALERT,	"alert" },
+  { PR_LOG_CRIT,	"crit" },
+  { PR_LOG_ERR,		"error" },
+  { PR_LOG_ERR,		"error" },
+  { PR_LOG_WARNING,	"warn" },
+  { PR_LOG_NOTICE,	"notice" },
+  { PR_LOG_INFO,	"info" },
+  { PR_LOG_DEBUG,	"debug" },
+  { 0,			NULL }
+};
+
+#endif /* Solaris 9 or 10 */
+
 static void pr_vsyslog(int sockfd, int pri, register const char *fmt,
     va_list ap) {
   time_t now;
@@ -143,6 +197,50 @@ static void pr_vsyslog(int sockfd, int pri, register const char *fmt,
     buflen = strlen(logbuf);
   }
 
+#if defined(SOLARIS2_9) || defined(SOLARIS2_10)
+  /* Add in the (IMHO stupid and nonportable) syslog "header" that was added
+   * to the Solaris 9/10 libc syslog(3) function.  Some sites apparently
+   * think that trying to use this header to generate reports of logging
+   * is a Good Idea; I'll have the last laugh when those sites try to move
+   * to a different platform with different syslog logging.
+   *
+   * The header to be added looks like:
+   *
+   *  "[ID %lu %s.%s]"
+   *
+   * where the ID is generated using STRLOG_MAKE_MSGID(), a macro defined
+   * in <sys/strlog.h>, and the following two strings are the syslog
+   * facility and level, respectively.
+   */
+
+  if (buflen < sizeof(logbuf)) {
+    register unsigned int i;
+    uint32_t msgid;
+    const char *facility_name = "unknown", *level_name = "unknown";
+
+    STRLOG_MAKE_MSGID(fmt, msgid);
+
+    for (i = 0; syslog_facility_names[i].name; i++) {
+      if (syslog_facility_names[i].facility == log_facility) {
+        facility_name = syslog_facility_names[i].name;
+        break;
+      }
+    }
+
+    for (i = 0; syslog_level_names[i].name; i++) {
+      if (syslog_level_names[i].level == (pri & LOG_PRIMASK)) {
+        level_name = syslog_level_names[i].name;
+        break;
+      }
+    }
+
+    snprintf(&(logbuf[buflen]), sizeof(logbuf) - buflen, "[ID %lu %s.%s] ",
+      (unsigned long) msgid, facility_name, level_name);
+    logbuf[sizeof(logbuf)-1] = '\0';
+    buflen = strlen(logbuf);
+  }
+#endif /* Solaris 9 or 10 */
+
   /* Restore errno for %m format.  */
   errno = saved_errno;
 
@@ -197,7 +295,7 @@ static struct sockaddr syslog_addr;
 #endif
 
 int pr_openlog(const char *ident, int opts, int facility) {
-  int sockfd = -1;
+  int sockfd;
 
   if (ident != NULL)
     log_ident = ident;
@@ -208,13 +306,15 @@ int pr_openlog(const char *ident, int opts, int facility) {
     log_facility = facility;
 
 #ifndef HAVE_DEV_LOG_STREAMS
+  sockfd = -1;
   while (1) {
     socklen_t addrlen = 0;
 
     if (sockfd == -1) {
       syslog_addr.sa_family = AF_UNIX;
 
-# if defined(DARWIN8) || defined(DARWIN9)
+# if defined(DARWIN8) || defined(DARWIN9) || defined(DARWIN10) || \
+     defined(DARWIN11)
       /* On Mac OSX, the sockaddr.sa_data field is 14 bytes.  It is possible
        * that PR_PATH_LOG exceeds that field length (which would truncate
        * the field, and ensure that proftpd cannot log via syslog).
