@@ -12,8 +12,8 @@ Library General Public License for more details.
 
 You should have received a copy of the GNU Library General Public
 License along with this library; see the file COPYING.LIB.  If
-not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite
-330, Boston, MA  02111-1307, USA.  */
+not, write to the Free Software Foundation, Inc., 51 Franklin Street,
+Suite 500, Boston, MA 02110-1335, USA.  */
 
 /* Required to tell conf.h not to include the standard ProFTPD
  * header files
@@ -137,7 +137,8 @@ static char *fgetbufline(char **buf, int *buflen, FILE *fp) {
     {
       char *new_buf;
 
-      if ((new_buf = realloc(*buf, *buflen)) == NULL)
+      new_buf = realloc(*buf, *buflen);
+      if (new_buf == NULL)
         break;
 
       *buf = new_buf;
@@ -176,10 +177,18 @@ static struct group *supp_getgrent(const char *buf) {
 
   i = strlen(buf) + 1;
 
-  if (!grpbuf)
+  if (!grpbuf) {
     grpbuf = malloc(i);
-  else
-    grpbuf = realloc(grpbuf, i);
+
+  } else {
+    char *new_buf;
+
+    new_buf = realloc(grpbuf, i);
+    if (new_buf == NULL)
+      return NULL;
+
+    grpbuf = new_buf;
+  }
 
   if (!grpbuf)
     return NULL;
