@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2005 The ProFTPD Project team
+ * Copyright (c) 2001-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -25,7 +25,7 @@
  */
 
 /* Unix authentication module for ProFTPD
- * $Id: mod_auth_unix.c,v 1.25 2005/07/03 18:52:02 castaglia Exp $
+ * $Id: mod_auth_unix.c,v 1.50 2011/05/23 21:11:56 castaglia Exp $
  */
 
 #include "conf.h"
@@ -38,7 +38,9 @@
 #endif
 
 #ifdef PR_USE_SHADOW
-# include <shadow.h>
+# ifdef HAVE_SHADOW_H
+#   include <shadow.h>
+# endif
 #endif
 
 #ifdef HAVE_SYS_SECURITY_H
@@ -49,6 +51,10 @@
 # include <krb.h>
 #endif
 
+#ifdef HAVE_LOGIN_H
+# include <login.h>
+#endif
+
 #if defined(HAVE_HPSECURITY_H) || defined(HPUX10) || defined(HPUX11)
 # include <hpsecurity.h>
 # ifndef COMSEC
@@ -60,6 +66,10 @@
 # include <prot.h>
 #endif
 
+#ifdef HAVE_USERSEC_H
+# include <usersec.h>
+#endif
+
 #ifdef PR_USE_SIA
 # ifdef HAVE_SIA_H
 #  include <sia.h>
@@ -87,47 +97,20 @@ extern void cygwin_set_impersonation_token (const HANDLE);
 
 #include "privs.h"
 
-const char *pwdfname = "/etc/passwd";
-const char *grpfname = "/etc/group";
+static const char *pwdfname = "/etc/passwd";
+static const char *grpfname = "/etc/group";
 
 #ifdef HAVE__PW_STAYOPEN
 extern int _pw_stayopen;
 #endif
 
-#define HASH_TABLE_SIZE		100
-
-typedef union idauth {
-  uid_t uid;
-  gid_t gid;
-} idauth_t;
-
-typedef struct _idmap {
-  struct _idmap *next,*prev;
-
-  /* This is a union because different OSs may give different types to UIDs
-   * and GIDs.  This presents a far more portable way to deal with this
-   * reality.
-   */
-  idauth_t id;
-
-  char *name;			/* user or group name */
-} idmap_t;
-
 module auth_unix_module;
 
-static xaset_t *uid_table[HASH_TABLE_SIZE];
-static xaset_t *gid_table[HASH_TABLE_SIZE];
-
 static FILE *pwdf = NULL;
 static FILE *grpf = NULL;
 
 extern unsigned char persistent_passwd;
 
-static int persistent_passwdf = 0, persistent_groupf = 0;
-
-#define PERSISTENT_PASSWD	(persistent_passwd || persistent_passwdf)
-#define PERSISTENT_GROUP	(persistent_passwd || persistent_groupf)
-
 #undef PASSWD
 #define PASSWD		pwdfname
 #undef GROUP
@@ -142,14 +125,24 @@ static int persistent_passwdf = 0, persistent_groupf = 0;
 
 #endif /* PR_USE_SHADOW */
 
+/* mod_auth_unix option flags */
+#define AUTH_UNIX_OPT_AIX_NO_RLOGIN		0x0001
+#define AUTH_UNIX_OPT_NO_GETGROUPLIST		0x0002
+#define AUTH_UNIX_OPT_MAGIC_TOKEN_CHROOT	0x0004
+
+static unsigned long auth_unix_opts = 0UL;
+
 static void p_setpwent(void) {
   if (pwdf)
     rewind(pwdf);
 
-  else
-    if ((pwdf = fopen(PASSWD,"r")) == NULL)
+  else {
+    pwdf = fopen(PASSWD, "r");
+    if (pwdf == NULL) {
       pr_log_pri(PR_LOG_ERR, "Unable to open password file %s for reading: %s",
         PASSWD, strerror(errno));
+    }
+  }
 }
 
 static void p_endpwent(void) {
@@ -163,10 +156,13 @@ static RETSETGRENTTYPE p_setgrent(void) {
   if (grpf)
     rewind(grpf);
 
-  else
-    if ((grpf = fopen(GROUP,"r")) == NULL)
+  else {
+    grpf = fopen(GROUP, "r");
+    if (grpf == NULL) {
       pr_log_pri(PR_LOG_ERR, "Unable to open group file %s for reading: %s",
         GROUP, strerror(errno));
+    }
+  }
 
 #ifndef SETGRENT_VOID
   return 0;
@@ -208,9 +204,13 @@ static struct passwd *p_getpwnam(const char *name) {
   struct passwd *pw = NULL;
 
   p_setpwent();
-  while ((pw = p_getpwent()) != NULL)
-    if (!strcmp(name,pw->pw_name))
+  while ((pw = p_getpwent()) != NULL) {
+    pr_signals_handle();
+
+    if (strcmp(name, pw->pw_name) == 0) {
       break;
+    }
+  }
 
   return pw;
 }
@@ -219,9 +219,12 @@ static struct passwd *p_getpwuid(uid_t uid) {
   struct passwd *pw = NULL;
 
   p_setpwent();
-  while ((pw = p_getpwent()) != NULL)
+  while ((pw = p_getpwent()) != NULL) {
+    pr_signals_handle();
+
     if (pw->pw_uid == uid)
       break;
+  }
 
   return pw;
 }
@@ -230,9 +233,12 @@ static struct group *p_getgrnam(const char *name) {
   struct group *gr = NULL;
 
   p_setgrent();
-  while ((gr = p_getgrent()) != NULL)
-    if (!strcmp(name,gr->gr_name))
+  while ((gr = p_getgrent()) != NULL) {
+    pr_signals_handle();
+
+    if (strcmp(name, gr->gr_name) == 0)
       break;
+  }
 
   return gr;
 }
@@ -241,125 +247,78 @@ static struct group *p_getgrgid(gid_t gid) {
   struct group *gr = NULL;
 
   p_setgrent();
-  while ((gr = p_getgrent()) != NULL)
+  while ((gr = p_getgrent()) != NULL) {
+    pr_signals_handle();
+
     if (gr->gr_gid == gid)
       break;
+  }
 
   return gr;
 }
 
-inline static int _compare_uid(idmap_t *m1, idmap_t *m2) {
-  if (m1->id.uid < m2->id.uid)
-    return -1;
-
-  if (m1->id.uid > m2->id.uid)
-    return 1;
-
-  return 0;
-}
-
-inline static int _compare_gid(idmap_t *m1, idmap_t *m2) {
-  if (m1->id.gid < m2->id.gid)
-    return -1;
-
-  if (m1->id.gid > m2->id.gid)
-    return 1;
-
-  return 0;
-}
-
-inline static int _compare_id(xaset_t **table, idauth_t id, idauth_t idcomp) {
-  if (table == uid_table)
-    return id.uid == idcomp.uid;
-  else
-    return id.gid == idcomp.gid;
-}
-
-static idmap_t *_auth_lookup_id(xaset_t **id_table, idauth_t id) {
-  int hash = ((id_table == uid_table) ? id.uid : id.gid) % HASH_TABLE_SIZE;
-  idmap_t *m;
-
-  if (!id_table[hash])
-    id_table[hash] = xaset_create(session.pool,
-      (id_table == uid_table) ? (XASET_COMPARE)_compare_uid :
-      (XASET_COMPARE)_compare_gid);
-
-  for (m = (idmap_t *) id_table[hash]->xas_list; m; m = m->next) {
-    if (_compare_id(id_table, m->id, id))
-      break;
-  }
-
-  if (!m || !_compare_id(id_table, m->id, id)) {
-    /* Isn't in the table */
-    m = (idmap_t *) pcalloc(id_table[hash]->pool, sizeof(idmap_t));
-
-    if (id_table == uid_table)
-      m->id.uid = id.uid;
-    else
-      m->id.gid = id.gid;
-
-    xaset_insert_sort(id_table[hash], (xasetmember_t *) m, FALSE);
-  }
-
-  return m;
-}
-
 MODRET pw_setpwent(cmd_rec *cmd) {
-  if (PERSISTENT_PASSWD)
+  if (persistent_passwd)
     p_setpwent();
+
   else
     setpwent();
 
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 MODRET pw_endpwent(cmd_rec *cmd) {
-  if (PERSISTENT_PASSWD)
+  if (persistent_passwd)
     p_endpwent();
+
   else
     endpwent();
 
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 MODRET pw_setgrent(cmd_rec *cmd) {
-  if (PERSISTENT_GROUP)
+  if (persistent_passwd)
     p_setgrent();
+
   else
     setgrent();
 
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 MODRET pw_endgrent(cmd_rec *cmd) {
-  if (PERSISTENT_GROUP)
+  if (persistent_passwd)
     p_endgrent();
+
   else
     endgrent();
 
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 MODRET pw_getgrent(cmd_rec *cmd) {
   struct group *gr;
 
-  if (PERSISTENT_GROUP)
+  if (persistent_passwd)
     gr = p_getgrent();
+
   else
     gr = getgrent();
 
-  return gr ? mod_create_data(cmd, gr) : DECLINED(cmd);
+  return gr ? mod_create_data(cmd, gr) : PR_DECLINED(cmd);
 }
 
 MODRET pw_getpwent(cmd_rec *cmd) {
   struct passwd *pw;
 
-  if (PERSISTENT_PASSWD)
+  if (persistent_passwd)
     pw = p_getpwent();
+
   else
     pw = getpwent();
 
-  return pw ? mod_create_data(cmd, pw) : DECLINED(cmd);
+  return pw ? mod_create_data(cmd, pw) : PR_DECLINED(cmd);
 }
 
 MODRET pw_getpwuid(cmd_rec *cmd) {
@@ -367,12 +326,13 @@ MODRET pw_getpwuid(cmd_rec *cmd) {
   uid_t uid;
 
   uid = *((uid_t *) cmd->argv[0]);
-  if (PERSISTENT_PASSWD)
+  if (persistent_passwd)
     pw = p_getpwuid(uid);
+
   else
     pw = getpwuid(uid);
 
-  return pw ? mod_create_data(cmd, pw) : DECLINED(cmd);
+  return pw ? mod_create_data(cmd, pw) : PR_DECLINED(cmd);
 }
 
 MODRET pw_getpwnam(cmd_rec *cmd) {
@@ -380,12 +340,94 @@ MODRET pw_getpwnam(cmd_rec *cmd) {
   const char *name;
 
   name = cmd->argv[0];
-  if (PERSISTENT_PASSWD)
+  if (persistent_passwd) {
     pw = p_getpwnam(name);
-  else
-    pw = getpwnam(name);
 
-  return pw ? mod_create_data(cmd, pw) : DECLINED(cmd);
+  } else {
+    pw = getpwnam(name);
+  }
+
+  if (auth_unix_opts & AUTH_UNIX_OPT_MAGIC_TOKEN_CHROOT) {
+    char *home_dir, *ptr;
+
+    /* Here is where we do the "magic token" chroot monstrosity inflicted
+     * on the world by wu-ftpd.
+     *
+     * If the magic token '/./' appears in the user's home directory, the
+     * directory portion before the token is the directory to use for
+     * the chroot; the directory portion after the token is the directory
+     * to use for the initial chdir.
+     */
+
+    home_dir = pstrdup(cmd->tmp_pool, pw->pw_dir);
+
+    /* We iterate through the home directory string since it is possible
+     * for the '.' character to appear without it being part of the magic
+     * token.
+     */
+    ptr = strchr(home_dir, '.');
+    while (ptr != NULL) {
+      pr_signals_handle();
+
+      /* If we're at the start of the home directory string, stop looking:
+       * this home directory is not really valid anyway.
+       */
+      if (ptr == home_dir) {
+        break;
+      }
+
+      /* Back up one character. */
+      ptr--;
+
+      /* If we're at the start of the home directory now, stop looking:
+       * this home directory cannot contain a valid magic token.  I.e.
+       *
+       * /./home/foo
+       *
+       * cannot be valid, as there is no directory portion before the
+       * token.
+       */
+      if (ptr == home_dir) {
+        break;
+      }
+
+      if (strncmp(ptr, "/./", 3) == 0) {
+        char *default_chdir;
+        config_rec *c;
+
+        *ptr = '\0';
+        default_chdir = pstrdup(cmd->tmp_pool, ptr + 2);
+
+        /* In order to make sure that this user is chrooted to this
+         * directory, we remove all DefaultRoot directives and add a new
+         * one.  Same for the DefaultChdir directive.
+         */
+
+        (void) remove_config(main_server->conf, "DefaultRoot", FALSE);
+        c = add_config_param_set(&main_server->conf, "DefaultRoot", 1, NULL);
+        c->argv[0] = pstrdup(c->pool, home_dir);
+
+        (void) remove_config(main_server->conf, "DefaultChdir", FALSE);
+        c = add_config_param_set(&main_server->conf, "DefaultChdir", 1, NULL);
+        c->argv[0] = pstrdup(c->pool, default_chdir);
+
+        pr_log_debug(DEBUG9, "AuthUnixOption magicTokenChroot: "
+          "found magic token in '%s', using 'DefaultRoot %s' and "
+          "'DefaultChdir %s'", pw->pw_dir, home_dir, default_chdir);
+
+        /* We need to use a long-lived memory pool for overwriting the
+         * normal home directory.
+         */
+        pw->pw_dir = pstrdup(session.pool, home_dir);
+
+        break;
+      }
+
+      ptr = strchr(ptr + 2, '.');
+    }
+  }
+
+  return pw ? mod_create_data(cmd, pw) : PR_DECLINED(cmd);
 }
 
 MODRET pw_getgrnam(cmd_rec *cmd) {
@@ -393,12 +435,13 @@ MODRET pw_getgrnam(cmd_rec *cmd) {
   const char *name;
 
   name = cmd->argv[0];
-  if (PERSISTENT_GROUP)
+  if (persistent_passwd)
     gr = p_getgrnam(name);
+
   else
     gr = getgrnam(name);
 
-  return gr ? mod_create_data(cmd, gr) : DECLINED(cmd);
+  return gr ? mod_create_data(cmd, gr) : PR_DECLINED(cmd);
 }
 
 MODRET pw_getgrgid(cmd_rec *cmd) {
@@ -406,12 +449,13 @@ MODRET pw_getgrgid(cmd_rec *cmd) {
   gid_t gid;
 
   gid = *((gid_t *) cmd->argv[0]);
-  if (PERSISTENT_GROUP)
+  if (persistent_passwd)
     gr = p_getgrgid(gid);
+
   else
     gr = getgrgid(gid);
 
-  return gr ? mod_create_data(cmd, gr) : DECLINED(cmd);
+  return gr ? mod_create_data(cmd, gr) : PR_DECLINED(cmd);
 }
 
 #ifdef PR_USE_SHADOW
@@ -426,67 +470,84 @@ static char *_get_pw_info(pool *p, const char *u, time_t *lstchg, time_t *min,
 #endif /* HAVE_SETSPENT */
 
   sp = getspnam(u);
-
-  if (sp) {
+  if (sp != NULL) {
     cpw = pstrdup(p, sp->sp_pwdp);
 
-    if (lstchg)
+    if (lstchg != NULL) {
       *lstchg = SP_CVT_DAYS(sp->sp_lstchg);
+    }
 
-    if (min)
+    if (min != NULL) {
       *min = SP_CVT_DAYS(sp->sp_min);
+    }
 
-    if (max)
+    if (max != NULL) {
       *max = SP_CVT_DAYS(sp->sp_max);
+    }
 
 #ifdef HAVE_SPWD_SP_WARN
-    if (warn)
+    if (warn != NULL) {
       *warn = SP_CVT_DAYS(sp->sp_warn);
+    }
 #endif /* HAVE_SPWD_SP_WARN */
 
 #ifdef HAVE_SPWD_SP_INACT
-    if (inact)
+    if (inact != NULL) {
       *inact = SP_CVT_DAYS(sp->sp_inact);
+    }
 #endif /* HAVE_SPWD_SP_INACT */
 
 #ifdef HAVE_SPWD_SP_EXPIRE
-    if (expire)
+    if (expire != NULL) {
       *expire = SP_CVT_DAYS(sp->sp_expire);
+    }
 #endif /* HAVE_SPWD_SP_EXPIRE */
   }
+
 #ifdef PR_USE_AUTO_SHADOW
-  else {
+  if (sp == NULL) {
     struct passwd *pw;
 
     endspent();
     PRIVS_RELINQUISH
 
-    if ((pw = getpwnam(u)) != NULL) {
+    pw = getpwnam(u);
+    if (pw != NULL) {
       cpw = pstrdup(p, pw->pw_passwd);
 
-      if (lstchg)
+      if (lstchg != NULL) {
         *lstchg = (time_t) -1;
+      }
 
-      if (min)
+      if (min != NULL) {
         *min = (time_t) -1;
+      }
 
-      if (max)
+      if (max != NULL) {
         *max = (time_t) -1;
+      }
 
-      if (warn)
+      if (warn != NULL) {
         *warn = (time_t) -1;
+      }
 
-      if (inact)
+      if (inact != NULL) {
         *inact = (time_t) -1;
+      }
 
-      if (expire)
+      if (expire != NULL) {
         *expire = (time_t) -1;
+      }
     }
+
+  } else {
+    PRIVS_RELINQUISH
   }
 #else
   endspent();
   PRIVS_RELINQUISH
 #endif /* PR_USE_AUTO_SHADOW */
+
   return cpw;
 }
 
@@ -525,7 +586,6 @@ static char *_get_pw_info(pool *p, const char *u, time_t *lstchg, time_t *min,
 #endif /* BSDI3 || BSDI4 */
 
   pw = getpwnam(u);
-
   if (pw) {
     cpw = pstrdup(p, pw->pw_passwd);
 
@@ -597,17 +657,6 @@ static char *_get_pw_info(pool *p, const char *u, time_t *lstchg, time_t *min,
 
 #endif /* PR_USE_SHADOW */
 
-static char *_get_ppw_info(pool *p, const char *u) {
-  struct passwd *pw;
-  char *cpw = NULL;
-
-  pw = p_getpwnam(u);
-  if (pw)
-    cpw = pstrdup(p, pw->pw_passwd);
-
-  return cpw;
-}
-
 /* High-level auth handlers
  */
 
@@ -624,40 +673,106 @@ MODRET pw_auth(cmd_rec *cmd) {
   name = cmd->argv[0];
   time(&now);
 
-  if (persistent_passwdf)
-    cpw = _get_ppw_info(cmd->tmp_pool, name);
-  else
-    cpw = _get_pw_info(cmd->tmp_pool, name, &lstchg, NULL, &max, NULL, &inact,
-      &disable);
+  cpw = _get_pw_info(cmd->tmp_pool, name, &lstchg, NULL, &max, NULL, &inact,
+    &disable);
 
   if (!cpw)
-    return DECLINED(cmd);
+    return PR_DECLINED(cmd);
 
   if (pr_auth_check(cmd->tmp_pool, cpw, cmd->argv[0], cmd->argv[1]))
-    return ERROR_INT(cmd, PR_AUTH_BADPWD);
+    return PR_ERROR_INT(cmd, PR_AUTH_BADPWD);
 
   if (lstchg > (time_t) 0 &&
       max > (time_t) 0 &&
       inact > (time_t)0)
     if (now > lstchg + max + inact)
-      return ERROR_INT(cmd, PR_AUTH_AGEPWD);
+      return PR_ERROR_INT(cmd, PR_AUTH_AGEPWD);
 
   if (disable > (time_t) 0 &&
       now > disable)
-    return ERROR_INT(cmd, PR_AUTH_DISABLEDPWD);
+    return PR_ERROR_INT(cmd, PR_AUTH_DISABLEDPWD);
 
   session.auth_mech = "mod_auth_unix.c";
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
-/* cmd->argv[0] = hashed password,
- * cmd->argv[1] = user,
+MODRET pw_authz(cmd_rec *cmd) {
+
+#ifdef HAVE_LOGINRESTRICTIONS
+  int code = 0, mode = S_RLOGIN;
+  char *reason = NULL;
+#endif
+
+  /* XXX Any other implementations here? */
+
+#ifdef HAVE_LOGINRESTRICTIONS
+
+  if (auth_unix_opts & AUTH_UNIX_OPT_AIX_NO_RLOGIN) {
+    mode = 0;
+  }
+
+  /* Check for account login restrictions and such using AIX-specific
+   * functions.
+   */
+  PRIVS_ROOT
+  if (loginrestrictions(cmd->argv[0], mode, NULL, &reason) != 0) {
+    PRIVS_RELINQUISH
+
+    if (reason &&
+        *reason) {
+      pr_log_auth(LOG_WARNING, "login restricted for user '%s': %.100s",
+        cmd->argv[0], reason);
+    }
+
+    pr_log_debug(DEBUG2, "AIX loginrestrictions() failed for user '%s': %s",
+      cmd->argv[0], strerror(errno));
+
+    return PR_ERROR_INT(cmd, PR_AUTH_DISABLEDPWD);
+  }
+
+  code = passwdexpired(cmd->argv[0], &reason);
+  PRIVS_RELINQUISH
+
+  switch (code) {
+    case 0:
+      /* Password not expired for user */
+      break;
+
+    case 1:
+      /* Password expired and needs to be changed */
+      pr_log_auth(LOG_WARNING, "password expired for user '%s': %.100s",
+        cmd->argv[0], reason);
+      return PR_ERROR_INT(cmd, PR_AUTH_AGEPWD);
+
+    case 2:
+      /* Password expired, requires sysadmin to change it */
+      pr_log_auth(LOG_WARNING,
+        "password expired for user '%s', requires sysadmin intervention: "
+        "%.100s", cmd->argv[0], reason);
+      return PR_ERROR_INT(cmd, PR_AUTH_AGEPWD);
+
+    default:
+      /* Other error */
+      pr_log_auth(LOG_WARNING, "AIX passwdexpired() failed for user '%s': "
+        "%.100s", cmd->argv[0], reason);
+      return PR_ERROR_INT(cmd, PR_AUTH_DISABLEDPWD);
+  }
+#endif /* !HAVE_LOGINRESTRICTIONS */
+
+  return PR_HANDLED(cmd);
+}
+
+/* cmd->argv[0] = hashed password
+ * cmd->argv[1] = user
  * cmd->argv[2] = cleartext
  */
 
 MODRET pw_check(cmd_rec *cmd) {
   const char *cpw = cmd->argv[0];
   const char *pw = cmd->argv[2];
+  modret_t *mr = NULL;
+  cmd_rec *cmd2 = NULL;
+  char *crypted_text = NULL;
 
 #ifdef PR_USE_SIA
   SIAENTITY *ent = NULL;
@@ -669,8 +784,9 @@ MODRET pw_check(cmd_rec *cmd) {
 
 #ifdef COMSEC
   if (iscomsec()) {
-    if (strcmp(bigcrypt((char *) pw, (char *) cpw), cpw) != 0)
-      return ERROR(cmd);
+    if (strcmp(bigcrypt((char *) pw, (char *) cpw), cpw) != 0) {
+      return PR_DECLINED(cmd);
+    }
 
   } else {
 #endif /* COMSEC */
@@ -699,7 +815,7 @@ MODRET pw_check(cmd_rec *cmd) {
       PRIVS_RELINQUISH
       pr_log_auth(PR_LOG_NOTICE, "sia_ses_authent() returned %d for user '%s'",
         res, user);
-      return ERROR(cmd);
+      return PR_ERROR(cmd);
     }
 
     res = sia_ses_estab(NULL, ent);
@@ -707,20 +823,21 @@ MODRET pw_check(cmd_rec *cmd) {
       PRIVS_RELINQUISH
       pr_log_auth(PR_LOG_NOTICE, "sia_ses_estab() returned %d for user '%s'",
         res, user);
-      return ERROR(cmd);
+      return PR_ERROR(cmd);
     }
 
     res = sia_ses_release(&ent);
     if (res != SIASUCCESS) {
       PRIVS_RELINQUISH
       pr_log_auth(PR_LOG_NOTICE, "sia_ses_release() returned %d", res);
-      return ERROR(cmd);
+      return PR_ERROR(cmd);
     }
   }
   PRIVS_RELINQUISH
 
-  if (res != SIASUCCESS)
-    return ERROR(cmd);
+  if (res != SIASUCCESS) {
+    return PR_DECLINED(cmd);
+  }
 
 #else /* !PR_USE_SIA */
 
@@ -730,46 +847,63 @@ MODRET pw_check(cmd_rec *cmd) {
    * http://cygwin.com/cygwin-ug-net/ntsec.html#NTSEC-SETUID
    */
   if (GetVersion() < 0x80000000) {
-    cmd_rec *tmp_cmd = NULL;
-    modret_t *mr = NULL;
     struct passwd *pwent = NULL;
     HANDLE token;
 
     /* A struct passwd * is needed.  To look one up via pw_getpwnam(), though,
      * we'll need a cmd_rec.
      */
-    tmp_cmd = pr_cmd_alloc(cmd->tmp_pool, 1, cmd->argv[1]);
+    cmd2 = pr_cmd_alloc(cmd->tmp_pool, 1, cmd->argv[1]);
 
     /* pw_getpwnam() returns a MODRET, so we need to handle that.  Yes, this
      * might have been easier if we'd used pr_auth_getpwnam(), but that would
      * dispatch through other auth modules, which is _not_ what we want.
      */
-    mr = pw_getpwnam(tmp_cmd);
+    mr = pw_getpwnam(cmd2);
 
     /* Note: we don't handle the case where pw_getpwnam() returns anything
      * other than HANDLED at the moment.
      */
 
-    if (MODRET_ISHANDLED(mr) && MODRET_HASDATA(mr)) {
+    if (MODRET_ISHANDLED(mr) &&
+        MODRET_HASDATA(mr)) {
       pwent = mr->data;
 
-      if ((token = cygwin_logon_user((const struct passwd *) pwent,
-          pw)) == INVALID_HANDLE_VALUE) {
+      token = cygwin_logon_user((const struct passwd *) pwent, pw);
+      if (token == INVALID_HANDLE_VALUE) {
         pr_log_pri(PR_LOG_NOTICE, "error authenticating Cygwin user: %s",
           strerror(errno));
-        return ERROR(cmd);
+        return PR_DECLINED(cmd);
       }
 
       cygwin_set_impersonation_token(token);
 
-    } else
-      return DECLINED(cmd);
+    } else {
+      return PR_DECLINED(cmd);
+    }
 
   } else
 # endif /* CYGWIN */
 
-  if (strcmp(crypt(pw, cpw), cpw) != 0)
-    return ERROR(cmd);
+  /* Call pw_authz here, to make sure the user is authorized to login. */
+
+  if (cmd2 == NULL)
+    cmd2 = pr_cmd_alloc(cmd->tmp_pool, 1, cmd->argv[1]);
+
+  mr = pw_authz(cmd2);
+  if (MODRET_ISDECLINED(mr)) {
+    return PR_DECLINED(cmd);
+  }
+
+  crypted_text = (char *) crypt(pw, cpw);
+  if (crypted_text == NULL) {
+    pr_log_pri(PR_LOG_NOTICE, "error error crypt(3): %s", strerror(errno));
+    return PR_DECLINED(cmd);
+  }
+
+  if (strcmp(crypted_text, cpw) != 0) {
+    return PR_DECLINED(cmd);
+  }
 #endif /* PR_USE_SIA */
 
 #ifdef COMSEC
@@ -777,62 +911,43 @@ MODRET pw_check(cmd_rec *cmd) {
 #endif /* COMSEC */
 
   session.auth_mech = "mod_auth_unix.c";
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 MODRET pw_uid2name(cmd_rec *cmd) {
-  idmap_t *m;
-  idauth_t id;
+  uid_t uid;
   struct passwd *pw;
 
-  id.uid = *((uid_t *) cmd->argv[0]);
-  m = _auth_lookup_id(uid_table, id);
+  uid = *((uid_t *) cmd->argv[0]);
 
-  if (!m->name) {
-    /* Wasn't cached, so perform a lookup */
+  if (persistent_passwd)
+    pw = p_getpwuid(uid);
 
-    if (PERSISTENT_PASSWD)
-      pw = p_getpwuid(id.uid);
-    else
-      pw = getpwuid(id.uid);
+  else
+    pw = getpwuid(uid);
 
-    if (pw) {
-      m->name = pstrdup(session.pool ? session.pool : permanent_pool,
-        pw->pw_name);
-      return mod_create_data(cmd, m->name);
-    }
+  if (pw)
+    return mod_create_data(cmd, pw->pw_name);
 
-    return DECLINED(cmd);
-  }
-
-  return mod_create_data(cmd, m->name);
+  return PR_DECLINED(cmd);
 }
 
 MODRET pw_gid2name(cmd_rec *cmd) {
-  idmap_t *m;
-  idauth_t id;
+  gid_t gid;
   struct group *gr;
 
-  id.gid = *((gid_t *) cmd->argv[0]);
+  gid = *((gid_t *) cmd->argv[0]);
 
-  m = _auth_lookup_id(gid_table, id);
+  if (persistent_passwd)
+    gr = p_getgrgid(gid);
 
-  if (!m->name) {
-    if (PERSISTENT_GROUP)
-      gr = p_getgrgid(id.gid);
-    else
-      gr = getgrgid(id.gid);
+  else
+    gr = getgrgid(gid);
 
-    if (gr) {
-      m->name = pstrdup(session.pool ? session.pool : permanent_pool,
-        gr->gr_name);
-      return mod_create_data(cmd, m->name);
-    }
+  if (gr) 
+    return mod_create_data(cmd, gr->gr_name);
 
-    return DECLINED(cmd);
-  }
-
-  return mod_create_data(cmd, m->name);
+  return PR_DECLINED(cmd);
 }
 
 MODRET pw_name2uid(cmd_rec *cmd) {
@@ -841,12 +956,13 @@ MODRET pw_name2uid(cmd_rec *cmd) {
 
   name = cmd->argv[0];
 
-  if (PERSISTENT_PASSWD)
+  if (persistent_passwd)
     pw = p_getpwnam(name);
+
   else
     pw = getpwnam(name);
 
-  return pw ? mod_create_data(cmd, (void *) &pw->pw_uid) : DECLINED(cmd);
+  return pw ? mod_create_data(cmd, (void *) &pw->pw_uid) : PR_DECLINED(cmd);
 }
 
 MODRET pw_name2gid(cmd_rec *cmd) {
@@ -855,12 +971,13 @@ MODRET pw_name2gid(cmd_rec *cmd) {
 
   name = cmd->argv[0];
 
-  if (PERSISTENT_GROUP)
+  if (persistent_passwd)
     gr = p_getgrnam(name);
+
   else
     gr = getgrnam(name);
 
-  return gr ? mod_create_data(cmd, (void *) &gr->gr_gid) : DECLINED(cmd);
+  return gr ? mod_create_data(cmd, (void *) &gr->gr_gid) : PR_DECLINED(cmd);
 }
 
 /* cmd->argv[0] = name
@@ -872,16 +989,17 @@ MODRET pw_getgroups(cmd_rec *cmd) {
   struct passwd *pw = NULL;
   struct group *gr = NULL;
   array_header *gids = NULL, *groups = NULL;
-  char **gr_member = NULL, *name = NULL;
+  char *name = NULL;
+  int use_getgrouplist = FALSE;
 
-  /* function pointers for which lookup functions to use */
+  /* Function pointers for which lookup functions to use */
   struct passwd *(*my_getpwnam)(const char *) = NULL;
   struct group *(*my_getgrgid)(gid_t) = NULL;
   struct group *(*my_getgrent)(void) = NULL;
   RETSETGRENTTYPE (*my_setgrent)(void) = NULL;
 
-  /* play function pointer games */
-  if (PERSISTENT_PASSWD) {
+  /* Play function pointer games */
+  if (persistent_passwd) {
     my_getpwnam = p_getpwnam;
     my_getgrgid = p_getgrgid;
     my_getgrent = p_getgrent;
@@ -894,6 +1012,29 @@ MODRET pw_getgroups(cmd_rec *cmd) {
     my_setgrent = setgrent;
   }
 
+#ifdef HAVE_GETGROUPLIST
+  /* Determine whether to use getgrouplist(3), if available.  Older glibc
+   * versions (i.e. 2.2.4 and older) had buggy getgrouplist() implementations
+   * which allowed for buffer overflows (see CVS-2003-0689); do not use
+   * getgrouplist() on such glibc versions.
+   */
+  use_getgrouplist = TRUE;
+
+# if defined(__GLIBC__) && \
+     defined(__GLIBC_MINOR__) && \
+     __GLIBC__ <= 2 && \
+     __GLIBC_MINOR__ < 3
+  use_getgrouplist = FALSE;
+# endif
+#endif /* !HAVE_GETGROUPLIST */
+
+  /* Use of getgrouplist(3) might have been disabled via the "noGetgrouplist"
+   * AuthUnixOption as well.
+   */
+  if (auth_unix_opts & AUTH_UNIX_OPT_NO_GETGROUPLIST) {
+    use_getgrouplist = FALSE;
+  }
+
   name = (char *) cmd->argv[0];
 
   /* Check for NULL values. */
@@ -905,7 +1046,7 @@ MODRET pw_getgroups(cmd_rec *cmd) {
 
   /* Retrieve the necessary info. */
   if (!name || !(pw = my_getpwnam(name)))
-    return DECLINED(cmd);
+    return PR_DECLINED(cmd);
 
   /* Populate the first group ID and name. */
   if (gids)
@@ -916,35 +1057,174 @@ MODRET pw_getgroups(cmd_rec *cmd) {
 
   my_setgrent();
 
-  /* This is where things get slow, expensive, and ugly.  Loop through
-   * everything, checking to make sure we haven't already added it.
-   */
-  while ((gr = my_getgrent()) != NULL && gr->gr_mem) {
+  if (use_getgrouplist) {
+#ifdef HAVE_GETGROUPLIST
+    gid_t group_ids[NGROUPS_MAX];
+    int ngroups = NGROUPS_MAX;
+    register unsigned int i;
 
-    /* Loop through each member name listed */
-    for (gr_member = gr->gr_mem; *gr_member; gr_member++) {
+    pr_trace_msg("auth", 4,
+      "using getgrouplist(3) to look up group membership");
 
-      /* If it matches the given username... */
-      if (!strcmp(*gr_member, pw->pw_name)) {
+    memset(group_ids, 0, sizeof(group_ids));
+    if (getgrouplist(pw->pw_name, pw->pw_gid, group_ids, &ngroups) < 0) {
+      pr_log_pri(PR_LOG_ERR, "getgrouplist error: %s", strerror(errno));
+      return PR_DECLINED(cmd);
+    }
 
-        /* ...add the GID and name */
-        if (gids)
+    for (i = 0; i < ngroups; i++) {
+      gr = my_getgrgid(group_ids[i]);
+      if (gr) {
+        if (gids && pw->pw_gid != gr->gr_gid)
           *((gid_t *) push_array(gids)) = gr->gr_gid;
 
-        if (groups && pw->pw_gid != gr->gr_gid)
+        if (groups && pw->pw_gid != gr->gr_gid) {
           *((char **) push_array(groups)) = pstrdup(session.pool,
             gr->gr_name);
+        }
       }
     }
+#endif /* !HAVE_GETGROUPLIST */
+
+  } else {
+#ifdef HAVE_GETGRSET
+    gid_t group_ids[NGROUPS_MAX];
+    unsigned int ngroups = 0;
+    register unsigned int i;
+    char *grgid, *grouplist, *ptr;
+
+    pr_trace_msg("auth", 4,
+      "using getgrset(3) to look up group membership");
+
+    grouplist = getgrset(pw->pw_name);
+    if (grouplist == NULL) {
+      pr_log_pri(PR_LOG_ERR, "getgrset error: %s", strerror(errno));
+      return PR_DECLINED(cmd);
+    }
+
+    ptr = grouplist;
+    memset(group_ids, 0, sizeof(group_ids));
+
+    /* The getgrset(3) function returns a string which is a comma-delimited
+     * list of group IDs.
+     */
+    grgid = strsep(&grouplist, ",");
+    while (grgid) {
+      long gid;
+
+      pr_signals_handle();
+
+      if (ngroups >= sizeof(group_ids)) {
+        /* Reached capacity of the group_ids array. */
+        break;
+      }
+
+      /* XXX Should we use strtoul(3) or even strtoull(3) here? */
+      gid = strtol(grgid, NULL, 10);
+
+      /* Skip the primary group. */
+      if ((gid_t) gid == pw->pw_gid) {
+        grgid = strsep(&grouplist, ",");
+        continue;
+      }
+
+      group_ids[ngroups] = (gid_t) gid;
+      ngroups++;
+
+      grgid = strsep(&grouplist, ",");
+    }
+
+    for (i = 0; i < ngroups; i++) {
+      gr = my_getgrgid(group_ids[i]);
+      if (gr) {
+        if (gids && pw->pw_gid != gr->gr_gid)
+          *((gid_t *) push_array(gids)) = gr->gr_gid;
+
+        if (groups && pw->pw_gid != gr->gr_gid) {
+          *((char **) push_array(groups)) = pstrdup(session.pool,
+            gr->gr_name);
+        }
+      }
+    }
+
+    free(ptr);
+#else
+    char **gr_member = NULL;
+    size_t pw_namelen;
+
+    /* This is where things get slow, expensive, and ugly.  Loop through
+     * everything, checking to make sure we haven't already added it.
+     */
+    pw_namelen = strlen(pw->pw_name);
+    while ((gr = my_getgrent()) != NULL && gr->gr_mem) {
+      pr_signals_handle();
+
+      /* Loop through each member name listed */
+      for (gr_member = gr->gr_mem; *gr_member; gr_member++) {
+
+        /* If it matches the given username... */
+        if (strncmp(*gr_member, pw->pw_name, pw_namelen + 1) == 0) {
+
+          /* ...add the GID and name */
+          if (gids)
+            *((gid_t *) push_array(gids)) = gr->gr_gid;
+
+          if (groups && pw->pw_gid != gr->gr_gid) {
+            *((char **) push_array(groups)) = pstrdup(session.pool,
+              gr->gr_name);
+          }
+        }
+      }
+    }
+#endif /* !HAVE_GETGRSET */
+  }
+
+  if (gids &&
+      gids->nelts > 0) {
+    return mod_create_data(cmd, (void *) &gids->nelts);
+
+  } else if (groups &&
+             groups->nelts > 0) {
+    return mod_create_data(cmd, (void *) &groups->nelts);
+  }
+
+  return PR_DECLINED(cmd);
+}
+
+/* usage: AuthUnixOptions opt1 ... */
+MODRET set_authunixoptions(cmd_rec *cmd) {
+  config_rec *c;
+  register unsigned int i;
+  unsigned long opts = 0UL;
+
+  if (cmd->argc == 1) {
+    CONF_ERROR(cmd, "wrong number of parameters");
+  }
+
+  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
+
+  c = add_config_param(cmd->argv[0], 1, NULL);
+
+  for (i = 1; i < cmd->argc; i++) {
+    if (strcmp(cmd->argv[i], "aixNoRLogin") == 0) {
+      opts |= AUTH_UNIX_OPT_AIX_NO_RLOGIN;
+
+    } else if (strcmp(cmd->argv[i], "noGetgrouplist") == 0) {
+      opts |= AUTH_UNIX_OPT_NO_GETGROUPLIST;
+
+    } else if (strcmp(cmd->argv[i], "magicTokenChroot") == 0) {
+      opts |= AUTH_UNIX_OPT_MAGIC_TOKEN_CHROOT;
+
+    } else {
+      CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": unknown AuthUnixOption '",
+        cmd->argv[i], "'", NULL));
+    }
   }
 
-  if (gids && gids->nelts > 0)
-    return mod_create_data(cmd, (void *) &gids->nelts);
+  c->argv[0] = pcalloc(c->pool, sizeof(unsigned long));
+  *((unsigned long *) c->argv[0]) = opts;
 
-  else if (groups && groups->nelts > 0)
-    return mod_create_data(cmd, (void *) &groups->nelts);
-
-  return DECLINED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 MODRET set_persistentpasswd(cmd_rec *cmd) {
@@ -958,7 +1238,7 @@ MODRET set_persistentpasswd(cmd_rec *cmd) {
 
   persistent_passwd = bool;
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* Events handlers
@@ -975,8 +1255,6 @@ static void auth_unix_exit_ev(const void *event_data, void *user_data) {
  */
 
 static int auth_unix_init(void) {
-  memset(uid_table, 0 ,sizeof(uid_table));
-  memset(gid_table, 0, sizeof(gid_table));
 
 #ifdef HAVE__PW_STAYOPEN
   _pw_stayopen = 1;
@@ -986,25 +1264,15 @@ static int auth_unix_init(void) {
 }
 
 static int auth_unix_sess_init(void) {
-  const char *file = NULL;
-
-  if ((file = get_param_ptr(main_server->conf, "AuthUserFile", FALSE))) {
-    endpwent();
-    persistent_passwdf = 1;		/* Force persistent mode */
-    pwdfname = file;
-    p_endpwent();
-    p_setpwent();
-  }
-
-  if ((file = get_param_ptr(main_server->conf, "AuthGroupFile", FALSE))) {
-    endgrent();
-    persistent_groupf = 1;
-    grpfname = file;
-    p_endgrent();
-    p_setgrent();
-  }
+  config_rec *c;
 
   pr_event_register(&auth_unix_module, "core.exit", auth_unix_exit_ev, NULL);
+
+  c = find_config(main_server->conf, CONF_PARAM, "AuthUnixOptions", FALSE);
+  if (c) {
+    auth_unix_opts = *((unsigned long *) c->argv[0]);
+  }
+ 
   return 0;
 }
 
@@ -1012,6 +1280,7 @@ static int auth_unix_sess_init(void) {
  */
 
 static conftable auth_unix_conftab[] = {
+  { "AuthUnixOptions",		set_authunixoptions,		NULL },
   { "PersistentPasswd",		set_persistentpasswd,		NULL },
   { NULL,			NULL,				NULL }
 };
@@ -1028,6 +1297,7 @@ static authtable auth_unix_authtab[] = {
   { 0,  "getgrnam",     pw_getgrnam },
   { 0,  "getgrgid",     pw_getgrgid },
   { 0,  "auth",         pw_auth	},
+  { 0,  "authorize",	pw_authz },
   { 0,  "check",	pw_check },
   { 0,  "uid2name",	pw_uid2name },
   { 0,  "gid2name",	pw_gid2name },
