@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2009 The ProFTPD Project team
+ * Copyright (c) 2001-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -25,17 +25,17 @@
  */
 
 /* Flexible logging module for proftpd
- * $Id: mod_log.c,v 1.88 2009/02/03 17:42:05 castaglia Exp $
+ * $Id: mod_log.c,v 1.124 2011/09/26 15:00:12 castaglia Exp $
  */
 
 #include "conf.h"
 #include "privs.h"
 
-extern pr_response_t *resp_list, *resp_err_list;
-
 module log_module;
 
-#define EXTENDED_LOG_BUFFER_SIZE		1025
+/* Max path length plus 128 bytes for additional info. */
+#define EXTENDED_LOG_BUFFER_SIZE		(PR_TUNABLE_PATH_MAX + 128)
+
 #define EXTENDED_LOG_MODE			0644
 
 typedef struct logformat_struc	logformat_t;
@@ -96,7 +96,16 @@ struct logfile_struc {
 #define META_DIR_PATH		24
 #define META_CMD_PARAMS		25
 #define META_RESPONSE_STR	26
-#define META_VERSION		27
+#define META_PROTOCOL		27
+#define META_VERSION		28
+#define META_RENAME_FROM	29
+#define META_FILE_MODIFIED	30
+#define META_UID		31
+#define META_GID		32
+#define META_RAW_BYTES_IN	33
+#define META_RAW_BYTES_OUT	34
+#define META_EOS_REASON		35
+#define META_VHOST_IP		36
 
 /* For tracking the size of deleted files. */
 static off_t log_dele_filesz = 0;
@@ -114,14 +123,18 @@ static xaset_t			*log_set = NULL;
    %c			- Class
    %D			- full directory path
    %d			- directory (for client)
+   %E			- End-of-session reason
    %{FOOBAR}e		- Contents of environment variable FOOBAR
    %F			- Transfer path (filename for client)
    %f			- Filename
+   %H                   - Local IP address of server handling session
    %h			- Remote client DNS name
+   %I                   - Total number of "raw" bytes read in from network
    %J                   - Request (command) arguments (file.txt, etc)
-   %L                   - Local server IP address
+   %L                   - Local IP address contacted by client
    %l			- Remote logname (from identd)
    %m			- Request (command) method (RETR, etc)
+   %O                   - Total number of "raw" bytes written out to network
    %P			- Process ID of child serving request
    %p			- Port of server serving request
    %r			- Full request (command)
@@ -134,6 +147,12 @@ static xaset_t			*log_set = NULL;
    %u			- Local user
    %V                   - DNS name of server serving request
    %v			- ServerName of server serving request
+   %w                   - RNFR path ("whence" a rename comes, i.e. the source)
+   %{file-modified}     - Indicates whether a file is being modified
+                          (i.e. already exists) or not.
+   %{protocol}          - Current protocol (e.g. "ftp", "sftp", etc)
+   %{uid}               - UID of logged-in user
+   %{gid}               - Primary GID of logged-in user
    %{version}           - ProFTPD version
 */
 
@@ -192,11 +211,35 @@ static void logformat(char *nickname, char *fmts) {
       arg = NULL;
       tmp++;
       for (;;) {
+        pr_signals_handle();
+ 
+        if (strncmp(tmp, "{file-modified}", 15) == 0) {
+          add_meta(&outs, META_FILE_MODIFIED, 0);
+          tmp += 15;
+          continue;
+        }
+
+        if (strncmp(tmp, "{gid}", 5) == 0) {
+          add_meta(&outs, META_GID, 0);
+          tmp += 5;
+          continue;
+        }
+
+        if (strncmp(tmp, "{protocol}", 10) == 0) {
+          add_meta(&outs, META_PROTOCOL, 0);
+          tmp += 10;
+          continue;
+        }
+
+        if (strncmp(tmp, "{uid}", 5) == 0) {
+          add_meta(&outs, META_UID, 0);
+          tmp += 5;
+          continue;
+        }
 
         if (strncmp(tmp, "{version}", 9) == 0) {
           add_meta(&outs, META_VERSION, 0);
           tmp += 9;
-          *outs++ = *tmp;
           continue;
         }
 
@@ -229,6 +272,10 @@ static void logformat(char *nickname, char *fmts) {
             add_meta(&outs, META_DIR_NAME, 0);
             break;
 
+          case 'E':
+            add_meta(&outs, META_EOS_REASON, 0);
+            break;
+
           case 'e':
             if (arg) {
               add_meta(&outs, META_ENV_VAR, 0);
@@ -244,10 +291,18 @@ static void logformat(char *nickname, char *fmts) {
             add_meta(&outs, META_XFER_PATH, 0);
             break;
 
+          case 'H':
+            add_meta(&outs, META_VHOST_IP, 0);
+            break;
+
           case 'h':
             add_meta(&outs, META_REMOTE_HOST, 0);
             break;
 
+          case 'I':
+            add_meta(&outs, META_RAW_BYTES_IN, 0);
+            break;
+
           case 'J':
             add_meta(&outs, META_CMD_PARAMS, 0);
             break;
@@ -264,6 +319,10 @@ static void logformat(char *nickname, char *fmts) {
             add_meta(&outs, META_METHOD, 0);
             break;
 
+          case 'O':
+            add_meta(&outs, META_RAW_BYTES_OUT, 0);
+            break;
+
           case 'p':
             add_meta(&outs, META_LOCAL_PORT, 0);
             break;
@@ -310,9 +369,17 @@ static void logformat(char *nickname, char *fmts) {
             add_meta(&outs, META_LOCAL_FQDN, 0);
             break;
 
+          case 'w':
+            add_meta(&outs, META_RENAME_FROM, 0);
+            break;
+
           case '%':
             *outs++ = '%';
             break;
+
+          default:
+            *outs++ = *tmp;
+            break;
         }
 
         tmp++;
@@ -347,18 +414,22 @@ MODRET set_logformat(cmd_rec *cmd) {
   return PR_HANDLED(cmd);
 }
 
-static int _parse_classes(char *s) {
-  int classes = 0;
+static int parse_classes(char *s) {
+  int classes = CL_NONE;
   char *nextp = NULL;
 
   do {
+    pr_signals_handle();
 
-    if ((nextp = strchr(s, ',')))
+    nextp = strchr(s, ',');
+    if (nextp != NULL) {
       *nextp++ = '\0';
 
-    if (!nextp) {
-      if ((nextp = strchr(s, '|')))
+    } else {
+      nextp = strchr(s, '|');
+      if (nextp != NULL) {
         *nextp++ = '\0';
+      }
     }
 
     if (strcasecmp(s, "NONE") == 0) {
@@ -392,10 +463,18 @@ static int _parse_classes(char *s) {
                strcasecmp(s, "SECURE") == 0) {
       classes |= CL_SEC;
 
-    } else
-      pr_log_pri(PR_LOG_NOTICE, "ExtendedLog class '%s' is not defined.", s);
+    } else if (strcasecmp(s, "EXIT") == 0) {
+      classes |= CL_EXIT;
 
-  } while ((s = nextp));
+    } else {
+      pr_log_pri(PR_LOG_NOTICE, "ExtendedLog class '%s' is not defined", s);
+      return -1;
+    }
+
+    /* Advance to the next class in the list. */
+    s = nextp;
+
+  } while (s);
 
   return classes;
 }
@@ -421,21 +500,35 @@ MODRET set_extendedlog(cmd_rec *cmd) {
       CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "unknown syslog level: '",
         tmp, "'", NULL));
 
-    } else
+    } else {
       c->argv[0] = pstrdup(log_pool, cmd->argv[1]);
+    }
 
   } else if (cmd->argv[1][0] != '/') {
     CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "relative paths not allowed: '",
         cmd->argv[1], "'", NULL));
 
-  } else
+  } else {
     c->argv[0] = pstrdup(log_pool, cmd->argv[1]);
+  }
 
-  if (argc > 2)
-    c->argv[1] = pstrdup(log_pool, cmd->argv[2]);
+  if (argc > 2) {
+    int res;
 
-  if (argc > 3)
+    /* Parse the given class names, to make sure that they are all valid. */
+    res = parse_classes(cmd->argv[2]);
+    if (res < 0) {
+      CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "invalid log class in '",
+        cmd->argv[2], "'", NULL));    
+    }
+
+    c->argv[1] = palloc(c->pool, sizeof(int));
+    *((int *) c->argv[1]) = res;
+  }
+
+  if (argc > 3) {
     c->argv[2] = pstrdup(log_pool, cmd->argv[3]);
+  }
 
   c->argc = argc-1;
   return PR_HANDLED(cmd);
@@ -473,7 +566,7 @@ MODRET set_serverlog(cmd_rec *cmd) {
 /* Syntax: SystemLog <filename> */
 MODRET set_systemlog(cmd_rec *cmd) {
   char *syslogfn = NULL;
-  int res;
+  int res, xerrno = 0;
 
   CHECK_ARGS(cmd, 1);
   CHECK_CONF(cmd, CONF_ROOT);
@@ -492,10 +585,14 @@ MODRET set_systemlog(cmd_rec *cmd) {
 
   pr_signals_block();
 
+  PRIVS_ROOT
   res = log_opensyslog(syslogfn);
   if (res < 0) {
-    int xerrno = errno;
+    xerrno = errno;
+  }
+  PRIVS_RELINQUISH
 
+  if (res < 0) {
     pr_signals_unblock();
 
     if (res == PR_LOG_WRITABLE_DIR) {
@@ -535,7 +632,7 @@ static struct tm *_get_gmtoff(int *tz) {
 
 static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
   unsigned char *m;
-  char arg[512] = {'\0'}, *argp = NULL, *pass;
+  char arg[PR_TUNABLE_PATH_MAX+1] = {'\0'}, *argp = NULL, *pass;
 
   /* This function can cause potential problems.  Custom logformats
    * might overrun the arg buffer.  Fixing this problem involves a
@@ -558,7 +655,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
     case META_ANON_PASS:
       argp = arg;
 
-      pass = get_param_ptr(cmd->server->conf, C_PASS, FALSE);
+      pass = pr_table_get(session.notes, "mod_auth.anon-passwd", NULL);
       if (!pass)
         pass = "UNKNOWN";
 
@@ -573,7 +670,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
         snprintf(argp, sizeof(arg), "%" PR_LU,
           (pr_off_t) session.xfer.total_bytes);
 
-      } else if (strcmp(cmd->argv[0], C_DELE) == 0) {
+      } else if (pr_cmd_cmp(cmd, PR_CMD_DELE_ID) == 0) {
         snprintf(argp, sizeof(arg), "%" PR_LU, (pr_off_t) log_dele_filesz);
 
       } else
@@ -592,14 +689,14 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
     case META_DIR_NAME:
       argp = arg;
 
-      if (strcmp(cmd->argv[0], C_CDUP) == 0 ||
-          strcmp(cmd->argv[0], C_CWD) == 0 ||
-          strcmp(cmd->argv[0], C_MKD) == 0 ||
-          strcmp(cmd->argv[0], C_RMD) == 0 ||
-          strcmp(cmd->argv[0], C_XCWD) == 0 ||
-          strcmp(cmd->argv[0], C_XCUP) == 0 ||
-          strcmp(cmd->argv[0], C_XMKD) == 0 ||
-          strcmp(cmd->argv[0], C_XRMD) == 0) {
+      if (pr_cmd_cmp(cmd, PR_CMD_CDUP_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_CWD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_MKD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_RMD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_XCWD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_XCUP_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_XMKD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_XRMD_ID) == 0) {
         char *path, *tmp;
 
         path = pr_fs_decode_path(p, cmd->arg);
@@ -608,7 +705,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
         sstrncpy(argp, tmp ? tmp : path, sizeof(arg));
 
       } else {
-        sstrncpy(argp, "", sizeof(arg));
+        sstrncpy(argp, pr_fs_getvwd(), sizeof(arg));
       }
 
       m++;
@@ -617,17 +714,17 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
     case META_DIR_PATH:
       argp = arg;
 
-      if (strcmp(cmd->argv[0], C_CDUP) == 0 ||
-          strcmp(cmd->argv[0], C_MKD) == 0 ||
-          strcmp(cmd->argv[0], C_RMD) == 0 ||
-          strcmp(cmd->argv[0], C_XCUP) == 0 ||
-          strcmp(cmd->argv[0], C_XMKD) == 0 ||
-          strcmp(cmd->argv[0], C_XRMD) == 0) {
+      if (pr_cmd_cmp(cmd, PR_CMD_CDUP_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_MKD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_RMD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_XCUP_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_XMKD_ID) == 0 ||
+          pr_cmd_cmp(cmd, PR_CMD_XRMD_ID) == 0) {
         sstrncpy(argp, dir_abs_path(p, pr_fs_decode_path(p, cmd->arg), TRUE),
           sizeof(arg));
 
-      } else if (strcmp(cmd->argv[0], C_CWD) == 0 ||
-                 strcmp(cmd->argv[0], C_XCWD) == 0) {
+      } else if (pr_cmd_cmp(cmd, PR_CMD_CWD_ID) == 0 ||
+                 pr_cmd_cmp(cmd, PR_CMD_XCWD_ID) == 0) {
 
         /* Note: by this point in the dispatch cycle, the current working
          * directory has already been changed.  For the CWD/XCWD commands,
@@ -643,21 +740,40 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
             pdircat(p, session.chroot_path, pr_fs_getvwd(), NULL) :
             session.chroot_path, sizeof(arg));
 
-        } else
+        } else {
 
           /* Non-chrooted session. */
           sstrncpy(arg, pr_fs_getcwd(), sizeof(arg));
+        }
 
-      } else
+      } else {
         sstrncpy(argp, "", sizeof(arg));
+      }
 
       m++;
       break;
 
+    case META_EOS_REASON: {
+      const char *reason_str;
+      char *details = NULL;
+
+      argp = arg;
+
+      reason_str = pr_session_get_disconnect_reason(&details);
+      sstrncpy(argp, reason_str, sizeof(arg));
+      if (details != NULL) {
+        sstrcat(argp, ": ", sizeof(arg));
+        sstrcat(argp, details, sizeof(arg));
+      }
+
+      m++;
+      break;
+    }
+
     case META_FILENAME:
       argp = arg;
 
-      if (strcmp(cmd->argv[0], C_RNTO) == 0) {
+      if (pr_cmd_cmp(cmd, PR_CMD_RNTO_ID) == 0) {
         sstrncpy(argp, dir_abs_path(p, pr_fs_decode_path(p, cmd->arg), TRUE),
           sizeof(arg));
 
@@ -665,7 +781,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
                  session.xfer.path) {
         sstrncpy(argp, dir_abs_path(p, session.xfer.path, TRUE), sizeof(arg));
 
-      } else if (strcmp(cmd->argv[0], C_SITE) == 0 &&
+      } else if (pr_cmd_cmp(cmd, PR_CMD_SITE_ID) == 0 &&
                  (strcasecmp(cmd->argv[1], "CHGRP") == 0 ||
                   strcasecmp(cmd->argv[1], "CHMOD") == 0)) {
         register unsigned int i;
@@ -683,17 +799,18 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
          * filenames that are not stored in the session.xfer structure; these
          * should be expanded properly as well.
          */
-        if (strcmp(cmd->argv[0], C_DELE) == 0 ||
-            strcmp(cmd->argv[0], C_MKD) == 0 ||
-            strcmp(cmd->argv[0], C_RMD) == 0 ||
-            strcmp(cmd->argv[0], C_XMKD) == 0 ||
-            strcmp(cmd->argv[0], C_XRMD) == 0)
+        if (pr_cmd_cmp(cmd, PR_CMD_DELE_ID) == 0 ||
+            pr_cmd_cmp(cmd, PR_CMD_MKD_ID) == 0 ||
+            pr_cmd_cmp(cmd, PR_CMD_RMD_ID) == 0 ||
+            pr_cmd_cmp(cmd, PR_CMD_XMKD_ID) == 0 ||
+            pr_cmd_cmp(cmd, PR_CMD_XRMD_ID) == 0) {
           sstrncpy(arg, dir_abs_path(p, pr_fs_decode_path(p, cmd->arg), TRUE),
             sizeof(arg));
 
-        else
+        } else {
           /* All other situations get a "-".  */
           sstrncpy(argp, "-", sizeof(arg));
+        }
       }
 
       m++;
@@ -702,7 +819,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
     case META_XFER_PATH:
       argp = arg;
 
-      if (strcmp(cmd->argv[0], C_RNTO) == 0) {
+      if (pr_cmd_cmp(cmd, PR_CMD_RNTO_ID) == 0) {
         char *path;
 
         path = dir_best_path(cmd->tmp_pool,
@@ -714,19 +831,24 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
         sstrncpy(argp, session.xfer.path, sizeof(arg));
 
       } else {
-        /* Some commands (i.e. DELE) have associated filenames that are not
-         * stored in the session.xfer structure; these should be expanded
-         * properly as well.
+        /* Some commands (i.e. DELE, MKD, XMKD, RMD, XRMD) have associated
+         * filenames that are not stored in the session.xfer structure; these
+         * should be expanded properly as well.
          */
-        if (strcmp(cmd->argv[0], C_DELE) == 0) {
+        if (pr_cmd_cmp(cmd, PR_CMD_DELE_ID) == 0 ||
+            pr_cmd_cmp(cmd, PR_CMD_MKD_ID) == 0 ||
+            pr_cmd_cmp(cmd, PR_CMD_XMKD_ID) == 0 ||
+            pr_cmd_cmp(cmd, PR_CMD_RMD_ID) == 0 ||
+            pr_cmd_cmp(cmd, PR_CMD_XRMD_ID) == 0) {
           char *path;
 
           path = dir_best_path(cmd->tmp_pool,
             pr_fs_decode_path(cmd->tmp_pool, cmd->arg));
           sstrncpy(arg, path, sizeof(arg));
 
-        } else
+        } else {
           sstrncpy(argp, "-", sizeof(arg));
+        }
       }
 
       m++;
@@ -762,6 +884,25 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
       m++;
       break;
 
+    case META_RENAME_FROM: {
+      char *rnfr_path = "-";
+
+      argp = arg;
+      if (pr_cmd_cmp(cmd, PR_CMD_RNTO_ID) == 0) {
+        rnfr_path = pr_table_get(session.notes, "mod_core.rnfr-path", NULL);
+        if (rnfr_path != NULL) {
+          rnfr_path = dir_abs_path(p, pr_fs_decode_path(p, rnfr_path), TRUE);
+
+        } else {
+          rnfr_path = "-";
+        }
+      }
+
+      sstrncpy(argp, rnfr_path, sizeof(arg));
+      m++;
+      break;
+    }
+
     case META_IDENT_USER: {
       char *rfc1413_ident;
 
@@ -779,7 +920,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
     case META_METHOD:
       argp = arg;
 
-      if (strcmp(cmd->argv[0], C_SITE) != 0) {
+      if (pr_cmd_cmp(cmd, PR_CMD_SITE_ID) != 0) {
         sstrncpy(argp, cmd->argv[0], sizeof(arg));
 
       } else {
@@ -813,7 +954,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
 
     case META_LOCAL_FQDN:
       argp = arg;
-      sstrncpy(argp, cmd->server->ServerFQDN, sizeof(arg));
+      sstrncpy(argp, pr_netaddr_get_dnsstr(session.c->local_addr), sizeof(arg));
       m++;
       break;
 
@@ -871,13 +1012,13 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
             session.xfer.start_time.tv_usec != 0) {
           struct timeval end_time;
 
-          gettimeofday(&end_time,NULL);
+          gettimeofday(&end_time, NULL);
           end_time.tv_sec -= session.xfer.start_time.tv_sec;
 
-          if (end_time.tv_usec >= session.xfer.start_time.tv_usec)
+          if (end_time.tv_usec >= session.xfer.start_time.tv_usec) {
             end_time.tv_usec -= session.xfer.start_time.tv_usec;
 
-          else {
+          } else {
             end_time.tv_usec = 1000000L - (session.xfer.start_time.tv_usec -
               end_time.tv_usec);
             end_time.tv_sec--;
@@ -886,11 +1027,13 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
           snprintf(argp, sizeof(arg), "%ld.%03ld", (long) end_time.tv_sec,
             (long) (end_time.tv_usec / 1000));
 
-        } else
+        } else {
           sstrncpy(argp, "-", sizeof(arg));
+        }
 
-      } else
+      } else {
         sstrncpy(argp, "-", sizeof(arg));
+      }
 
       m++;
       break;
@@ -898,7 +1041,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
     case META_COMMAND:
       argp = arg;
 
-      if (strcasecmp(cmd->argv[0], C_PASS) == 0 &&
+      if (pr_cmd_cmp(cmd, PR_CMD_PASS_ID) == 0 &&
           session.hide_password) {
         sstrncpy(argp, "PASS (hidden)", sizeof(arg));
 
@@ -911,7 +1054,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
 
     case META_CMD_PARAMS:
       argp = arg;
-      if (strcasecmp(cmd->argv[0], C_PASS) == 0 &&
+      if (pr_cmd_cmp(cmd, PR_CMD_PASS_ID) == 0 &&
           session.hide_password) {
         sstrncpy(argp, "(hidden)", sizeof(arg));
 
@@ -931,10 +1074,11 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
     case META_USER:
       argp = arg;
 
-      if (!session.user) {
+      if (session.user == NULL) {
         char *u = get_param_ptr(cmd->server->conf, "UserName", FALSE);
-        if (!u)
+        if (u == NULL) {
           u = "root";
+        }
 
         sstrncpy(argp, u, sizeof(arg));
 
@@ -946,31 +1090,35 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
       break;
 
     case META_ORIGINAL_USER: {
-      char *login_user = get_param_ptr(main_server->conf, C_USER, FALSE);
+      char *login_user;
+
       argp = arg;
 
-      if (login_user)
+      login_user = pr_table_get(session.notes, "mod_auth.orig-user", NULL);
+      if (login_user) {
         sstrncpy(argp, login_user, sizeof(arg));
-      else
+
+      } else {
         sstrncpy(argp, "(none)", sizeof(arg));
+      }
 
       m++;
       break;
     }
 
     case META_RESPONSE_CODE: {
-      pr_response_t *r;
+      char *resp_code = NULL;
+      int res;
 
       argp = arg;
-      r = (resp_list ? resp_list : resp_err_list);
 
-      for (; r && !r->num; r = r->next) ;
-      if (r &&
-          r->num) {
-        sstrncpy(argp, r->num, sizeof(arg));
+      res = pr_response_get_last(cmd->tmp_pool, &resp_code, NULL);
+      if (res == 0 &&
+          resp_code != NULL) {
+        sstrncpy(argp, resp_code, sizeof(arg));
 
       /* Hack to add return code for proper logging of QUIT command. */
-      } else if (strcasecmp(cmd->argv[0], C_QUIT) == 0) {
+      } else if (pr_cmd_cmp(cmd, PR_CMD_QUIT_ID) == 0) {
         sstrncpy(argp, R_221, sizeof(arg));
 
       } else {
@@ -982,15 +1130,15 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
     }
 
     case META_RESPONSE_STR: {
-      pr_response_t *r;
+      char *resp_msg = NULL;
+      int res;
 
       argp = arg;
-      r = (resp_list ? resp_list : resp_err_list);
 
-      for (; r && !r->msg; r = r->next) ;
-      if (r &&
-          r->msg) {
-        sstrncpy(argp, r->msg, sizeof(arg));
+      res = pr_response_get_last(cmd->tmp_pool, NULL, &resp_msg);
+      if (res == 0 &&
+          resp_msg != NULL) {
+        sstrncpy(argp, resp_msg, sizeof(arg));
 
       } else {
         sstrncpy(argp, "-", sizeof(arg));
@@ -1000,12 +1148,65 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
       break;
     }
 
+    case META_PROTOCOL:
+      argp = arg;
+      sstrncpy(argp, pr_session_get_protocol(0), sizeof(arg));
+      m++;
+      break;
+
+    case META_UID:
+      argp = arg;
+      snprintf(argp, sizeof(arg), "%lu", (unsigned long) session.login_uid);
+      m++;
+      break;
+
+    case META_GID:
+      argp = arg;
+      snprintf(argp, sizeof(arg), "%lu", (unsigned long) session.login_gid);
+      m++;
+      break;
+
     case META_VERSION:
       argp = arg;
       sstrncpy(argp, PROFTPD_VERSION_TEXT, sizeof(arg));
       m++;
       break;
 
+    case META_FILE_MODIFIED: {
+      char *modified;
+
+      argp = arg;
+
+      modified = pr_table_get(cmd->notes, "mod_xfer.file-modified", NULL);
+      if (modified) {
+        sstrncpy(argp, modified, sizeof(arg));
+
+      } else {
+        sstrncpy(argp, "false", sizeof(arg));
+      }
+
+      m++;
+      break;
+    }
+
+    case META_RAW_BYTES_IN:
+      argp = arg;
+      snprintf(argp, sizeof(arg), "%" PR_LU, (pr_off_t) session.total_raw_in);
+      m++;
+      break;
+
+    case META_RAW_BYTES_OUT:
+      argp = arg;
+      snprintf(argp, sizeof(arg), "%" PR_LU, (pr_off_t) session.total_raw_out);
+      m++;
+      break;
+
+    case META_VHOST_IP:
+      argp = arg;
+      sstrncpy(argp, cmd->server->ServerAddress, sizeof(arg));
+      m++;
+      break;
+
   }
  
   *f = m;
@@ -1025,15 +1226,17 @@ static void do_log(cmd_rec *cmd, logfile_t *lf) {
   char logbuf[EXTENDED_LOG_BUFFER_SIZE] = {'\0'};
   logformat_t *fmt = NULL;
   char *s, *bp;
+  size_t logbuflen;
 
   fmt = lf->lf_format;
   f = fmt->lf_format;
   bp = logbuf;
 
   while (*f && size) {
+    pr_signals_handle();
+
     if (*f == META_START) {
       s = get_next_meta(cmd->tmp_pool, cmd, &f);
-
       if (s) {
         size_t tmp;
 
@@ -1055,14 +1258,21 @@ static void do_log(cmd_rec *cmd, logfile_t *lf) {
   *bp++ = '\n';
   *bp = '\0';
 
+  logbuflen = strlen(logbuf);
+
   if (lf->lf_fd != EXTENDED_LOG_SYSLOG) {
-    if (write(lf->lf_fd, logbuf, strlen(logbuf)) < 0) {
+    pr_log_event_generate(PR_LOG_TYPE_EXTLOG, lf->lf_fd, -1, logbuf, logbuflen);
+
+    if (write(lf->lf_fd, logbuf, logbuflen) < 0) {
       pr_log_pri(PR_LOG_ERR, "error: cannot write ExtendedLog to fd %d: %s",
         lf->lf_fd, strerror(errno));
     }
 
-  } else
+  } else {
+    pr_log_event_generate(PR_LOG_TYPE_EXTLOG, syslog_sockfd,
+      lf->lf_syslog_level, logbuf, logbuflen);
     pr_syslog(syslog_sockfd, lf->lf_syslog_level, "%s", logbuf);
+  }
 }
 
 MODRET log_any(cmd_rec *cmd) {
@@ -1071,7 +1281,17 @@ MODRET log_any(cmd_rec *cmd) {
   /* If not in anon mode, only handle logs for main servers */
   for (lf = logs; lf; lf = lf->next) {
     if (lf->lf_fd != -1 &&
-        (cmd->class & lf->lf_classes)) {
+
+        /* If the logging class of this command is one of the classes
+         * configured for this ExtendedLog...
+         */ 
+        ((cmd->class & lf->lf_classes) ||
+
+         /* ...or if the logging class of this command is unknown (defaults to
+          * zero), and this ExtendedLog is configured to log ALL commands, then
+          * log it.
+          */
+         (cmd->class == 0 && lf->lf_classes == CL_ALL))) {
 
       if (!session.anon_config &&
           lf->lf_conf &&
@@ -1085,6 +1305,18 @@ MODRET log_any(cmd_rec *cmd) {
   return PR_DECLINED(cmd);
 }
 
+/* Event handlers
+ */
+
+static void log_exit_ev(const void *event_data, void *user_data) {
+  cmd_rec *cmd;
+
+  cmd = pr_cmd_alloc(session.pool, 1, "EXIT");
+  cmd->class |= CL_EXIT;
+
+  (void) log_any(cmd);
+}
+
 static void log_restart_ev(const void *event_data, void *user_data) {
   destroy_pool(log_pool);
 
@@ -1101,6 +1333,19 @@ static void log_restart_ev(const void *event_data, void *user_data) {
   return;
 }
 
+static void log_xfer_stalled_ev(const void *event_data, void *user_data) {
+  if (session.curr_cmd_rec != NULL) {
+    /* Automatically dispatch the current command, at the LOG_CMD_ERR phase,
+     * so that the ExtendedLog entry for the command gets written out.  This
+     * should handle any LIST/MLSD/NLST commands as well (Bug#3696).
+     */
+    (void) log_any(session.curr_cmd_rec);
+  }
+}
+
+/* Initialization handlers
+ */
+
 static int log_init(void) {
   log_pool = make_sub_pool(permanent_pool);
   pr_pool_tag(log_pool, "mod_log pool");
@@ -1114,10 +1359,9 @@ static int log_init(void) {
 
 static void find_extendedlogs(void) {
   config_rec *c;
-  char *logfname;
+  char *logfname, *logfmt_s;
   int logclasses = CL_ALL;
   logformat_t *logfmt;
-  char *logfmt_s = NULL;
   logfile_t *extlog = NULL;
 
   /* We _do_ actually want the recursion here.  The reason is that we want
@@ -1129,28 +1373,39 @@ static void find_extendedlogs(void) {
    */
 
   c = find_config(main_server->conf, CONF_PARAM, "ExtendedLog", TRUE);
-
   while (c) {
+    pr_signals_handle();
+
     logfname = c->argv[0];
+    logfmt_s = NULL;
 
     if (c->argc > 1) {
-      logclasses = _parse_classes(c->argv[1]);
-      if (c->argc > 2)
+      logclasses = *((int *) c->argv[1]);
+
+      if (c->argc > 2) {
         logfmt_s = c->argv[2];
+      }
     }
 
-    /* No logging for this round.
+    /* No logging for this round.  If, however, this was found in an
+     * <Anonymous> section, add a logfile entry for it anyway; the anonymous
+     * directive might be trying to override a higher-level config; see
+     * Bug#1908.
      */
-    if (logclasses == CL_NONE)
+    if (logclasses == CL_NONE &&
+        (c->parent != NULL && c->parent->config_type != CONF_ANON)) {
       goto loop_extendedlogs;
+    }
 
     if (logfmt_s) {
       /* search for the format-nickname */
-      for (logfmt = formats; logfmt; logfmt = logfmt->next)
-        if (strcmp(logfmt->lf_nickname, logfmt_s) == 0)
+      for (logfmt = formats; logfmt; logfmt = logfmt->next) {
+        if (strcmp(logfmt->lf_nickname, logfmt_s) == 0) {
           break;
+        }
+      }
 
-      if (!logfmt) {
+      if (logfmt == NULL) {
         pr_log_pri(PR_LOG_NOTICE,
           "ExtendedLog '%s' uses unknown format nickname '%s'", logfname,
           logfmt_s);
@@ -1209,10 +1464,12 @@ MODRET log_post_pass(cmd_rec *cmd) {
    */
   if (!session.anon_config) {
     for (lf = logs; lf; lf = lf->next) {
-      if (lf->lf_fd != -1 && lf->lf_fd != EXTENDED_LOG_SYSLOG &&
-          lf->lf_conf && lf->lf_conf->config_type == CONF_ANON) {
-        pr_log_debug(DEBUG7, "mod_log: closing ExtendedLog '%s'",
-          lf->lf_filename);
+      if (lf->lf_fd != -1 &&
+          lf->lf_fd != EXTENDED_LOG_SYSLOG &&
+          lf->lf_conf &&
+          lf->lf_conf->config_type == CONF_ANON) {
+        pr_log_debug(DEBUG7, "mod_log: closing ExtendedLog '%s' (fd %d)",
+          lf->lf_filename, lf->lf_fd);
         close(lf->lf_fd);
         lf->lf_fd = -1;
       }
@@ -1223,10 +1480,12 @@ MODRET log_post_pass(cmd_rec *cmd) {
      * context.
      */
     for (lf = logs; lf; lf = lf->next) {
-      if (lf->lf_fd != -1 && lf->lf_fd != EXTENDED_LOG_SYSLOG &&
-          lf->lf_conf && lf->lf_conf != session.anon_config) {
-        pr_log_debug(DEBUG7, "mod_log: closing ExtendedLog '%s'",
-          lf->lf_filename);
+      if (lf->lf_fd != -1 &&
+          lf->lf_fd != EXTENDED_LOG_SYSLOG &&
+          lf->lf_conf &&
+          lf->lf_conf != session.anon_config) {
+        pr_log_debug(DEBUG7, "mod_log: closing ExtendedLog '%s' (fd %d)",
+          lf->lf_filename, lf->lf_fd);
         close(lf->lf_fd);
         lf->lf_fd = -1;
       }
@@ -1236,7 +1495,8 @@ MODRET log_post_pass(cmd_rec *cmd) {
      * close the outer (this allows overriding inside <Anonymous>).
      */
     for (lf = logs; lf; lf = lf->next) {
-      if (lf->lf_conf && lf->lf_conf == session.anon_config) {
+      if (lf->lf_conf &&
+          lf->lf_conf == session.anon_config) {
         /* This should "override" any lower-level extendedlog with the
          * same filename.
          */
@@ -1247,8 +1507,8 @@ MODRET log_post_pass(cmd_rec *cmd) {
               lfi->lf_fd != EXTENDED_LOG_SYSLOG &&
               !lfi->lf_conf &&
               strcmp(lfi->lf_filename, lf->lf_filename) == 0) {
-            pr_log_debug(DEBUG7, "mod_log: closing ExtendedLog '%s'",
-              lf->lf_filename);
+            pr_log_debug(DEBUG7, "mod_log: closing ExtendedLog '%s' (fd %d)",
+              lf->lf_filename, lfi->lf_fd);
             close(lfi->lf_fd);
             lfi->lf_fd = -1;
           }
@@ -1258,8 +1518,8 @@ MODRET log_post_pass(cmd_rec *cmd) {
         if (lf->lf_fd != -1 &&
             lf->lf_fd != EXTENDED_LOG_SYSLOG &&
             lf->lf_classes == CL_NONE) {
-          pr_log_debug(DEBUG7, "mod_log: closing ExtendedLog '%s'",
-            lf->lf_filename);
+          pr_log_debug(DEBUG7, "mod_log: closing ExtendedLog '%s' (fd %d)",
+            lf->lf_filename, lf->lf_fd);
           close(lf->lf_fd);
           lf->lf_fd = -1;
         }
@@ -1276,8 +1536,8 @@ static int log_sess_init(void) {
   logfile_t *lf = NULL;
 
   /* Open the ServerLog, if present. */
-  if ((serverlog_name = get_param_ptr(main_server->conf, "ServerLog",
-      FALSE)) != NULL) {
+  serverlog_name = get_param_ptr(main_server->conf, "ServerLog", FALSE);
+  if (serverlog_name != NULL) {
     PRIVS_ROOT
     log_closesyslog();
     log_opensyslog(serverlog_name);
@@ -1288,7 +1548,6 @@ static int log_sess_init(void) {
   find_extendedlogs();
 
   for (lf = logs; lf; lf = lf->next) {
-
     if (lf->lf_fd == -1) {
 
       /* Is this ExtendedLog to be written to a file, or to syslog? */
@@ -1300,26 +1559,23 @@ static int log_sess_init(void) {
 
         pr_signals_block();
         PRIVS_ROOT
-        res = pr_log_openfile(lf->lf_filename, &lf->lf_fd, EXTENDED_LOG_MODE);
+        res = pr_log_openfile(lf->lf_filename, &(lf->lf_fd), EXTENDED_LOG_MODE);
         PRIVS_RELINQUISH
         pr_signals_unblock();
 
-        if (res == -1) {
-          pr_log_pri(PR_LOG_NOTICE, "unable to open ExtendedLog '%s': %s",
-            lf->lf_filename, strerror(errno));
-          continue;
+        if (res < 0) {
+          if (res == -1) {
+            pr_log_pri(PR_LOG_NOTICE, "unable to open ExtendedLog '%s': %s",
+              lf->lf_filename, strerror(errno));
 
-        } else if (res == PR_LOG_WRITABLE_DIR) {
-          pr_log_pri(PR_LOG_NOTICE, "unable to open ExtendedLog '%s': "
-            "containing directory is world writable", lf->lf_filename);
-          continue;
+          } else if (res == PR_LOG_WRITABLE_DIR) {
+            pr_log_pri(PR_LOG_NOTICE, "unable to open ExtendedLog '%s': "
+              "containing directory is world writable", lf->lf_filename);
 
-        } else if (res == PR_LOG_SYMLINK) {
-          pr_log_pri(PR_LOG_NOTICE, "unable to open ExtendedLog '%s': "
-            "%s is a symbolic link", lf->lf_filename, lf->lf_filename);
-          close(lf->lf_fd);
-          lf->lf_fd = -1;
-          continue;
+          } else if (res == PR_LOG_SYMLINK) {
+            pr_log_pri(PR_LOG_NOTICE, "unable to open ExtendedLog '%s': "
+              "%s is a symbolic link", lf->lf_filename, lf->lf_filename);
+          }
         }
 
       } else {
@@ -1331,6 +1587,11 @@ static int log_sess_init(void) {
     }
   }
 
+  /* Register event handlers for the session. */
+  pr_event_register(&log_module, "core.exit", log_exit_ev, NULL);
+  pr_event_register(&log_module, "core.timeout-stalled", log_xfer_stalled_ev,
+    NULL);
+
   return 0;
 }
 
