@@ -1,7 +1,7 @@
 /*
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -23,15 +23,11 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/*
- * "SITE" commands module for ProFTPD
- * $Id: mod_site.c,v 1.44 2004/12/12 05:59:27 castaglia Exp $
+/* "SITE" commands module for ProFTPD
+ * $Id: mod_site.c,v 1.57 2011/05/23 21:11:56 castaglia Exp $
  */
 
 #include "conf.h"
-#ifdef HAVE_REGEX_H
-#include <regex.h>
-#endif
 
 /* From mod_core.c */
 extern int core_chmod(cmd_rec *cmd, char *dir, mode_t mode);
@@ -52,27 +48,34 @@ static struct {
 
 static char *_get_full_cmd(cmd_rec *cmd) {
   char *res = "";
+  size_t reslen = 0;
   int i;
 
   for (i = 0; i < cmd->argc; i++)
-    res = pstrcat(cmd->tmp_pool,res,cmd->argv[i]," ",NULL);
+    res = pstrcat(cmd->tmp_pool, res, cmd->argv[i], " ", NULL);
 
-  while (res[strlen(res)-1] == ' ')
-    res[strlen(res)-1] = '\0';
+  reslen = strlen(res);
+  while (reslen >= 1 &&
+         res[reslen-1] == ' ') {
+    res[reslen-1] = '\0';
+    reslen = strlen(res);
+  }
 
   return res;
 }
 
 MODRET site_chgrp(cmd_rec *cmd) {
+  int res;
   gid_t gid;
   char *path = NULL, *tmp = NULL, *arg = "";
   register unsigned int i = 0;
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-  regex_t *preg;
+#ifdef PR_USE_REGEX
+  pr_regex_t *pre;
 #endif
 
   if (cmd->argc < 3) {
-    pr_response_add_err(R_500, "'SITE %s' not understood", _get_full_cmd(cmd));
+    pr_response_add_err(R_500, _("'SITE %s' not understood"),
+      _get_full_cmd(cmd));
     return NULL;
   }
 
@@ -80,25 +83,26 @@ MODRET site_chgrp(cmd_rec *cmd) {
    * the mode, separating them with spaces.
    */
   for (i = 2; i <= cmd->argc-1; i++)
-    arg = pstrcat(cmd->tmp_pool, arg, *arg ? " " : "", cmd->argv[i], NULL);
+    arg = pstrcat(cmd->tmp_pool, arg, *arg ? " " : "",
+      pr_fs_decode_path(cmd->tmp_pool, cmd->argv[i]), NULL);
 
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-  preg = (regex_t *) get_param_ptr(CURRENT_CONF, "PathAllowFilter", FALSE);
-
-  if (preg && regexec(preg, arg, 0, NULL, 0) != 0) {
+#ifdef PR_USE_REGEX
+  pre = get_param_ptr(CURRENT_CONF, "PathAllowFilter", FALSE);
+  if (pre != NULL &&
+      pr_regexp_exec(pre, arg, 0, NULL, 0, 0, 0) != 0) {
     pr_log_debug(DEBUG2, "'%s %s' denied by PathAllowFilter", cmd->argv[0],
-      cmd->arg);
-    pr_response_add_err(R_550, "%s: Forbidden filename", arg);
-    return ERROR(cmd);
+      arg);
+    pr_response_add_err(R_550, _("%s: Forbidden filename"), cmd->arg);
+    return PR_ERROR(cmd);
   }
 
-  preg = (regex_t *) get_param_ptr(CURRENT_CONF, "PathDenyFilter", FALSE);
-
-  if (preg && regexec(preg, arg, 0, NULL, 0) == 0) {
+  pre = get_param_ptr(CURRENT_CONF, "PathDenyFilter", FALSE);
+  if (pre != NULL &&
+      pr_regexp_exec(pre, arg, 0, NULL, 0, 0, 0) == 0) {
     pr_log_debug(DEBUG2, "'%s %s' denied by PathDenyFilter", cmd->argv[0],
-      cmd->arg);
-    pr_response_add_err(R_550, "%s: Forbidden filename", arg);
-    return ERROR(cmd);
+      arg);
+    pr_response_add_err(R_550, _("%s: Forbidden filename"), cmd->arg);
+    return PR_ERROR(cmd);
   }
 #endif
 
@@ -106,7 +110,7 @@ MODRET site_chgrp(cmd_rec *cmd) {
 
   if (!path) {
     pr_response_add_err(R_550, "%s: %s", arg, strerror(errno));
-    return ERROR(cmd);
+    return PR_ERROR(cmd);
   }
 
   /* Map the given group argument, if a string, to a GID.  If already a
@@ -116,34 +120,49 @@ MODRET site_chgrp(cmd_rec *cmd) {
 
   if (tmp && *tmp) {
 
-    /* Try the parameter as a user name. */
+    /* Try the parameter as a group name. */
     gid = pr_auth_name2gid(cmd->tmp_pool, cmd->argv[1]);
     if (gid == (gid_t) -1) {
+      pr_log_debug(DEBUG9,
+        "SITE CHGRP: Unable to resolve group name '%s' to GID", cmd->argv[1]);
       pr_response_add_err(R_550, "%s: %s", arg, strerror(EINVAL));
-      return ERROR(cmd);
+      return PR_ERROR(cmd);
     }
   }
 
-  if (core_chgrp(cmd, path, (uid_t) -1, gid) == -1) {
-    pr_response_add_err(R_550, "%s: %s", arg, strerror(errno));
-    return ERROR(cmd);
+  res = core_chgrp(cmd, path, (uid_t) -1, gid);
+  if (res < 0) {
+    int xerrno = errno;
 
-  } else
-    pr_response_add(R_200, "SITE %s command successful", cmd->argv[0]);
+    (void) pr_trace_msg("fileperms", 1, "%s, user '%s' (UID %lu, GID %lu): "
+      "error chown'ing '%s' to GID %lu: %s", cmd->argv[0], session.user,
+      (unsigned long) session.uid, (unsigned long) session.gid,
+      path, (unsigned long) gid, strerror(xerrno));
 
-  return HANDLED(cmd);
+    pr_response_add_err(R_550, "%s: %s", arg, strerror(xerrno));
+
+    errno = xerrno;
+    return PR_ERROR(cmd);
+
+  } else {
+    pr_response_add(R_200, _("SITE %s command successful"), cmd->argv[0]);
+  }
+
+  return PR_HANDLED(cmd);
 }
 
 MODRET site_chmod(cmd_rec *cmd) {
+  int res;
   mode_t mode = 0;
   char *dir, *endp, *tmp, *arg = "";
   register unsigned int i = 0;
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-  regex_t *preg;
+#ifdef PR_USE_REGEX
+  pr_regex_t *pre;
 #endif
 
   if (cmd->argc < 3) {
-    pr_response_add_err(R_500, "'SITE %s' not understood", _get_full_cmd(cmd));
+    pr_response_add_err(R_500, _("'SITE %s' not understood"),
+      _get_full_cmd(cmd));
     return NULL;
   }
 
@@ -151,25 +170,26 @@ MODRET site_chmod(cmd_rec *cmd) {
    * the mode, separating them with spaces.
    */
   for (i = 2; i <= cmd->argc-1; i++)
-    arg = pstrcat(cmd->tmp_pool, arg, *arg ? " " : "", cmd->argv[i], NULL);
+    arg = pstrcat(cmd->tmp_pool, arg, *arg ? " " : "",
+      pr_fs_decode_path(cmd->tmp_pool, cmd->argv[i]), NULL);
 
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-  preg = (regex_t *) get_param_ptr(CURRENT_CONF, "PathAllowFilter", FALSE);
-
-  if (preg && regexec(preg, arg, 0, NULL, 0) != 0) {
+#ifdef PR_USE_REGEX
+  pre = get_param_ptr(CURRENT_CONF, "PathAllowFilter", FALSE);
+  if (pre != NULL &&
+      pr_regexp_exec(pre, arg, 0, NULL, 0, 0, 0) != 0) {
     pr_log_debug(DEBUG2, "'%s %s %s' denied by PathAllowFilter", cmd->argv[0],
       cmd->argv[1], arg);
-    pr_response_add_err(R_550, "%s: Forbidden filename", arg);
-    return ERROR(cmd);
+    pr_response_add_err(R_550, _("%s: Forbidden filename"), cmd->arg);
+    return PR_ERROR(cmd);
   }
 
-  preg = (regex_t *) get_param_ptr(CURRENT_CONF, "PathDenyFilter", FALSE);
-
-  if (preg && regexec(preg, arg, 0, NULL, 0) == 0) {
+  pre = get_param_ptr(CURRENT_CONF, "PathDenyFilter", FALSE);
+  if (pre != NULL &&
+      pr_regexp_exec(pre, arg, 0, NULL, 0, 0, 0) == 0) {
     pr_log_debug(DEBUG2, "'%s %s %s' denied by PathDenyFilter", cmd->argv[0],
       cmd->argv[1], arg);
-    pr_response_add_err(R_550, "%s: Forbidden filename", arg);
-    return ERROR(cmd);
+    pr_response_add_err(R_550, _("%s: Forbidden filename"), cmd->arg);
+    return PR_ERROR(cmd);
   }
 #endif
 
@@ -177,7 +197,7 @@ MODRET site_chmod(cmd_rec *cmd) {
 
   if (!dir) {
     pr_response_add_err(R_550, "%s: %s", arg, strerror(errno));
-    return ERROR(cmd);
+    return PR_ERROR(cmd);
   }
 
   /* If the first character isn't '0', prepend it and attempt conversion.
@@ -185,42 +205,49 @@ MODRET site_chmod(cmd_rec *cmd) {
    * case where an octal number is sent without the leading '0'.
    */
 
-  if (cmd->argv[1][0] != '0')
+  if (cmd->argv[1][0] != '0') {
     tmp = pstrcat(cmd->tmp_pool, "0", cmd->argv[1], NULL);
-  else
+
+  } else {
     tmp = cmd->argv[1];
+  }
 
   mode = strtol(tmp,&endp,0);
   if (endp && *endp) {
     /* It's not an absolute number, try symbolic */
     char *cp = cmd->argv[1];
-    int mask = 0,mode_op = 0,curmode = 0,curumask = umask(0);
+    int mask = 0, mode_op = 0, curmode = 0, curumask = umask(0);
     int invalid = 0;
-    char *who,*how,*what;
-    struct stat sbuf;
+    char *who, *how, *what;
+    struct stat st;
 
     umask(curumask);
     mode = 0;
 
-    if (pr_fsio_stat(dir, &sbuf) != -1)
-      curmode = sbuf.st_mode;
+    if (pr_fsio_stat(dir, &st) != -1)
+      curmode = st.st_mode;
 
     while (TRUE) {
-      who = pstrdup(cmd->tmp_pool,cp);
-      if ((tmp = strpbrk(who,"+-=")) != NULL) {
-        how = pstrdup(cmd->tmp_pool,tmp);
+      who = pstrdup(cmd->tmp_pool, cp);
+
+      tmp = strpbrk(who, "+-=");
+      if (tmp != NULL) {
+        how = pstrdup(cmd->tmp_pool, tmp);
         if (*how != '=')
           mode = curmode;
 
         *tmp = '\0';
+
       } else {
         invalid++;
         break;
       }
 
-      if ((tmp = strpbrk(how,"rwxXstugo")) != NULL) {
-        what = pstrdup(cmd->tmp_pool,tmp);
+      tmp = strpbrk(how, "rwxXstugo");
+      if (tmp != NULL) {
+        what = pstrdup(cmd->tmp_pool, tmp);
         *tmp = '\0';
+
       } else {
         invalid++;
         break;
@@ -229,117 +256,150 @@ MODRET site_chmod(cmd_rec *cmd) {
       cp = what;
       while (cp) {
         switch (*who) {
-        case 'u':
-          mask = 0077;
-          break;
-        case 'g':
-          mask = 0707;
-          break;
-        case 'o':
-          mask = 0770;
-          break;
-        case 'a':
-          mask = 0000;
-          break;
-        case '\0':
-          mask = curumask;
-          break;
-        default:
-          invalid++;
-          break;
+          case 'u':
+            mask = 0077;
+            break;
+
+          case 'g':
+            mask = 0707;
+            break;
+
+          case 'o':
+            mask = 0770;
+            break;
+
+          case 'a':
+            mask = 0000;
+            break;
+
+          case '\0':
+            mask = curumask;
+            break;
+
+          default:
+            invalid++;
+            break;
         }
 
-        if (invalid) break;
+        if (invalid)
+          break;
 
         switch (*how) {
-        case '+':
-        case '-':
-        case '=':
-          break;
-        default:
-          invalid++;
+          case '+':
+          case '-':
+          case '=':
+            break;
+
+          default:
+            invalid++;
         }
 
-        if (invalid) break;
+        if (invalid)
+          break;
 
         switch (*cp) {
-        case 'r':
-          mode_op |= (S_IRUSR|S_IRGRP|S_IROTH);
-          break;
-        case 'w':
-          mode_op |= (S_IWUSR|S_IWGRP|S_IWOTH);
-          break;
-        case 'x':
-          mode_op |= (S_IXUSR|S_IXGRP|S_IXOTH);
-          break;
-        /* 'X' not implemented */
-        case 's':
-          /* setuid */
-          mode_op |= S_ISUID;
-          break;
-        case 't':
-          /* sticky */
-          mode_op |= S_ISVTX;
-          break;
-        case 'o':
-          mode_op |= curmode & S_IRWXO;
-          mode_op |= (curmode & S_IRWXO) << 3;
-          mode_op |= (curmode & S_IRWXO) << 6;
-          break;
-        case 'g':
-          mode_op |= (curmode & S_IRWXG) >> 3;
-          mode_op |= curmode & S_IRWXG;
-          mode_op |= (curmode & S_IRWXG) << 3;
-          break;
-        case 'u':
-          mode_op |= (curmode & S_IRWXO) >> 6;
-          mode_op |= (curmode & S_IRWXO) >> 3;
-          mode_op |= curmode & S_IRWXU;
-          break;
-        case '\0':
-          /* Apply the mode and move on */
-          switch (*how) {
-          case '+':
-          case '=':
-            mode |= (mode_op & ~mask);
+          case 'r':
+            mode_op |= (S_IRUSR|S_IRGRP|S_IROTH);
             break;
-          case '-':
-            mode &= ~(mode_op & ~mask);
-            break;
-          }
 
-          mode_op = 0;
-          if (*who && *(who+1)) {
-            who++;
-            cp = what;
-            continue;
-          } else
-            cp = NULL;
-          break;
-        default:
-          invalid++;
+          case 'w':
+            mode_op |= (S_IWUSR|S_IWGRP|S_IWOTH);
+            break;
+          case 'x':
+            mode_op |= (S_IXUSR|S_IXGRP|S_IXOTH);
+            break;
+
+          /* 'X' not implemented */
+          case 's':
+            /* setuid */
+            mode_op |= S_ISUID;
+            break;
+
+          case 't':
+            /* sticky */
+            mode_op |= S_ISVTX;
+            break;
+
+          case 'o':
+            mode_op |= curmode & S_IRWXO;
+            mode_op |= (curmode & S_IRWXO) << 3;
+            mode_op |= (curmode & S_IRWXO) << 6;
+            break;
+
+          case 'g':
+            mode_op |= (curmode & S_IRWXG) >> 3;
+            mode_op |= curmode & S_IRWXG;
+            mode_op |= (curmode & S_IRWXG) << 3;
+            break;
+
+          case 'u':
+            mode_op |= (curmode & S_IRWXO) >> 6;
+            mode_op |= (curmode & S_IRWXO) >> 3;
+            mode_op |= curmode & S_IRWXU;
+            break;
+
+          case '\0':
+            /* Apply the mode and move on */
+            switch (*how) {
+              case '+':
+              case '=':
+                mode |= (mode_op & ~mask);
+                break;
+
+              case '-':
+                mode &= ~(mode_op & ~mask);
+                break;
+            }
+
+            mode_op = 0;
+            if (*who && *(who+1)) {
+              who++;
+              cp = what;
+              continue;
+
+            } else
+              cp = NULL;
+
+            break;
+
+          default:
+            invalid++;
         }
 
-        if (invalid) break;
-        if (cp) cp++;
+        if (invalid)
+          break;
+
+        if (cp)
+          cp++;
       }
       break;
     }
 
     if (invalid) {
-      pr_response_add_err(R_550, "'%s': invalid mode.", cmd->argv[1]);
-      return ERROR(cmd);
+      pr_response_add_err(R_550, _("'%s': invalid mode"), cmd->argv[1]);
+      return PR_ERROR(cmd);
     }
   }
 
-  if (core_chmod(cmd, dir, mode) == -1) {
-    pr_response_add_err(R_550, "%s: %s", arg, strerror(errno));
-    return ERROR(cmd);
+  res = core_chmod(cmd, dir, mode);
+  if (res < 0) {
+    int xerrno = errno;
 
-  } else
-    pr_response_add(R_200, "SITE %s command successful", cmd->argv[0]);
+    (void) pr_trace_msg("fileperms", 1, "%s, user '%s' (UID %lu, GID %lu): "
+      "error chmod'ing '%s' to %04o: %s", cmd->argv[0], session.user,
+      (unsigned long) session.uid, (unsigned long) session.gid,
+      dir, (unsigned int) mode, strerror(xerrno));
 
-  return HANDLED(cmd);
+    pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(xerrno));
+
+    errno = xerrno;
+    return PR_ERROR(cmd);
+
+  } else {
+    pr_response_add(R_200, _("SITE %s command successful"), cmd->argv[0]);
+  }
+
+  return PR_HANDLED(cmd);
 }
 
 MODRET site_help(cmd_rec *cmd) {
@@ -351,10 +411,10 @@ MODRET site_help(cmd_rec *cmd) {
    */
 
   if (cmd->argc == 1 || (cmd->argc == 2 &&
-      ((!strcasecmp(cmd->argv[0], "SITE") &&
-        !strcasecmp(cmd->argv[1], "HELP")) ||
-       (!strcasecmp(cmd->argv[0], "HELP") &&
-        !strcasecmp(cmd->argv[1], "SITE"))))) {
+      ((strcasecmp(cmd->argv[0], "SITE") == 0 &&
+        strcasecmp(cmd->argv[1], "HELP") == 0) ||
+       (strcasecmp(cmd->argv[0], "HELP") == 0 &&
+        strcasecmp(cmd->argv[1], "SITE") == 0)))) {
 
     for (i = 0; _help[i].cmd; i++) {
       if (_help[i].implemented)
@@ -371,17 +431,17 @@ MODRET site_help(cmd_rec *cmd) {
       *cp = toupper(*cp);
 
     for (i = 0; _help[i].cmd; i++)
-      if (!strcasecmp(cmd->argv[1], _help[i].cmd)) {
-        pr_response_add(R_214, "Syntax: SITE %s %s",
+      if (strcasecmp(cmd->argv[1], _help[i].cmd) == 0) {
+        pr_response_add(R_214, _("Syntax: SITE %s %s"),
           cmd->argv[1], _help[i].syntax);
-        return HANDLED(cmd);
+        return PR_HANDLED(cmd);
       }
 
-    pr_response_add_err(R_502, "Unknown command 'SITE %s'", cmd->arg);
-    return ERROR(cmd);
+    pr_response_add_err(R_502, _("Unknown command 'SITE %s'"), cmd->arg);
+    return PR_ERROR(cmd);
   }
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* The site_commands table is local only, and not registered with our
@@ -399,23 +459,23 @@ modret_t *site_dispatch(cmd_rec *cmd) {
   register unsigned int i = 0;
 
   if (!cmd->argc) {
-    pr_response_add_err(R_500, "'SITE' requires parameters");
-    return ERROR(cmd);
+    pr_response_add_err(R_500, _("'SITE' requires parameters"));
+    return PR_ERROR(cmd);
   }
 
   for (i = 0; site_commands[i].command; i++)
-    if (!strcmp(cmd->argv[0], site_commands[i].command)) {
+    if (strcmp(cmd->argv[0], site_commands[i].command) == 0) {
       if (site_commands[i].requires_auth && cmd_auth_chk &&
           !cmd_auth_chk(cmd)) {
-        pr_response_send(R_530, "Please login with " C_USER " and " C_PASS);
-        return ERROR(cmd);
+        pr_response_send(R_530, _("Please login with USER and PASS"));
+        return PR_ERROR(cmd);
 
       } else
         return site_commands[i].handler(cmd);
     }
 
-  pr_response_add_err(R_500, "'SITE %s' not understood", cmd->argv[0]);
-  return ERROR(cmd);
+  pr_response_add_err(R_500, _("'SITE %s' not understood"), cmd->argv[0]);
+  return PR_ERROR(cmd);
 }
 
 /* Command handlers
@@ -424,24 +484,26 @@ modret_t *site_dispatch(cmd_rec *cmd) {
 MODRET site_pre_cmd(cmd_rec *cmd) {
   if (cmd->argc > 1 && !strcasecmp(cmd->argv[1], "help"))
     pr_response_add(R_214,
-      "The following SITE commands are recognized (* =>'s unimplemented).");
-  return DECLINED(cmd);
+      _("The following SITE commands are recognized (* =>'s unimplemented)"));
+  return PR_DECLINED(cmd);
 }
 
 MODRET site_cmd(cmd_rec *cmd) {
   char *cp = NULL;
   cmd_rec *tmpcmd = NULL;
 
-  /* Make a copy of the cmd structure for passing to call_module */
+  /* Make a copy of the cmd structure for passing to pr_module_call(). */
   tmpcmd = pcalloc(cmd->pool, sizeof(cmd_rec));
   memcpy(tmpcmd, cmd, sizeof(cmd_rec));
 
   tmpcmd->argc--;
   tmpcmd->argv++;
 
-  if (tmpcmd->argc)
-    for (cp = tmpcmd->argv[0]; *cp; cp++)
+  if (tmpcmd->argc) {
+    for (cp = tmpcmd->argv[0]; *cp; cp++) {
       *cp = toupper((int) *cp);
+    }
+  }
 
   tmpcmd->notes = cmd->notes;
 
@@ -449,10 +511,12 @@ MODRET site_cmd(cmd_rec *cmd) {
 }
 
 MODRET site_post_cmd(cmd_rec *cmd) {
-  if (cmd->argc > 1 && !strcasecmp(cmd->argv[1], "help"))
-    pr_response_add(R_214, "Direct comments to %s.",
+  if (cmd->argc > 1 &&
+      strcasecmp(cmd->argv[1], "help") == 0)
+    pr_response_add(R_214, _("Direct comments to %s"),
       (cmd->server->ServerAdmin ? cmd->server->ServerAdmin : "ftp-admin"));
-  return DECLINED(cmd);
+
+  return PR_DECLINED(cmd);
 }
 
 /* Initialization routines
@@ -461,7 +525,7 @@ MODRET site_post_cmd(cmd_rec *cmd) {
 static int site_init(void) {
 
   /* Add the commands handled by this module to the HELP list. */ 
-  pr_help_add(C_SITE, "<sp> string", TRUE);
+  pr_help_add(C_SITE, _("<sp> string"), TRUE);
 
   return 0;
 }
