@@ -1,8 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 1997, 1998 Public Flood Software
- * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2011 The ProFTPD Project team
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -25,13 +23,11 @@
  */
 
 /* ProFTPD Controls command-line client
- *
- * $Id: ftpdctl.c,v 1.1 2003/11/09 01:55:28 castaglia Exp $
+ * $Id: ftpdctl.c,v 1.18 2011/05/23 21:22:24 castaglia Exp $
  */
 
 #include "conf.h"
-
-#include <signal.h>
+#include "privs.h"
 
 #ifdef HAVE_GETOPT_H
 # include <getopt.h>
@@ -48,32 +44,97 @@ void *get_param_ptr(xaset_t *set, const char *name, int recurse) {
   return NULL;
 }
 
-char *pr_inet_validate(char *buf) {
-  return NULL;
-}
-
 void pr_alarms_block(void) {
 }
 
 void pr_alarms_unblock(void) {
 }
 
+gid_t pr_auth_name2gid(pool *p, const char *name) {
+  return (gid_t) -1;
+}
+
+uid_t pr_auth_name2uid(pool *p, const char *name) {
+  return (uid_t) -1;
+}
+
+void pr_event_generate(const char *event, const void *event_data) {
+  (void) event;
+  (void) event_data;
+}
+
+int pr_event_listening(const char *event) {
+  return -1;
+}
+
+int pr_fs_get_usable_fd(int fd) {
+  return -1;
+}
+
+int pr_privs_root(const char *file, int lineno) {
+  return 0;
+}
+
+int pr_privs_relinquish(const char *file, int lineno) {
+  return 0;
+}
+
 void pr_signals_block(void) {
 }
 
 void pr_signals_unblock(void) {
 }
 
-char *sstrcat(char *dest, const char *src, size_t n) {
-  register char *d;
+void pr_signals_handle(void) {
+}
 
-  for (d = dest; *d && n > 1; d++, n--) ;
+pr_table_t *pr_table_alloc(pool *p, int flags) {
+  errno = EPERM;
+  return NULL;
+}
 
-  while(n-- > 1 && *src)
-    *d++ = *src++;
+int pr_table_add(pr_table_t *tab, const char *k, void *v, size_t sz) {
+  errno = EPERM;
+  return -1;
+}
 
-  *d = 0;
-  return dest;
+int pr_table_count(pr_table_t *tab) {
+  errno = EPERM;
+  return -1;
+}
+
+int pr_table_empty(pr_table_t *tab) {
+  errno = EPERM;
+  return -1;
+}
+
+int pr_table_exists(pr_table_t *tab, const char *k) {
+  errno = EPERM;
+  return -1;
+}
+
+int pr_table_free(pr_table_t *tab) {
+  errno = EPERM;
+  return -1;
+}
+
+void *pr_table_get(pr_table_t *tab, const char *k, size_t *sz) {
+  errno = EPERM;
+  return NULL;
+}
+
+int pr_table_set(pr_table_t *tab, const char *k, void *v, size_t sz) {
+  errno = EPERM;
+  return -1;
+}
+
+struct tm *pr_localtime(pool *p, const time_t *t) {
+  return localtime(t);
+}
+
+int pr_trace_msg(const char *channel, int level, const char *fmt, ...) {
+  errno = ENOSYS;
+  return -1;
 }
 
 char *sstrncpy(char *dest, const char *src, size_t n) {
@@ -82,6 +143,9 @@ char *sstrncpy(char *dest, const char *src, size_t n) {
   if (!dest)
     return NULL;
 
+  if (n == 0)
+    return NULL;
+
   if (src && *src) {
     for (; *src && n > 1; n--)
       *d++ = *src++;
@@ -91,7 +155,7 @@ char *sstrncpy(char *dest, const char *src, size_t n) {
   return dest;
 }
 
-#ifdef USE_CTRLS
+#ifdef PR_USE_CTRLS
 
 /* need a SIGPIPE handler */
 static RETSIGTYPE sig_pipe(int sig) {
@@ -113,7 +177,7 @@ int main(int argc, char *argv[]) {
   const char *cmdopts = "hs:v";
 
   register int i = 0;
-  char *socket_file = RUN_DIR "/proftpd.sock";
+  char *socket_file = PR_RUN_DIR "/proftpd.sock";
   int sockfd = -1, optc = 0, status = 0, respargc = 0;
   unsigned int reqargc = 0;
   pool *ctl_pool = NULL;
@@ -179,13 +243,15 @@ int main(int argc, char *argv[]) {
   /* Don't forget to NULL-terminate the array. */
   *((char **) push_array(reqargv)) = NULL;
 
-  /* Open a connection to the socket maintained by mod_ftpdctl. */
+  /* Open a connection to the socket maintained by mod_ctrls. */
   if (verbose)
-    fprintf(stdout, "%s: contacting server\n", program);
+    fprintf(stdout, "%s: contacting server using '%s'\n", program,
+      socket_file);
 
-  if ((sockfd = pr_ctrls_connect(socket_file)) < 0) {
-    fprintf(stderr, "%s: error contacting server: %s\n", program,
-      strerror(errno));
+  sockfd = pr_ctrls_connect(socket_file);
+  if (sockfd < 0) {
+    fprintf(stderr, "%s: error contacting server using '%s': %s\n", program,
+      socket_file, strerror(errno));
     exit(1);
   }
 
@@ -238,4 +304,4 @@ int main(int argc, char *argv[]) {
   return 1;
 }
 
-#endif /* USE_CTRLS */
+#endif /* PR_USE_CTRLS */
