@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001-2008 The ProFTPD Project team
+ * Copyright (c) 2001-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, the ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,11 +23,10 @@
  */
 
 /* NetIO routines
- * $Id: netio.c,v 1.38 2008/11/17 01:42:17 castaglia Exp $
+ * $Id: netio.c,v 1.53 2011/09/21 14:48:41 castaglia Exp $
  */
 
 #include "conf.h"
-#include <signal.h>
 
 /* See RFC 854 for the definition of these Telnet values */
 
@@ -103,19 +102,21 @@ static pr_netio_stream_t *netio_stream_alloc(pool *parent_pool) {
 }
 
 static pr_buffer_t *netio_buffer_alloc(pr_netio_stream_t *nstrm) {
+  size_t bufsz;
   pr_buffer_t *pbuf = NULL;
 
   pbuf = pcalloc(nstrm->strm_pool, sizeof(pr_buffer_t));
 
   /* Allocate a buffer. */
-  pbuf->buf = pcalloc(nstrm->strm_pool, PR_TUNABLE_BUFFER_SIZE);
-  pbuf->buflen = PR_TUNABLE_BUFFER_SIZE;
+  bufsz = pr_config_get_server_xfer_bufsz(nstrm->strm_mode);
+  pbuf->buf = pcalloc(nstrm->strm_pool, bufsz);
+  pbuf->buflen = bufsz;
 
   /* Position the offset at the start of the buffer, and set the
    * remaining bytes value accordingly.
    */
   pbuf->current = pbuf->buf;
-  pbuf->remaining = PR_TUNABLE_BUFFER_SIZE;
+  pbuf->remaining = bufsz;
 
   /* Add this buffer to the given stream. */
   nstrm->strm_buf = pbuf;
@@ -657,6 +658,8 @@ int pr_netio_printf_async(pr_netio_stream_t *nstrm, char *fmt, ...) {
 
 int pr_netio_write(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
   int bwritten = 0, total = 0;
+  pr_buffer_t *pbuf;
+  pool *sub_pool;
 
   /* Sanity check */
   if (!nstrm) {
@@ -669,6 +672,43 @@ int pr_netio_write(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
     return -1;
   }
 
+  /* Before we send out the data to the client, generate an event
+   * for any listeners which may want to examine this data.  To do this, we
+   * need to allocate a pr_buffer_t for sending the buffer data to the
+   * listeners.
+   *
+   * We could just use nstrm->strm_pool, but for a long-lived control
+   * connection, this would amount to a slow memory increase.  So instead,
+   * we create a subpool from the stream's pool, and allocate the
+   * pr_buffer_t out of that.  Then simply destroy the subpool when done.
+   */
+
+  sub_pool = pr_pool_create_sz(nstrm->strm_pool, 64);
+  pbuf = pcalloc(sub_pool, sizeof(pr_buffer_t));
+  pbuf->buf = buf;
+  pbuf->buflen = buflen;
+  pbuf->current = pbuf->buf;
+  pbuf->remaining = 0;
+
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      pr_event_generate("core.ctrl-write", pbuf);
+      break;
+
+    case PR_NETIO_STRM_DATA:
+      pr_event_generate("core.data-write", pbuf);
+      break;
+
+    case PR_NETIO_STRM_OTHR:
+      pr_event_generate("core.othr-write", pbuf);
+      break;
+  }
+
+  /* The event listeners may have changed the data to write out. */
+  buf = pbuf->buf;
+  buflen = pbuf->buflen - pbuf->remaining;
+  destroy_pool(sub_pool);
+
   while (buflen) {
 
     switch (pr_netio_poll(nstrm)) {
@@ -718,12 +758,14 @@ int pr_netio_write(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
     buflen -= bwritten;
   }
 
+  session.total_raw_out += total;
   return total;
 }
 
 int pr_netio_write_async(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
   int flags = 0;
   int bwritten = 0, total = 0;
+  pr_buffer_t *pbuf;
 
   /* Sanity check */
   if (!nstrm) {
@@ -743,6 +785,34 @@ int pr_netio_write_async(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
   if (fcntl(nstrm->strm_fd, F_SETFL, flags|O_NONBLOCK) == -1)
     return -1;
 
+  /* Before we send out the data to the client, generate an event
+   * for any listeners which may want to examine this data.
+   */
+
+  pbuf = pcalloc(nstrm->strm_pool, sizeof(pr_buffer_t));
+  pbuf->buf = buf;
+  pbuf->buflen = buflen;
+  pbuf->current = pbuf->buf;
+  pbuf->remaining = 0;
+
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      pr_event_generate("core.ctrl-write", pbuf);
+      break;
+
+    case PR_NETIO_STRM_DATA:
+      pr_event_generate("core.data-write", pbuf);
+      break;
+
+    case PR_NETIO_STRM_OTHR:
+      pr_event_generate("core.othr-write", pbuf);
+      break;
+  }
+
+  /* The event listeners may have changed the data to write out. */
+  buf = pbuf->buf;
+  buflen = pbuf->buflen - pbuf->remaining;
+
   while (buflen) {
     do {
 
@@ -852,8 +922,24 @@ int pr_netio_read(pr_netio_stream_t *nstrm, char *buf, size_t buflen,
           }
 
 #ifdef EAGAIN
-	  if (bread == -1 && errno == EAGAIN)
+	  if (bread == -1 &&
+              errno == EAGAIN) {
+            int xerrno = EAGAIN;
+
+            /* Treat this as an interrupted call, call pr_signals_handle()
+             * (which will delay for a few msecs because of EINTR), and try
+             * again.
+             *
+             * This should avoid a tightly spinning loop if read(2) returns
+             * EAGAIN, as on a data transfer (Bug#3639).
+             */
+
+            errno = EINTR;
+            pr_signals_handle();
+
+            errno = xerrno;
             goto polling;
+          }
 #endif
 
         } while (bread == -1 && errno == EINTR);
@@ -883,6 +969,7 @@ int pr_netio_read(pr_netio_stream_t *nstrm, char *buf, size_t buflen,
     buflen -= bread;
   }
 
+  session.total_raw_in += total;
   return total;
 }
 
@@ -928,10 +1015,12 @@ char *pr_netio_gets(char *buf, size_t buflen, pr_netio_stream_t *nstrm) {
 
   buflen--;
 
-  if (nstrm->strm_buf)
+  if (nstrm->strm_buf) {
     pbuf = nstrm->strm_buf;
-  else
+
+  } else {
     pbuf = netio_buffer_alloc(nstrm);
+  }
 
   while (buflen) {
 
@@ -954,8 +1043,17 @@ char *pr_netio_gets(char *buf, size_t buflen, pr_netio_stream_t *nstrm) {
       pbuf->remaining = pbuf->buflen - toread;
       pbuf->current = pbuf->buf;
 
-    } else
-      toread = pbuf->buflen - pbuf->remaining;
+      pbuf->remaining = pbuf->buflen - toread;
+      pbuf->current = pbuf->buf;
+
+      /* Before we begin iterating through the data read in from the
+       * network, generate an event for any listeners which may want to
+       * examine this data as well.
+       */
+      pr_event_generate("core.othr-read", pbuf);
+    }
+
+    toread = pbuf->buflen - pbuf->remaining;
 
     while (buflen && *pbuf->current != '\n' && toread--) {
       if (*pbuf->current & 0x80)
@@ -1016,7 +1114,7 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
         pbuf->remaining == pbuf->buflen) {
 
       toread = pr_netio_read(in_nstrm, pbuf->buf,
-        (buflen < pbuf->buflen ?  buflen : pbuf->buflen), 1);
+        (buflen < pbuf->buflen ? buflen : pbuf->buflen), 1);
 
       if (toread <= 0) {
         if (bp != buf) {
@@ -1031,8 +1129,14 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
       pbuf->remaining = pbuf->buflen - toread;
       pbuf->current = pbuf->buf;
 
-    } else
-      toread = pbuf->buflen - pbuf->remaining;
+      /* Before we begin iterating through the data read in from the
+       * network, handing any Telnet characters and such, generate an event
+       * for any listeners which may want to examine this data as well.
+       */
+      pr_event_generate("core.ctrl-read", pbuf);
+    }
+
+    toread = pbuf->buflen - pbuf->remaining;
 
     while (buflen && toread > 0 && *pbuf->current != '\n' && toread--) {
       cp = *pbuf->current++;
@@ -1048,7 +1152,6 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
               case TELNET_DONT:
               case TELNET_IP:
               case TELNET_DM:
-
                 /* Why do we do this crazy thing where we set the "telnet mode"
                  * to be the action, and let the while loop, on the next pass,
                  * handle that action?  It's because we don't know, right now,
@@ -1060,6 +1163,16 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
                 telnet_mode = cp;
                 continue;
 
+              case TELNET_IAC:
+                /* In this case, we know that the previous byte was TELNET_IAC,
+                 * and that the current byte is another TELNET_IAC.  The
+                 * first TELNET_IAC thus "escapes" the second, telling us
+                 * that the current byte (TELNET_IAC) should be written out
+                 * as is (Bug#3697).
+                 */
+                telnet_mode = 0;
+                break;
+
               default:
                 /* In this case, we know that the previous byte was TELNET_IAC,
                  * but the current byte is not a value we care about.  So
@@ -1098,6 +1211,16 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
         }
       }
 
+      /* In the situation where the previous byte was an IAC, we wrote IAC
+       * into the output buffer, and decremented buflen (size of the output
+       * buffer remaining).  Thus we need to check here if buflen is zero,
+       * before trying to decrement buflen again (and possibly underflowing
+       * the buflen size_t data type).
+       */
+      if (buflen == 0) {
+        break;
+      }
+
       *bp++ = cp;
       buflen--;
     }
