@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2004, 2005 The ProFTPD Project team
+ * Copyright (c) 2004-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,7 +23,7 @@
  */
 
 /* Table API implementation
- * $Id: table.c,v 1.9 2006/05/26 16:43:15 castaglia Exp $
+ * $Id: table.c,v 1.24 2011/05/23 21:22:24 castaglia Exp $
  */
 
 #include "conf.h"
@@ -47,7 +47,7 @@ struct table_rec {
   pr_table_entry_t *tab_iter_ent;
 
   /* For iterating through all of the possible multiple values for a single
-   * key.
+   * key.  Only used if the PR_TABLE_FL_MULTI_VALUE flag is set.
    */
   pr_table_entry_t *val_iter_ent;
 
@@ -63,12 +63,40 @@ struct table_rec {
   void (*entremove)(pr_table_entry_t **, pr_table_entry_t *);
 };
 
+static int handling_signal = FALSE;
+
+static const char *trace_channel = "table";
+
 /* Default table callbacks
  */
 
 static int key_cmp(const void *key1, size_t keysz1, const void *key2,
     size_t keysz2) {
-  return strcmp((const char *) key1, (const char *) key2);
+  const char *k1, *k2;
+
+  if (keysz1 != keysz2) {
+    return keysz1 < keysz2 ? -1 : 1;
+  }
+
+  k1 = key1;
+  k2 = key2;
+
+  if (keysz1 >= 1) {
+    /* Basic check of the first character in each key, trying to reduce
+     * the chances of calling strncmp(3) if not needed.
+     */
+
+    if (k1[0] != k2[0]) {
+      return k1[0] < k2[0] ? -1 : 1;
+    }
+
+    /* Special case (unlikely, but possible). */
+    if (keysz1 == 1) {
+      return 0;
+    }
+  }
+
+  return strncmp((const char *) key1, (const char *) key2, keysz1);
 }
 
 /* Use Perl's hashing algorithm by default. */
@@ -79,10 +107,11 @@ static unsigned int key_hash(const void *key, size_t keysz) {
   while (sz--) {
     const char *k = key;
     unsigned int c = *k;
-    k++;
 
-    /* Always handle signals in potentially long-running while loops. */
-    pr_signals_handle();
+    if (!handling_signal) {
+      /* Always handle signals in potentially long-running while loops. */
+      pr_signals_handle();
+    }
 
     i = (i * 33) + c;
   }
@@ -96,7 +125,10 @@ static unsigned int key_hash(const void *key, size_t keysz) {
 static void entry_insert(pr_table_entry_t **h, pr_table_entry_t *e) {
   pr_table_entry_t *ei;
 
-  for (ei = *h; ei && ei->next; ei = ei->next);
+  if (*h == NULL)
+    return;
+
+  for (ei = *h; ei != NULL && ei->next; ei = ei->next);
 
   /* Now, ei points to the last entry in the chain. */
   ei->next = e;
@@ -106,19 +138,23 @@ static void entry_insert(pr_table_entry_t **h, pr_table_entry_t *e) {
 /* Default removal is simply to remove the entry from the chain. */
 static void entry_remove(pr_table_entry_t **h, pr_table_entry_t *e) {
 
-  if (e->next)
+  if (e->next) {
     e->next->prev = e->prev;
+  }
 
-  if (e->prev)
+  if (e->prev) {
     e->prev->next = e->next;
+  }
 
-  if (e == *h &&
-      e->next == NULL)
-    /* This entry is the head, and is the only entry in this chain. */
-    *h = NULL;
+  if (e == *h) {
+    if (e->next == NULL) {
+      /* This entry is the head, and is the only entry in this chain. */
+      *h = NULL;
 
-  else 
-    *h = e->next;
+    } else {
+      *h = e->next;
+    }
+  }
 
   e->prev = e->next = NULL;
   return;
@@ -155,7 +191,10 @@ static void tab_key_free(pr_table_t *tab, pr_table_key_t *k) {
 
     /* Scan to the end of the list. */
     while (i->next != NULL) {
-      pr_signals_handle();
+      if (!handling_signal) {
+        pr_signals_handle();
+      }
+
       i = i->next;
     }
 
@@ -196,7 +235,10 @@ static void tab_entry_free(pr_table_t *tab, pr_table_entry_t *e) {
 
     /* Scan to the end of the list. */
     while (i->next != NULL) {
-      pr_signals_handle();
+      if (!handling_signal) {
+        pr_signals_handle();
+      }
+
       i = i->next;
     }
 
@@ -264,7 +306,9 @@ static pr_table_entry_t *tab_entry_next(pr_table_t *tab) {
 }
 
 static void tab_entry_remove(pr_table_t *tab, pr_table_entry_t *e) {
-  pr_table_entry_t *h = tab->chains[e->idx];
+  pr_table_entry_t *h;
+
+  h = tab->chains[e->idx];
 
   tab->entremove(&h, e);
   tab->chains[e->idx] = h;
@@ -286,7 +330,9 @@ int pr_table_kadd(pr_table_t *tab, const void *key_data, size_t key_datasz,
   unsigned int h, idx;
   pr_table_entry_t *e, *n;
 
-  if (!tab || !key_data) {
+  if (tab == NULL ||
+      key_data == NULL ||
+      (value_datasz > 0 && value_data == NULL)) {
     errno = EINVAL;
     return -1;
   }
@@ -323,7 +369,8 @@ int pr_table_kadd(pr_table_t *tab, const void *key_data, size_t key_datasz,
        * is identical.  If so, we have multiple values for the same key.
        */
 
-      if (tab->keycmp(ei->key->key_data, 0, key_data, 0) == 0) {
+      if (tab->keycmp(ei->key->key_data, ei->key->key_datasz,
+          key_data, key_datasz) == 0) {
 
         /* Check if this table allows multivalues. */
         if (!(tab->flags & PR_TABLE_FL_MULTI_VALUE)) {
@@ -394,7 +441,8 @@ int pr_table_kexists(pr_table_t *tab, const void *key_data, size_t key_datasz) {
   }
 
   for (ent = head; ent; ent = ent->next) {
-    if (ent->key->hash != h)
+    if (ent->key == NULL ||
+        ent->key->hash != h)
       continue;
 
     /* Matching hashes.  Now to see if the keys themselves match. */
@@ -456,11 +504,10 @@ void *pr_table_kget(pr_table_t *tab, const void *key_data, size_t key_datasz,
              tab->cache_ent->key->key_data == key_data) {
 
      /* If the cached lookup entry matches, we'll use it. */
-     head = tab->cache_ent;
+     head = tab->cache_ent->next;
 
   } else {
     unsigned int idx = h % tab->nchains;
-
     head = tab->chains[idx];
   }
 
@@ -473,7 +520,8 @@ void *pr_table_kget(pr_table_t *tab, const void *key_data, size_t key_datasz,
   }
 
   for (ent = head; ent; ent = ent->next) {
-    if (ent->key->hash != h)
+    if (ent->key == NULL ||
+        ent->key->hash != h)
       continue;
 
     /* Matching hashes.  Now to see if the keys themselves match. */
@@ -483,7 +531,8 @@ void *pr_table_kget(pr_table_t *tab, const void *key_data, size_t key_datasz,
       if (tab->flags & PR_TABLE_FL_USE_CACHE) 
         tab->cache_ent = ent;
 
-      tab->val_iter_ent = ent;
+      if (tab->flags & PR_TABLE_FL_MULTI_VALUE)
+        tab->val_iter_ent = ent;
 
       if (value_datasz)
         *value_datasz = ent->value_datasz;
@@ -546,7 +595,8 @@ void *pr_table_kremove(pr_table_t *tab, const void *key_data,
   }
 
   for (ent = head; ent; ent = ent->next) {
-    if (ent->key->hash != h)
+    if (ent->key == NULL ||
+        ent->key->hash != h)
       continue;
 
     /* Matching hashes.  Now to see if the keys themselves match. */
@@ -578,7 +628,9 @@ int pr_table_kset(pr_table_t *tab, const void *key_data, size_t key_datasz,
 
   /* XXX Should callers be allowed to set NULL values for keys? */
 
-  if (!tab || !key_data || !value_data) {
+  if (tab == NULL ||
+      key_data == NULL ||
+      (value_datasz > 0 && value_data == NULL)) {
     errno = EINVAL;
     return -1;
   }
@@ -620,7 +672,8 @@ int pr_table_kset(pr_table_t *tab, const void *key_data, size_t key_datasz,
   }
 
   for (ent = head; ent; ent = ent->next) {
-    if (ent->key->hash != h)
+    if (ent->key == NULL ||
+        ent->key->hash != h)
       continue;
 
     /* Matching hashes.  Now to see if the keys themselves match. */
@@ -638,7 +691,8 @@ int pr_table_kset(pr_table_t *tab, const void *key_data, size_t key_datasz,
       if (tab->flags & PR_TABLE_FL_USE_CACHE)
         tab->cache_ent = ent;
 
-      tab->val_iter_ent = ent;
+      if (tab->flags & PR_TABLE_FL_MULTI_VALUE)
+        tab->val_iter_ent = ent;
 
       return 0;
     }
@@ -654,12 +708,14 @@ int pr_table_kset(pr_table_t *tab, const void *key_data, size_t key_datasz,
 int pr_table_add(pr_table_t *tab, const char *key_data, void *value_data,
     size_t value_datasz) {
 
-  if (!tab || !key_data) {
+  if (tab == NULL ||
+      key_data == NULL) {
     errno = EINVAL;
     return -1;
   }
 
-  if (value_data && value_datasz == 0)
+  if (value_data &&
+      value_datasz == 0)
     value_datasz = strlen((char *) value_data) + 1;
 
   return pr_table_kadd(tab, key_data, strlen(key_data) + 1, value_data,
@@ -680,6 +736,9 @@ int pr_table_add_dup(pr_table_t *tab, const char *key_data, void *value_data,
     return -1;
   }
 
+  if (value_data && value_datasz == 0)
+    value_datasz = strlen((char *) value_data) + 1;
+
   dup_data = pcalloc(tab->pool, value_datasz);
   memcpy(dup_data, value_data, value_datasz);
  
@@ -690,7 +749,8 @@ pr_table_t *pr_table_nalloc(pool *p, int flags, unsigned int nchains) {
   pr_table_t *tab;
   pool *tab_pool;
 
-  if (!p) {
+  if (p == NULL ||
+      nchains == 0) {
     errno = EINVAL;
     return NULL;
   }
@@ -744,8 +804,10 @@ int pr_table_do(pr_table_t *tab, int (*cb)(const void *key_data,
 
     while (ent) {
       int res;
- 
-      pr_signals_handle();
+
+      if (!handling_signal) { 
+        pr_signals_handle();
+      }
 
       res = cb(ent->key->key_data, ent->key->key_datasz, ent->value_data,
         ent->value_datasz, user_data);
@@ -777,7 +839,9 @@ int pr_table_empty(pr_table_t *tab) {
     pr_table_entry_t *e = tab->chains[i];
 
     while (e) {
-      pr_signals_handle();
+      if (!handling_signal) {
+        pr_signals_handle();
+      }
 
       tab_entry_remove(tab, e);
       tab_entry_free(tab, e);
@@ -843,7 +907,9 @@ void *pr_table_next(pr_table_t *tab) {
 
   ent = tab_entry_next(tab);
   while (ent) {
-    pr_signals_handle();
+    if (!handling_signal) {
+      pr_signals_handle();
+    }
 
     if (prev &&
         ent->key == prev->key) {
@@ -891,6 +957,10 @@ int pr_table_set(pr_table_t *tab, const char *key_data, void *value_data,
     return -1;
   }
 
+  if (value_data &&
+      value_datasz == 0)
+    value_datasz = strlen((char *) value_data) + 1;
+
   return pr_table_kset(tab, key_data, strlen(key_data) + 1, value_data,
     value_datasz);
 }
@@ -942,8 +1012,21 @@ int pr_table_ctl(pr_table_t *tab, int cmd, void *arg) {
         (void (*)(pr_table_entry_t **, pr_table_entry_t *)) entry_remove;
       return 0;
 
-    case PR_TABLE_CTL_SET_NCHAINS:
-      tab->nchains = *((unsigned int *) arg);
+    case PR_TABLE_CTL_SET_NCHAINS: {
+      unsigned int new_nchains;
+
+      if (arg == NULL) {
+        errno = EINVAL;
+        return -1;
+      }
+
+      new_nchains = *((unsigned int *) arg);
+      if (new_nchains == 0) {
+        errno = EINVAL;
+        return -1;
+      }
+
+      tab->nchains = new_nchains;
       
       /* Note: by not freeing the memory of the previously allocated
        * chains, this constitutes a minor leak of the table's memory pool.
@@ -951,13 +1034,12 @@ int pr_table_ctl(pr_table_t *tab, int cmd, void *arg) {
       tab->chains = pcalloc(tab->pool,
         sizeof(pr_table_entry_t *) * tab->nchains);
       return 0;
+    }
 
     default:
       errno = EINVAL;
-      return -1;
   }
 
-  errno = EACCES;
   return -1;
 }
 
@@ -971,16 +1053,34 @@ void *pr_table_pcalloc(pr_table_t *tab, size_t sz) {
   return pcalloc(tab->pool, sz);
 }
 
+static void table_printf(const char *fmt, ...) {
+  char buf[PR_TUNABLE_BUFFER_SIZE+1];
+  va_list msg;
+
+  memset(buf, '\0', sizeof(buf));
+  va_start(msg, fmt);
+  vsnprintf(buf, sizeof(buf)-1, fmt, msg);
+  va_end(msg);
+
+  buf[sizeof(buf)-1] = '\0';
+  pr_trace_msg(trace_channel, 19, "dump: %s", buf);
+}
+
 void pr_table_dump(void (*dumpf)(const char *fmt, ...), pr_table_t *tab) {
   register unsigned int i;
 
-  if (!tab)
+  if (tab == NULL) {
     return;
+  }
 
-  if (tab->flags == 0)
+  if (dumpf == NULL) {
+    dumpf = table_printf;
+  }
+
+  if (tab->flags == 0) {
     dumpf("%s", "[table flags]: None");
 
-  else {
+  } else {
     if ((tab->flags & PR_TABLE_FL_MULTI_VALUE) &&
         (tab->flags & PR_TABLE_FL_USE_CACHE)) {
       dumpf("%s", "[table flags]: MultiValue, UseCache");
@@ -1006,9 +1106,12 @@ void pr_table_dump(void (*dumpf)(const char *fmt, ...), pr_table_t *tab) {
     pr_table_entry_t *ent = tab->chains[i];
 
     while (ent) {
-      pr_signals_handle();
+      if (!handling_signal) {
+        pr_signals_handle();
+      }
 
-      dumpf("[chain %u#%u] '%s' => '%s' (%u)", i, j++, ent->key->key_data,
+      dumpf("[hash %u (%u chains) chain %u#%u] '%s' => '%s' (%u)",
+        ent->key->hash, tab->nchains, i, j++, ent->key->key_data,
         ent->value_data, ent->value_datasz);
       ent = ent->next;
     }
@@ -1016,3 +1119,14 @@ void pr_table_dump(void (*dumpf)(const char *fmt, ...), pr_table_t *tab) {
 
   return;
 }
+
+int table_handling_signal(int bool) {
+  if (bool == TRUE ||
+      bool == FALSE) {
+    handling_signal = bool;
+    return 0;
+  }
+
+  errno = EINVAL;
+  return -1;
+}
