@@ -1,10 +1,9 @@
 package ProFTPD::Tests::Commands::SIZE;
 
 use lib qw(t/lib);
-use base qw(Test::Unit::TestCase ProFTPD::TestSuite::Child);
+use base qw(ProFTPD::TestSuite::Child);
 use strict;
 
-use File::Path qw(mkpath rmtree);
 use File::Spec;
 use IO::Handle;
 
@@ -46,29 +45,6 @@ sub list_tests {
   return testsuite_get_runnable_tests($TESTS);
 }
 
-sub set_up {
-  my $self = shift;
-  $self->{tmpdir} = testsuite_get_tmp_dir();
-
-  # Create temporary scratch dir
-  eval { mkpath($self->{tmpdir}) };
-  if ($@) {
-    my $abs_path = File::Spec->rel2abs($self->{tmpdir});
-    die("Can't create dir $abs_path: $@");
-  }
-}
-
-sub tear_down {
-  my $self = shift;
-
-  # Remove temporary scratch dir
-  if ($self->{tmpdir}) {
-    eval { rmtree($self->{tmpdir}) };
-  }
-
-  undef $self;
-};
-
 sub size_ok {
   my $self = shift;
   my $tmpdir = $self->{tmpdir};
@@ -160,7 +136,6 @@ sub size_ok {
         test_msg("Expected $expected, got $resp_code"));
 
       $expected = 14;
-      chomp($resp_msg);
       $self->assert($expected == $resp_msg,
         test_msg("Expected $expected, got $resp_msg"));
     };
@@ -285,7 +260,6 @@ sub size_fails_ascii {
         test_msg("Expected $expected, got $resp_code"));
 
       $expected = "SIZE not allowed in ASCII mode";
-      chomp($resp_msg);
       $self->assert($expected eq $resp_msg,
         test_msg("Expected '$expected', got '$resp_msg'"));
     };
@@ -409,7 +383,6 @@ sub size_fails_enoent {
         test_msg("Expected $expected, got $resp_code"));
 
       $expected = "$test_file: No such file or directory";
-      chomp($resp_msg);
       $self->assert($expected eq $resp_msg,
         test_msg("Expected '$expected', got '$resp_msg'"));
     };
@@ -547,7 +520,6 @@ sub size_fails_eperm {
         test_msg("Expected $expected, got $resp_code"));
 
       $expected = "$test_file: Permission denied";
-      chomp($resp_msg);
       $self->assert($expected eq $resp_msg,
         test_msg("Expected '$expected', got '$resp_msg'"));
     };
