@@ -1,10 +1,11 @@
 package ProFTPD::Tests::Config::Directory::Limits;
 
 use lib qw(t/lib);
-use base qw(Test::Unit::TestCase ProFTPD::TestSuite::Child);
+use base qw(ProFTPD::TestSuite::Child);
 use strict;
 
-use File::Path qw(mkpath rmtree);
+use Cwd;
+use File::Path qw(mkpath);
 use File::Spec;
 use IO::Handle;
 
@@ -62,6 +63,22 @@ my $TESTS = {
     order => ++$order,
     test_class => [qw(forking)],
   },
+
+  limits_one_char_dir_bug3337 => {
+    order => ++$order,
+    test_class => [qw(bug forking)],
+  },
+
+  limits_symlink_dir_bug3166 => {
+    order => ++$order,
+    test_class => [qw(bug forking rootprivs)],
+  },
+
+  limits_anon_dir_abs_path_bug3283 => {
+    order => ++$order,
+    test_class => [qw(bug forking rootprivs)],
+  },
+
 };
 
 sub new {
@@ -72,29 +89,6 @@ sub list_tests {
   return testsuite_get_runnable_tests($TESTS);
 }
 
-sub set_up {
-  my $self = shift;
-  $self->{tmpdir} = testsuite_get_tmp_dir();
-
-  # Create temporary scratch dir
-  eval { mkpath($self->{tmpdir}) };
-  if ($@) {
-    my $abs_path = File::Spec->rel2abs($self->{tmpdir});
-    die("Can't create dir $abs_path: $@");
-  }
-}
-
-sub tear_down {
-  my $self = shift;
-
-  # Remove temporary scratch dir
-  if ($self->{tmpdir}) {
-    eval { rmtree($self->{tmpdir}) };
-  }
-
-  undef $self;
-};
-
 sub limits_with_glob_then_nonglob_dirs_for_same_path {
   my $self = shift;
   my $tmpdir = $self->{tmpdir};
@@ -368,7 +362,7 @@ EOD
       $self->assert($expected == $resp_code,
         test_msg("Expected $expected, got $resp_code"));
 
-      $expected = "test.txt: No such file or directory";
+      $expected = "test.txt: Operation not permitted";
       $self->assert($expected eq $resp_msg,
         test_msg("Expected '$expected', got '$resp_msg'"));
     };
@@ -531,7 +525,7 @@ sub limits_with_glob_denied_delete_bug3146 {
       $self->assert($expected == $resp_code,
         test_msg("Expected $expected, got $resp_code"));
 
-      $expected = "test.txt: No such file or directory",
+      $expected = "test.txt: Operation not permitted";
       $self->assert($expected eq $resp_msg,
         test_msg("Expected '$expected', got '$resp_msg'"));
     };
@@ -695,7 +689,7 @@ sub limits_without_glob_denied_delete_bug3146 {
       $self->assert($expected == $resp_code,
         test_msg("Expected $expected, got $resp_code"));
 
-      $expected = "test.txt: No such file or directory",
+      $expected = "test.txt: Operation not permitted";
       $self->assert($expected eq $resp_msg,
         test_msg("Expected '$expected', got '$resp_msg'"));
     };
@@ -1486,7 +1480,7 @@ sub limits_rename_dir_failed_rnfr_denied {
       $self->assert($expected == $resp_code,
         test_msg("Expected $expected, got $resp_code"));
 
-      $expected = "upload: No such file or directory";
+      $expected = "upload: Operation not permitted";
       $self->assert($expected eq $resp_msg,
         test_msg("Expected '$expected', got '$resp_msg'"));
     };
@@ -1520,4 +1514,530 @@ sub limits_rename_dir_failed_rnfr_denied {
   unlink($log_file);
 }
 
+sub limits_one_char_dir_bug3337 {
+  my $self = shift;
+  my $tmpdir = $self->{tmpdir};
+
+  my $config_file = "$tmpdir/dir.conf";
+  my $pid_file = File::Spec->rel2abs("$tmpdir/dir.pid");
+  my $scoreboard_file = File::Spec->rel2abs("$tmpdir/dir.scoreboard");
+
+  my $log_file = File::Spec->rel2abs('tests.log');
+
+  my $auth_user_file = File::Spec->rel2abs("$tmpdir/dir.passwd");
+  my $auth_group_file = File::Spec->rel2abs("$tmpdir/dir.group");
+
+  my $user = 'proftpd';
+  my $passwd = 'test';
+  my $home_dir = File::Spec->rel2abs($tmpdir);
+  my $uid = 500;
+  my $gid = 500;
+
+  my $a_dir = File::Spec->rel2abs("$home_dir/a");
+  mkpath($a_dir);
+
+  my $ab_dir = File::Spec->rel2abs("$home_dir/ab");
+  mkpath($ab_dir);
+
+  # Make sure that, if we're running as root, that the home directory has
+  # permissions/privs set for the account we create
+  if ($< == 0) {
+    unless (chmod(0755, $home_dir, $a_dir, $ab_dir)) {
+      die("Can't set perms on $home_dir to 0755: $!");
+    }
+
+    unless (chown($uid, $gid, $home_dir, $a_dir, $ab_dir)) {
+      die("Can't set owner of $home_dir to $uid/$gid: $!");
+    }
+  }
+ 
+  auth_user_write($auth_user_file, $user, $passwd, $uid, $gid, $home_dir,
+    '/bin/bash');
+  auth_group_write($auth_group_file, 'ftpd', $gid, $user);
+
+  # Reported in
+  #
+  #  http://forums.proftpd.org/smf/index.php/topic,4170.0.html
+
+  my $config = {
+    PidFile => $pid_file,
+    ScoreboardFile => $scoreboard_file,
+    SystemLog => $log_file,
+    TraceLog => $log_file,
+    Trace => 'DEFAULT:10 fsio:0 directory:10 lock:0',
+
+    AuthUserFile => $auth_user_file,
+    AuthGroupFile => $auth_group_file,
+    DefaultChdir => '~',
+
+    Directory => {
+      "$home_dir" => {
+        Limit => {
+          WRITE => {
+            DenyUser => $user,
+          },
+        },
+      },
+
+      "$a_dir" => {
+        Limit => {
+          WRITE => {
+            AllowUser => $user,
+          },
+        },
+      },
+
+      "$ab_dir" => {
+        Limit => {
+          WRITE => {
+            AllowUser => $user,
+          },
+        },
+      },
+    },
+
+    IfModules => {
+      'mod_delay.c' => {
+        DelayEngine => 'off',
+      },
+    },
+  };
+
+  my ($port, $config_user, $config_group) = config_write($config_file, $config);
+
+  # Open pipes, for use between the parent and child processes.  Specifically,
+  # the child will indicate when it's done with its test by writing a message
+  # to the parent.
+  my ($rfh, $wfh);
+  unless (pipe($rfh, $wfh)) {
+    die("Can't open pipe: $!");
+  }
+
+  my $ex;
+
+  # Fork child
+  $self->handle_sigchld();
+  defined(my $pid = fork()) or die("Can't fork: $!");
+  if ($pid) {
+    eval {
+      my $client = ProFTPD::TestSuite::FTP->new('127.0.0.1', $port);
+      $client->login($user, $passwd);
+
+      $client->cwd('ab');
+      my $conn = $client->stor_raw('test.txt');
+      unless ($conn) {
+        die("STOR test.txt failed: " . $client->response_code() . " " .
+          $client->response_msg());
+      }
+
+      $conn->close();
+
+      my $resp_code = $client->response_code();
+      my $resp_msg = $client->response_msg();
+
+      my $expected;
+
+      $expected = 226;
+      $self->assert($expected == $resp_code,
+        test_msg("Expected $expected, got $resp_code"));
+
+      $expected = "Transfer complete";
+      $self->assert($expected eq $resp_msg,
+        test_msg("Expected '$expected', got '$resp_msg'"));
+
+      $client->cwd('../a');
+      $conn = $client->stor_raw('test.txt');
+      unless ($conn) {
+        die("STOR test.txt failed: " . $client->response_code() . " " .
+          $client->response_msg());
+      }
+
+      $conn->close();
+
+      $resp_code = $client->response_code();
+      $resp_msg = $client->response_msg();
+
+      $expected = 226;
+      $self->assert($expected == $resp_code,
+        test_msg("Expected $expected, got $resp_code"));
+
+      $expected = "Transfer complete";
+      $self->assert($expected eq $resp_msg,
+        test_msg("Expected '$expected', got '$resp_msg'"));
+    };
+
+    if ($@) {
+      $ex = $@;
+    }
+
+    $wfh->print("done\n");
+    $wfh->flush();
+
+  } else {
+    eval { server_wait($config_file, $rfh) };
+    if ($@) {
+      warn($@);
+      exit 1;
+    }
+
+    exit 0;
+  }
+
+  # Stop server
+  server_stop($pid_file);
+
+  $self->assert_child_ok($pid);
+
+  if ($ex) {
+    die($ex);
+  }
+
+  unlink($log_file);
+}
+
+sub limits_symlink_dir_bug3166 {
+  my $self = shift;
+  my $tmpdir = $self->{tmpdir};
+
+  my $config_file = "$tmpdir/dir.conf";
+  my $pid_file = File::Spec->rel2abs("$tmpdir/dir.pid");
+  my $scoreboard_file = File::Spec->rel2abs("$tmpdir/dir.scoreboard");
+
+  my $log_file = File::Spec->rel2abs('tests.log');
+
+  my $auth_user_file = File::Spec->rel2abs("$tmpdir/dir.passwd");
+  my $auth_group_file = File::Spec->rel2abs("$tmpdir/dir.group");
+
+  my $user = 'proftpd';
+  my $passwd = 'test';
+  my $home_dir = File::Spec->rel2abs($tmpdir);
+  my $uid = 500;
+  my $gid = 500;
+
+  my $src_dir = File::Spec->rel2abs("$tmpdir/dispatch/$user");
+  mkpath($src_dir);
+
+  my $dst_dir = File::Spec->rel2abs("$tmpdir/writable/dir1");
+  mkpath($dst_dir);
+
+  # Change to the $src_dir in order to create a symlink with a relative
+  # path (so that it will work in the chroot).
+
+  my $cwd = getcwd();
+  unless (chdir($src_dir)) {
+    die("Can't chdir to $src_dir: $!");
+  }
+
+  unless (symlink("../../writable/dir1", "dir1")) {
+    die("Can't symlink '../../writable/dir1' to 'dir1': $!");
+  }
+
+  unless (chdir($cwd)) {
+    die("Can't chdir to $cwd: $!");
+  }
+
+  # Make sure that, if we're running as root, that the home directory has
+  # permissions/privs set for the account we create
+  if ($< == 0) {
+    unless (chmod(0755, $home_dir, $src_dir, $dst_dir)) {
+      die("Can't set perms on $home_dir to 0755: $!");
+    }
+
+    unless (chown($uid, $gid, $home_dir, $src_dir, $dst_dir)) {
+      die("Can't set owner of $home_dir to $uid/$gid: $!");
+    }
+  }
+ 
+  auth_user_write($auth_user_file, $user, $passwd, $uid, $gid, $home_dir,
+    '/bin/bash');
+  auth_group_write($auth_group_file, 'ftpd', $gid, $user);
+
+  my $config = {
+    PidFile => $pid_file,
+    ScoreboardFile => $scoreboard_file,
+    SystemLog => $log_file,
+    TraceLog => $log_file,
+    Trace => 'DEFAULT:10 fsio:0 directory:10 lock:0',
+
+    AuthUserFile => $auth_user_file,
+    AuthGroupFile => $auth_group_file,
+    DefaultRoot => '~',
+
+    IfModules => {
+      'mod_delay.c' => {
+        DelayEngine => 'off',
+      },
+    },
+  };
+
+  my ($port, $config_user, $config_group) = config_write($config_file, $config);
+
+  if (open(my $fh, ">> $config_file")) {
+    print $fh <<EOD;
+<Directory $src_dir>
+  <Limit ALL>
+    Order allow, deny
+    AllowUser $user
+    DenyAll
+  </Limit>
+</Directory>
+
+<Directory $home_dir/writable>
+  <Limit ALL>
+    Order allow, deny
+    AllowUser admin
+    DenyAll
+  </Limit>
+
+  <Limit DIRS PORT PASV EPRT EPSV>
+    Order allow, deny
+    AllowUser admin
+    AllowUser test
+    DenyAll
+  </Limit>
+</Directory>
+
+<Directory $dst_dir>
+
+  # Bug#3166 was caused by using just ALL in this limit; the parent
+  # directory's limit included the DIRS command group, which took precedence
+  # over ALL.  By adding DIRS to the limit (which includes the test user),
+  # the desired behavior is achieved.
+
+  <Limit ALL DIRS>
+    Order allow, deny
+    AllowUser admin
+    AllowUser $user
+    DenyAll
+  </Limit>
+</Directory>
+EOD
+    close($fh);
+
+  } else {
+    die("Can't read $config_file: $!");
+  }
+
+  # Open pipes, for use between the parent and child processes.  Specifically,
+  # the child will indicate when it's done with its test by writing a message
+  # to the parent.
+  my ($rfh, $wfh);
+  unless (pipe($rfh, $wfh)) {
+    die("Can't open pipe: $!");
+  }
+
+  my $ex;
+
+  # Fork child
+  $self->handle_sigchld();
+  defined(my $pid = fork()) or die("Can't fork: $!");
+  if ($pid) {
+    eval {
+      my $client = ProFTPD::TestSuite::FTP->new('127.0.0.1', $port);
+      $client->login($user, $passwd);
+
+      $client->cwd('dispatch');
+      $client->nlst();
+
+      $client->cwd($user);
+      $client->nlst();
+
+      $client->cwd('dir1');
+
+      my ($resp_code, $resp_msg) = $client->pwd();
+
+      my $expected;
+
+      $expected = 257;
+      $self->assert($expected == $resp_code,
+        test_msg("Expected $expected, got $resp_code"));
+
+      $expected = '"/writable/dir1" is the current directory';
+      $self->assert($expected eq $resp_msg,
+        test_msg("Expected '$expected', got '$resp_msg'"));
+    };
+
+    if ($@) {
+      $ex = $@;
+    }
+
+    $wfh->print("done\n");
+    $wfh->flush();
+
+  } else {
+    eval { server_wait($config_file, $rfh) };
+    if ($@) {
+      warn($@);
+      exit 1;
+    }
+
+    exit 0;
+  }
+
+  # Stop server
+  server_stop($pid_file);
+
+  $self->assert_child_ok($pid);
+
+  if ($ex) {
+    die($ex);
+  }
+
+  unlink($log_file);
+}
+
+sub limits_anon_dir_abs_path_bug3283 {
+  my $self = shift;
+  my $tmpdir = $self->{tmpdir};
+
+  my $config_file = "$tmpdir/dir.conf";
+  my $pid_file = File::Spec->rel2abs("$tmpdir/dir.pid");
+  my $scoreboard_file = File::Spec->rel2abs("$tmpdir/dir.scoreboard");
+
+  my $log_file = File::Spec->rel2abs('tests.log');
+
+  my $auth_user_file = File::Spec->rel2abs("$tmpdir/dir.passwd");
+  my $auth_group_file = File::Spec->rel2abs("$tmpdir/dir.group");
+
+  my ($user, $group) = config_get_identity();
+
+  my $passwd = 'test';
+  my $anon_dir = File::Spec->rel2abs($tmpdir);
+  my $uid = 500;
+  my $gid = 500;
+
+  my $hide_dir = File::Spec->rel2abs("$tmpdir/hide");
+  mkpath($hide_dir);
+
+  # Make sure that, if we're running as root, that the home directory has
+  # permissions/privs set for the account we create
+  if ($< == 0) {
+    unless (chmod(0755, $anon_dir, $hide_dir)) {
+      die("Can't set perms on $anon_dir to 0755: $!");
+    }
+
+    unless (chown($uid, $gid, $anon_dir, $hide_dir)) {
+      die("Can't set owner of $anon_dir to $uid/$gid: $!");
+    }
+  }
+ 
+  auth_user_write($auth_user_file, $user, $passwd, $uid, $gid, '/tmp',
+    '/bin/bash');
+  auth_group_write($auth_group_file, 'ftpd', $gid, $user);
+
+  my $config = {
+    PidFile => $pid_file,
+    ScoreboardFile => $scoreboard_file,
+    SystemLog => $log_file,
+    TraceLog => $log_file,
+    Trace => 'DEFAULT:10 fsio:0 directory:10 lock:0',
+
+    AuthUserFile => $auth_user_file,
+    AuthGroupFile => $auth_group_file,
+
+    IfModules => {
+      'mod_delay.c' => {
+        DelayEngine => 'off',
+      },
+    },
+  };
+
+  my ($port, $config_user, $config_group) = config_write($config_file, $config);
+
+  if (open(my $fh, ">> $config_file")) {
+    print $fh <<EOD;
+<Anonymous $anon_dir>
+  User $user
+  Group $group
+  UserAlias anonymous $user
+  RequireValidShell off
+
+  <Directory $hide_dir>
+    <Limit ALL>
+      DenyAll
+    </Limit>
+  </Directory>
+</Anonymous>
+EOD
+    close($fh);
+
+  } else {
+    die("Can't read $config_file: $!");
+  }
+
+  # Open pipes, for use between the parent and child processes.  Specifically,
+  # the child will indicate when it's done with its test by writing a message
+  # to the parent.
+  my ($rfh, $wfh);
+  unless (pipe($rfh, $wfh)) {
+    die("Can't open pipe: $!");
+  }
+
+  my $ex;
+
+  # Fork child
+  $self->handle_sigchld();
+  defined(my $pid = fork()) or die("Can't fork: $!");
+  if ($pid) {
+    eval {
+      my $client = ProFTPD::TestSuite::FTP->new('127.0.0.1', $port);
+      $client->login($user, $passwd);
+
+      my $conn = $client->list_raw();
+      unless ($conn) {
+        die("LIST failed: " . $client->response_code() . " " .
+          $client->response_msg());
+      }
+
+      my $buf;
+      $conn->read($buf, 32768, 30);
+      $conn->close();
+
+      my $lines = [split(/\n/, $buf)];
+
+      foreach my $line (@$lines) {
+        if ($line =~ /\s+(\S+)$/) {
+          my $dir_name = $1;
+
+          # The <Directory> with <Limit ALL>DenyAll should not appear in
+          # the returned list
+
+          my $unexpected = 'hide';
+          $self->assert($unexpected ne $dir_name,
+            test_msg("Unexpectedly saw '$dir_name' in LIST"));
+        }
+      }
+
+      $client->quit();
+    };
+
+    if ($@) {
+      $ex = $@;
+    }
+
+    $wfh->print("done\n");
+    $wfh->flush();
+
+  } else {
+    eval { server_wait($config_file, $rfh) };
+    if ($@) {
+      warn($@);
+      exit 1;
+    }
+
+    exit 0;
+  }
+
+  # Stop server
+  server_stop($pid_file);
+
+  $self->assert_child_ok($pid);
+
+  if ($ex) {
+    die($ex);
+  }
+
+  unlink($log_file);
+}
+
 1;
