@@ -1,10 +1,9 @@
 package ProFTPD::Tests::Logins;
 
 use lib qw(t/lib);
-use base qw(Test::Unit::TestCase ProFTPD::TestSuite::Child);
+use base qw(ProFTPD::TestSuite::Child);
 use strict;
 
-use File::Path qw(mkpath rmtree);
 use File::Spec;
 use IO::Handle;
 
@@ -26,6 +25,11 @@ my $TESTS = {
     test_class => [qw(forking rootprivs)],
   },
 
+  login_anonymous_with_delay_ok => {
+    order => ++$order,
+    test_class => [qw(forking rootprivs)],
+  },
+
   login_anonymous_user_alias_ok => {
     order => ++$order,
     test_class => [qw(forking rootprivs)],
@@ -43,7 +47,7 @@ my $TESTS = {
 
   login_regular_with_anon_defined_bug3307 => {
     order => ++$order,
-    test_class => [qw(forking)],
+    test_class => [qw(bug forking)],
   },
 
 };
@@ -56,29 +60,6 @@ sub list_tests {
   return testsuite_get_runnable_tests($TESTS);
 }
 
-sub set_up {
-  my $self = shift;
-  $self->{tmpdir} = testsuite_get_tmp_dir();
-
-  # Create temporary scratch dir
-  eval { mkpath($self->{tmpdir}) };
-  if ($@) {
-    my $abs_path = File::Spec->rel2abs($self->{tmpdir});
-    die("Can't create dir $abs_path: $@");
-  }
-}
-
-sub tear_down {
-  my $self = shift;
-
-  # Remove temporary scratch dir
-  if ($self->{tmpdir}) {
-    eval { rmtree($self->{tmpdir}) };
-  }
-
-  undef $self;
-}
-
 sub login_plaintext_fails {
   my $self = shift;
   my $tmpdir = $self->{tmpdir};
@@ -283,6 +264,134 @@ sub login_anonymous_ok {
   unlink($log_file);
 }
 
+sub login_anonymous_with_delay_ok {
+  my $self = shift;
+  my $tmpdir = $self->{tmpdir};
+
+  my $config_file = "$tmpdir/login.conf";
+  my $pid_file = File::Spec->rel2abs("$tmpdir/login.pid");
+  my $scoreboard_file = File::Spec->rel2abs("$tmpdir/login.scoreboard");
+
+  my $log_file = File::Spec->rel2abs('tests.log');
+
+  my $auth_user_file = File::Spec->rel2abs("$tmpdir/login.passwd");
+  my $auth_group_file = File::Spec->rel2abs("$tmpdir/login.group");
+
+  my ($config_user, $config_group) = config_get_identity();
+
+  my $anon_dir = File::Spec->rel2abs($tmpdir);
+  my $uid = 500;
+  my $gid = 500;
+
+  # Make sure that, if we're running as root, that the home directory has
+  # permissions/privs set for the account we create
+  if ($< == 0) {
+    unless (chmod(0755, $anon_dir)) {
+      die("Can't set perms on $anon_dir to 0755: $!");
+    }
+
+    unless (chown($uid, $gid, $anon_dir)) {
+      die("Can't set owner of $anon_dir to $uid/$gid: $!");
+    }
+  }
+
+  auth_user_write($auth_user_file, $config_user, 'foo', $uid, $gid,
+    '/tmp', '/bin/bash');
+  auth_group_write($auth_group_file, $config_group, $gid, $config_user);
+
+  my $delay_tab = File::Spec->rel2abs("$tmpdir/delay.tab");
+
+  my $config = {
+    PidFile => $pid_file,
+    ScoreboardFile => $scoreboard_file,
+    SystemLog => $log_file,
+    TraceLog => $log_file,
+    Trace => 'DEFAULT:10 privs:10',
+
+    AuthUserFile => $auth_user_file,
+    AuthGroupFile => $auth_group_file,
+
+    IfModules => {
+      'mod_delay.c' => {
+        DelayTable => $delay_tab,
+      },
+    },
+
+    Anonymous => {
+      $anon_dir => {
+        User => $config_user,
+        Group => $config_group,
+        UserAlias => "anonymous $config_user",
+        RequireValidShell => 'off',
+      },
+    },
+  };
+
+  my ($port, $user, $group) = config_write($config_file, $config);
+
+  # Open pipes, for use between the parent and child processes.  Specifically,
+  # the child will indicate when it's done with its test by writing a message
+  # to the parent.
+  my ($rfh, $wfh);
+  unless (pipe($rfh, $wfh)) {
+    die("Can't open pipe: $!");
+  }
+
+  my $ex;
+
+  # Fork child
+  $self->handle_sigchld();
+  defined(my $pid = fork()) or die("Can't fork: $!");
+  if ($pid) {
+    eval {
+      my $client = ProFTPD::TestSuite::FTP->new('127.0.0.1', $port);
+
+      # In parent process, login anonymously to server using a plaintext
+      # password which SHOULD work.
+      my ($resp_code, $resp_msg) = $client->login($config_user, 'ftp@nospam.org');
+
+      my $expected;
+
+      $expected = 230;
+      $self->assert($expected == $resp_code,
+        test_msg("Expected $expected, got $resp_code"));
+
+      $expected = 'Anonymous access granted, restrictions apply';
+      $self->assert($expected eq $resp_msg,
+        test_msg("Expected '$expected', got '$resp_msg'"));
+
+      $client->quit();
+    };
+
+    if ($@) {
+      $ex = $@;
+    }
+
+    $wfh->print("done\n");
+    $wfh->flush();
+
+  } else {
+    eval { server_wait($config_file, $rfh) };
+    if ($@) {
+      warn($@);
+      exit 1;
+    }
+
+    exit 0;
+  }
+
+  # Stop server
+  server_stop($pid_file);
+
+  $self->assert_child_ok($pid);
+
+  if ($ex) {
+    die($ex);
+  }
+
+  unlink($log_file);
+}
+
 sub login_anonymous_user_alias_ok {
   my $self = shift;
   my $tmpdir = $self->{tmpdir};
