@@ -1,7 +1,7 @@
 /*
  * ProFTPD - ftptop: a utility for monitoring proftpd sessions
  * Copyright (c) 2000-2002 TJ Saunders <tj@castaglia.org>
- * Copyright (c) 2003-2006 The ProFTPD Project team
+ * Copyright (c) 2003-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders and other respective copyright holders
  * give permission to link this program with OpenSSL, and distribute the
@@ -26,7 +26,7 @@
 /* Shows who is online via proftpd, in a manner similar to top.  Uses the
  * scoreboard files.
  *
- * $Id: ftptop.c,v 1.33 2007/05/22 21:21:25 castaglia Exp $
+ * $Id: ftptop.c,v 1.42 2011/05/23 20:46:20 castaglia Exp $
  */
 
 #define FTPTOP_VERSION "ftptop/0.9"
@@ -34,10 +34,13 @@
 #include "utils.h"
 
 #include <ctype.h>
-#include <signal.h>
 #include <stdlib.h>
 #include <time.h>
 
+#if defined(PR_USE_NLS) && defined(HAVE_LOCALE_H)
+# include <locale.h>
+#endif
+
 static const char *program = "ftptop";
 
 /* ncurses is preferred...*/
@@ -65,11 +68,11 @@ static const char *program = "ftptop";
 
 /* These are for displaying "PID S USER CLIENT SERVER TIME COMMAND" */
 #define FTPTOP_REG_HEADER_FMT	"%-5s %s %-8s %-20s %-15s %-4s %-*s\n"
-#define FTPTOP_REG_DISPLAY_FMT	"%-5u %s %-8.8s %-20.20s %-15s %-6.6s %4s %-*.*s\n"
+#define FTPTOP_REG_DISPLAY_FMT	"%-5u %s %-*.*s %-*.*s %-15s %-6.6s %4s %-*.*s\n"
 
 /* These are for displaying tranfer data: "PID S USER CLIENT KB/s %DONE" */
 #define FTPTOP_XFER_HEADER_FMT	"%-5s %s %-8s %-44s %-10s %-*s\n"
-#define FTPTOP_XFER_DISPLAY_FMT	"%-5u %s %-8.8s %-44.44s %-10.2f %-*s\n"
+#define FTPTOP_XFER_DISPLAY_FMT	"%-5u %s %-*.*s %-*.*s %-10.2f %-*.*s\n"
 
 #define FTPTOP_REG_ARG_MIN_SIZE		20
 #define FTPTOP_XFER_DONE_MIN_SIZE	6
@@ -83,6 +86,7 @@ static const char *program = "ftptop";
 #define FTPTOP_SHOW_DOWNLOAD		0x0001
 #define FTPTOP_SHOW_UPLOAD		0x0002
 #define FTPTOP_SHOW_IDLE		0x0004
+#define FTPTOP_SHOW_AUTH		0x0008
 #define	FTPTOP_SHOW_REG \
   (FTPTOP_SHOW_DOWNLOAD|FTPTOP_SHOW_UPLOAD|FTPTOP_SHOW_IDLE)
 #define FTPTOP_SHOW_RATES		0x0010
@@ -112,7 +116,8 @@ static void usage(void);
 
 static void clear_counters(void) {
 
-  if (ftp_sessions && ftp_nsessions) {
+  if (ftp_sessions &&
+      ftp_nsessions > 0) {
     register unsigned int i = 0;
 
     for (i = 0; i < ftp_nsessions; i++)
@@ -156,6 +161,64 @@ static char *calc_percent_done(off_t size, off_t done) {
   return sbuf;
 }
 
+/* Given a NUL-terminated string -- possibly UTF8-encoded -- and a maximum
+ * CHARACTER length, return the number of bytes in the string which can fit in
+ * that max length  without truncating a character.  This is needed since UTF8
+ * characters are variable-width.
+ */
+static int str_getscreenlen(const char *str, size_t max_chars) {
+#ifdef PR_USE_NLS
+  register unsigned int i = 0;
+  int nbytes = 0, nchars = 0;
+
+  while (str[i] > 0 &&
+         i < max_chars) {
+ascii:
+    i++;
+    nbytes++;
+    nchars++;
+  }
+
+  while (str[i] &&
+         nchars < max_chars) {
+    size_t len;
+
+    if (str[i] > 0) {
+      goto ascii;
+    }
+
+    len = 0;
+
+    switch (str[i] & 0xF0) {
+      case 0xE0:
+        len = 3;
+        break;
+
+      case 0xF0:
+        len = 4;
+        break;
+
+      default:
+        len = 2;
+        break;
+    }
+
+    /* Increment the index with the given length, but increment the
+     * character count only one.
+     */
+
+    i += len;
+    nbytes += len;
+    nchars++;
+  }
+
+  return nbytes;
+#else
+  /* No UTF8 support in this proftpd build; just return the max characters. */
+  return (int) max_chars;
+#endif /* !PR_USE_NLS */
+}
+
 /* Borrowed from ftpwho.c */
 static const char *show_time(time_t *i) {
   time_t now = time(NULL);
@@ -168,11 +231,13 @@ static const char *show_time(time_t *i) {
   memset(sbuf, '\0', sizeof(sbuf));
   l = now - *i;
 
-  if (l < 3600)
+  if (l < 3600) {
     snprintf(sbuf, sizeof(sbuf), "%lum%lus",(l / 60),(l % 60));
-  else
+
+  } else {
     snprintf(sbuf, sizeof(sbuf), "%luh%lum",(l / 3600),
-    ((l - (l / 3600) * 3600) / 60));
+      ((l - (l / 3600) * 3600) / 60));
+  }
 
   return sbuf;
 }
@@ -215,93 +280,31 @@ static const char *show_ftpd_uptime(void) {
   upminutes = upminutes % 60;
 
   if (uphours) {
-    pos += snprintf(buf + pos, sizeof(buf) - pos, "%2d hr%s %02d min", uphours,
+    snprintf(buf + pos, sizeof(buf) - pos, "%2d hr%s %02d min", uphours,
       (uphours != 1) ? "s" : "", upminutes);
 
   } else {
-    pos += snprintf(buf + pos, sizeof(buf) - pos, "%d min", upminutes);
+    snprintf(buf + pos, sizeof(buf) - pos, "%d min", upminutes);
   }
 
   return buf;
 }
 
-/* scan_config_file() is a kludge for 1.2 which does a very simplistic attempt
- * at determining what the "ScoreboardFile" directive is set to.  It will be
- * replaced in 1.3 with the abstracted configure system (hopefully).
- */
-static void scan_config_file(void) {
-  FILE *fp = NULL;
-  char buf[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
-  char *cp, *file = NULL;
-
-  if (!config_filename || (fp = fopen(config_filename,"r")) == NULL)
-    return;
-
-  while (!file && fgets(buf, sizeof(buf) - 1, fp)) {
-    int i = strlen(buf);
-
-    if (i && buf[i - 1] == '\n')
-      buf[i-1] = '\0';
-
-    for (cp = buf; *cp && isspace((int) *cp); cp++);
-
-    if (*cp == '#' || !*cp)
-      continue;
-
-    i = strlen("ScoreboardFile");
-
-    if (strncasecmp(cp, "ScoreboardFile", i) != 0)
-      continue;
-
-    /* Found it! */
-    cp += i;
-
-    /* strip whitespace */
-    while (*cp && isspace((int) *cp))
-      cp++;
-
-    file = cp;
-
-    /* If the scoreboard file argument is quoted, dequote */
-    if (*cp == '"') {
-      char *src = cp;
-
-      cp++;
-      file++;
-
-      while (*++src) {
-        switch (*src) {
-          case '\\':
-            if (*++src)
-              *cp++ = *src;
-            break;
-
-          case '"':
-            src++;
-            break;
-
-          default:
-            *cp++ = *src;
-        }
-      }
-
-      *cp = '\0';
-    }
-  }
-
-  fclose(fp);
-
-  /* If we got something out of all this, go ahead and set it. */
-  if (file)
-    util_set_scoreboard(file);
-}
-
 static void process_opts(int argc, char *argv[]) {
   int optc = 0;
-  const char *prgopts = "DS:d:f:hIiUV";
+  const char *prgopts = "AaDS:d:f:hIiUV";
 
   while ((optc = getopt(argc, argv, prgopts)) != -1) {
     switch (optc) {
+      case 'A':
+        display_mode = 0U;
+        display_mode |= FTPTOP_SHOW_AUTH;
+        break;
+
+      case 'a':
+        display_mode &= ~FTPTOP_SHOW_AUTH;
+        break;
+ 
       case 'D':
         display_mode = 0U;
         display_mode |= FTPTOP_SHOW_DOWNLOAD;
@@ -370,7 +373,13 @@ static void process_opts(int argc, char *argv[]) {
    * incorrect, try the config file kludge.
    */
   if (check_scoreboard_file() < 0) {
-    scan_config_file();
+    char *path;
+
+    path = util_scan_config(config_filename, "ScoreboardFile");
+    if (path) {
+      util_set_scoreboard(path);
+      free(path);
+    }
 
     if (check_scoreboard_file() < 0) {
       fprintf(stderr, "%s: %s\n", util_get_scoreboard(), strerror(errno));
@@ -389,75 +398,108 @@ static void read_scoreboard(void) {
   static char buf[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
   pr_scoreboard_entry_t *score = NULL;
 
-  if ((ftp_sessions = calloc(chunklen, sizeof(char *))) == NULL)
+  ftp_sessions = calloc(chunklen, sizeof(char *));
+  if (ftp_sessions == NULL)
     exit(1);
 
   if (scoreboard_open() < 0)
     return;
 
   /* Iterate through the scoreboard. */
-  while ((score = util_scoreboard_read_entry()) != NULL) {
+  while ((score = util_scoreboard_entry_read()) != NULL) {
 
     /* Default status: "A" for "authenticating" */
     char *status = "A";
 
     /* If a ServerName was given, skip unless the scoreboard entry matches. */
-    if (server_name && strcmp(server_name, score->sce_server_label) != 0)
+    if (server_name &&
+        strcmp(server_name, score->sce_server_label) != 0)
       continue;
 
     /* Clear the buffer for this run. */
     memset(buf, '\0', sizeof(buf));
 
-    /* Determine the status symbol to display. */
-    if (strcmp(score->sce_cmd, "idle") == 0) {
-      status = "I";
-      ftp_nidles++;
+    /* Has the user authenticated yet? */
+    if (strcmp(score->sce_user, "(none)") != 0) {
 
-      if (display_mode != FTPTOP_SHOW_RATES &&
-          !(display_mode & FTPTOP_SHOW_IDLE))
-        continue;
+      /* Determine the status symbol to display. */
+      if (strcmp(score->sce_cmd, "idle") == 0) {
+        status = "I";
+        ftp_nidles++;
 
-    } else if (strcmp(score->sce_cmd, "RETR") == 0) {
-      status = "D";
-      ftp_ndownloads++;
+        if (display_mode != FTPTOP_SHOW_RATES &&
+            !(display_mode & FTPTOP_SHOW_IDLE))
+          continue;
 
-      if (display_mode != FTPTOP_SHOW_RATES &&
-          !(display_mode & FTPTOP_SHOW_DOWNLOAD))
-        continue;
+      } else if (strcmp(score->sce_cmd, "RETR") == 0 ||
+                 strcmp(score->sce_cmd, "READ") == 0 ||
+                 strcmp(score->sce_cmd, "scp download") == 0) {
+        status = "D";
+        ftp_ndownloads++;
 
-    } else if (strcmp(score->sce_cmd, "STOR") == 0 ||
-        strcmp(score->sce_cmd, "APPE") == 0 ||
-        strcmp(score->sce_cmd, "STOU") == 0) {
-      status = "U";
-      ftp_nuploads++;
+        if (display_mode != FTPTOP_SHOW_RATES &&
+            !(display_mode & FTPTOP_SHOW_DOWNLOAD))
+          continue;
 
-      if (display_mode != FTPTOP_SHOW_RATES &&
-          !(display_mode & FTPTOP_SHOW_UPLOAD))
-        continue;
+      } else if (strcmp(score->sce_cmd, "STOR") == 0 ||
+                 strcmp(score->sce_cmd, "APPE") == 0 ||
+                 strcmp(score->sce_cmd, "STOU") == 0 ||
+                 strcmp(score->sce_cmd, "WRITE") == 0 ||
+                 strcmp(score->sce_cmd, "scp upload") == 0) {
+        status = "U";
+        ftp_nuploads++;
 
-    } else if (strcmp(score->sce_cmd, "LIST") == 0 ||
-        strcmp(score->sce_cmd, "NLST") == 0)
-      status = "L";
+        if (display_mode != FTPTOP_SHOW_RATES &&
+            !(display_mode & FTPTOP_SHOW_UPLOAD))
+          continue;
+
+      } else if (strcmp(score->sce_cmd, "LIST") == 0 ||
+                 strcmp(score->sce_cmd, "NLST") == 0 ||
+                 strcmp(score->sce_cmd, "MLST") == 0 ||
+                 strcmp(score->sce_cmd, "MLSD") == 0 ||
+                 strcmp(score->sce_cmd, "READDIR") == 0) {
+        status = "L";
+      }
+
+    } else {
+      status = "A";
+
+      /* Overwrite the "command", for display purposes */
+      util_sstrncpy(score->sce_cmd, "(authenticating)", sizeof(score->sce_cmd));
+    }
 
     if (display_mode != FTPTOP_SHOW_RATES) {
+      int user_namelen, client_namelen, cmd_arglen;
+
+      user_namelen = str_getscreenlen(score->sce_user, 8);
+      client_namelen = str_getscreenlen(score->sce_client_name, 20);
+      cmd_arglen = str_getscreenlen(score->sce_cmd_arg, FTPTOP_REG_ARG_SIZE);
+
       snprintf(buf, sizeof(buf), FTPTOP_REG_DISPLAY_FMT,
-        (unsigned int) score->sce_pid, status, score->sce_user,
-        score->sce_client_name, score->sce_server_addr,
+        (unsigned int) score->sce_pid, status,
+        user_namelen, user_namelen, score->sce_user,
+        client_namelen, client_namelen, score->sce_client_name,
+        score->sce_server_addr,
         show_time(&score->sce_begin_session), score->sce_cmd,
-       FTPTOP_REG_ARG_SIZE, FTPTOP_REG_ARG_SIZE, score->sce_cmd_arg);
+        cmd_arglen, cmd_arglen, score->sce_cmd_arg);
       buf[sizeof(buf)-1] = '\0';
 
     } else {
+      int user_namelen, client_namelen;
+
+      user_namelen = str_getscreenlen(score->sce_user, 8);
+      client_namelen = str_getscreenlen(score->sce_client_name, 44);
 
       /* Skip sessions unless they are actually transferring data */
       if (*status != 'U' && *status != 'D')
         continue;
 
       snprintf(buf, sizeof(buf), FTPTOP_XFER_DISPLAY_FMT,
-        (unsigned int) score->sce_pid, status, score->sce_user,
-        score->sce_client_name,
+        (unsigned int) score->sce_pid, status,
+        user_namelen, user_namelen, score->sce_user,
+        client_namelen, client_namelen, score->sce_client_name,
         (score->sce_xfer_len / 1024.0) / (score->sce_xfer_elapsed / 1000),
-        FTPTOP_XFER_DONE_SIZE,
+        FTPTOP_XFER_DONE_SIZE, FTPTOP_XFER_DONE_SIZE,
         *status == 'D' ?
           calc_percent_done(score->sce_xfer_size, score->sce_xfer_done) :
           "(n/a)");
@@ -467,14 +509,21 @@ static void read_scoreboard(void) {
     /* Make sure there is enough memory allocated in the session list.
      * Allocate more if needed.
      */
-    if (ftp_nsessions && ftp_nsessions % chunklen == 0) {
-      if ((ftp_sessions = realloc(ftp_sessions,
-          (ftp_nsessions + chunklen) * sizeof(char *))) == NULL)
+    if (ftp_nsessions &&
+        ftp_nsessions % chunklen == 0) {
+      ftp_sessions = realloc(ftp_sessions,
+        (ftp_nsessions + chunklen) * sizeof(char *));
+
+      if (ftp_sessions == NULL) {
         exit(1);
+      }
     }
 
-    if ((ftp_sessions[ftp_nsessions] = calloc(1, strlen(buf) + 1)) == NULL)
+    ftp_sessions[ftp_nsessions] = calloc(1, strlen(buf) + 1);
+    if (ftp_sessions[ftp_nsessions] == NULL) {
       exit(1);
+    }
+
     strncpy(ftp_sessions[ftp_nsessions++], buf, strlen(buf) + 1);
   }
 
@@ -488,7 +537,8 @@ static void scoreboard_close(void) {
 static int scoreboard_open(void) {
   int res = 0;
 
-  if ((res = util_open_scoreboard(O_RDONLY)) < 0) {
+  res = util_open_scoreboard(O_RDONLY);
+  if (res < 0) {
     switch (res) {
       case UTIL_SCORE_ERR_BAD_MAGIC:
         fprintf(stderr, "%s: error opening scoreboard: bad/corrupted file\n",
@@ -546,20 +596,24 @@ static void show_sessions(void) {
 
   attron(A_REVERSE);
 
-  if (display_mode != FTPTOP_SHOW_RATES)
+  if (display_mode != FTPTOP_SHOW_RATES) {
     printw(FTPTOP_REG_HEADER_FMT, "PID", "S", "USER", "CLIENT", "SERVER",
       "TIME", FTPTOP_REG_ARG_SIZE, "COMMAND");
-  else
+
+  } else {
     printw(FTPTOP_XFER_HEADER_FMT, "PID", "S", "USER", "CLIENT", "KB/s", FTPTOP_XFER_DONE_SIZE, "%DONE");
+  }
 
   attroff(A_REVERSE);
 
   /* Write out the scoreboard entries. */
-  if (ftp_sessions && ftp_nsessions) {
+  if (ftp_sessions &&
+      ftp_nsessions > 0) {
     register unsigned int i = 0;
 
-    for (i = 0; i < ftp_nsessions; i++)
+    for (i = 0; i < ftp_nsessions; i++) {
       printw("%s", ftp_sessions[i]);
+    }
   }
 
   wrefresh(stdscr);
@@ -571,10 +625,12 @@ static void toggle_mode(void) {
   if (cached_mode == 0)
     cached_mode = display_mode;
 
-  if (display_mode != FTPTOP_SHOW_RATES)
+  if (display_mode != FTPTOP_SHOW_RATES) {
     display_mode = FTPTOP_SHOW_RATES;
-  else
+
+  } else {
     display_mode = cached_mode;
+  }
 }
 
 static void show_version(void) {
@@ -584,6 +640,8 @@ static void show_version(void) {
 
 static void usage(void) {
   fprintf(stdout, "usage: ftptop [options]\n\n");
+  fprintf(stdout, "\t-A      \t\tshow only authenticatng sessions\n");
+  fprintf(stdout, "\t-a      \t\tignores authenticating connections when listing\n");
   fprintf(stdout, "\t-D      \t\tshow only downloading sessions\n");
   fprintf(stdout, "\t-d <num>\t\trefresh delay in seconds\n");
   fprintf(stdout, "\t-f      \t\tconfigures the ScoreboardFile to use\n");
@@ -619,6 +677,10 @@ int main(int argc, char *argv[]) {
   signal(SIGINT, finish);
   signal(SIGTERM, finish);
 
+#if defined(PR_USE_NLS) && defined(HAVE_LOCALE_H)
+  (void) setlocale(LC_ALL, "");
+#endif
+
   /* Initialize the display. */
   initscr();
   cbreak();
@@ -644,11 +706,13 @@ int main(int argc, char *argv[]) {
 #endif
 
     if (c != -1) {
-      if (tolower(c) == 'q')
-        finish(0);
+      if (tolower(c) == 'q') {
+        break;
+      }
 
-      if (tolower(c) == 't')
+      if (tolower(c) == 't') {
         toggle_mode();
+      }
     }
 
     show_sessions();
@@ -656,6 +720,7 @@ int main(int argc, char *argv[]) {
 
   /* done */
   finish(0);
+  return 0;
 }
 
 #else /* defined(HAVE_CURSES) || defined(HAVE_NCURSES) */
